;;; $Id: 3-6-letrec3.scm,v 1.6 2005/03/21 21:58:37 leavens Exp $
;;; adapted from 3-5.scm and 3-6-letrec3.scm in the EOPL2e distribution

(let ((time-stamp "Time-stamp: <2000-12-12 19:04:49 wand>"))
  (eopl:printf
    "letrec3.scm -- AST rep of environments with circular links ~a~%"
    (substring time-stamp 13 29)))

;;;;;;;;;;;;;;;; environments ;;;;;;;;;;;;;;;;

;; This can't be a module, because it uses procval->expressed, closure, etc.

;; type predicate
(deftype environment? (type-predicate-for environment))

;; constructors
(deftype empty-env (-> () environment))
(deftype extend-env
  (-> ((list-of symbol) (list-of Expressed-Value) environment) environment))
(deftype extend-env-recursively            ; for section 3.6
  (-> ((list-of symbol) (list-of (list-of symbol)) (list-of expression)
       environment)
      environment))

;; observers
(deftype apply-env
  (-> (environment symbol) Expressed-Value))
(deftype defined-in-env?                   ; added
  (-> (environment symbol) boolean))

(require (lib "list-index-find.scm" "lib342"))

(define-datatype environment environment?
  (empty-env)
  (extended-env-record
    (syms (list-of symbol?))
    (vec vector?)              ; can use this for anything.
    (env environment?))
  )

(define extend-env
  (lambda (syms vals env)
    (extended-env-record syms (list->vector vals) env)))

(define extend-env-recursively
  (lambda (proc-names idss bodies old-env)
    (let ((len (length proc-names)))
      (let ((vec (make-vector
		  len
		  (procval->expressed (closure '() (lit-exp 0) (empty-env))))))
        (let ((env (extended-env-record
                     proc-names vec old-env)))
          (for-each
            (lambda (pos ids body)
              (vector-set! vec pos
			   (procval->expressed (closure ids body env))))
            (iota len) idss bodies)
          env)))))

(define apply-env
  (lambda (env sym)
    (cases environment env
      (empty-env ()
        (eopl:error 'apply-env "No binding for ~s" sym))
      (extended-env-record (syms vals env)
        (let ((position (list-index sym syms)))
          (if (<= 0 position)
              (vector-ref vals position)
              (apply-env env sym)))))))

(define defined-in-env?
  (lambda (env sym)
    (cases environment env
      (empty-env () #f)
      (extended-env-record (syms vals env)
        (let ((position (list-index sym syms)))
          (or (<= 0 position)
              (defined-in-env? env sym)))))))

(deftype iota (-> (number) (list-of number)))
(define iota
  (lambda (end)
    (let loop ((next 0))
      (if (>= next end) '()
        (cons next (loop (+ 1 next)))))))
