;;; $Id: 3-9.scm,v 1.11 2005/05/04 04:12:24 leavens Exp $

(let ((time-stamp "Time-stamp: <2001-05-10 16:22:06 dfried>"))
  (eopl:printf "3-9.scm: language with statements ~a~%"
    (substring time-stamp 13 29)))

;;; this file adapted from one written by Will Clinger.  This may
;;; differ in some details from the code in section 3.9 and in other
;;; interpreters. 

;;;;;;;;;;;;;;;; top level and tests ;;;;;;;;;;;;;;;;

(deftype run (-> (string) datum))
(define run
  (lambda (string)
    (begin
      (execute-program (scan&parse string))
      (has-type datum #t))))

(deftype run-all (-> () void))
(define run-all
  (lambda ()
    (run-experiment run use-execution-outcome
      '(lang3-9) all-tests)))

(deftype run-one (-> (symbol) void))
(define run-one
  (lambda (test-name)
    (run-test run test-name)))

;; needed for testing
(deftype equal-external-reps? (-> (datum datum) boolean))
(define equal-external-reps? equal?)

;;;;;;;;;;;;;;;; grammatical specification ;;;;;;;;;;;;;;;;

(define the-lexical-spec
  '((whitespace (whitespace) skip)
    (comment ("%" (arbno (not #\newline))) skip)
    (identifier
      (letter (arbno (or letter digit "_" "-" "?")))
      symbol)
    (number (digit (arbno digit)) number)))

(define the-grammar
  '((program
      (statement)
      a-program)
    
    (statement
      (identifier "=" expression)
      assign-statement)
    (statement
      ("print" "(" expression ")")
      print-statement)
    (statement
      ("{" (separated-list statement ";") "}")
      compound-statement)
    (statement
      ("if" expression statement statement)
      if-statement)
    (statement
      ("while" expression "do" statement)
      while-statement)
    (statement
      ("var" (separated-list identifier ",") ";" statement)
      block-statement)
    
    (expression (number) lit-exp)
    (expression (identifier) var-exp)   
    (expression
      (primitive "(" (separated-list expression ",") ")")
      primapp-exp)
    (expression
      ("if" expression "then" expression "else" expression)
      if-exp)
    (expression
      ("let" (arbno  identifier "=" expression) "in" expression)
      let-exp)
    (expression
      ("proc" "(" (separated-list identifier ",") ")" expression)
      proc-exp)
    (expression
      ("(" expression (arbno expression) ")")
      app-exp)
    (expression
      ("begin" expression (arbno ";" expression) "end")
      begin-exp)
    (expression                         
      ("letrec"
        (arbno identifier "(" (separated-list identifier ",") ")"
          "=" expression)
        "in" expression)
      letrec-exp)
    (expression ("set" identifier "=" expression) varassign-exp)

    (primitive ("+")     add-prim)
    (primitive ("-")     subtract-prim)
    (primitive ("*")     mult-prim)
    (primitive ("add1")  incr-prim)
    (primitive ("sub1")  decr-prim)
    (primitive ("zero?")     zero-test-prim)
    (primitive ("equal?")    equal-prim)
    (primitive ("greater?")  greater-prim)
    (primitive ("less?")     less-prim)
    ))

(deftype show-the-datatypes (-> () (list-of datum)))
(define show-the-datatypes
  (lambda ()
    (sllgen:make-define-datatypes the-lexical-spec the-grammar)
    (sllgen:list-define-datatypes the-lexical-spec the-grammar)))

;; The following is generated by calling show-the-datatypes
;; (and then removing the outer lists, indenting, and renaming fields).
(define-datatype
  program
  program?
  (a-program (stmt statement?)))

(define-datatype
  statement
  statement?
  (assign-statement
   (id symbol?)
   (exp expression?))
  (print-statement
   (exp expression?))
  (compound-statement
   (stmts (list-of statement?)))
  (if-statement
   (exp expression?)
   (true-stmt statement?)
   (false-stmt statement?))
  (while-statement
   (exp expression?)
   (stmt statement?))
  (block-statement
   (ids (list-of symbol?))
   (body statement?)))

(define-datatype
  expression
  expression?
  (lit-exp (datum number?))
  (var-exp (id symbol?))
  (primapp-exp
   (prim primitive?)
   (rands (list-of expression?)))
  (if-exp
   (test-exp expression?)
   (true-exp expression?)
   (false-exp expression?))
  (let-exp
   (ids (list-of symbol?))
   (rands (list-of expression?))
   (body expression?))
  (proc-exp
   (ids (list-of symbol?))
   (body expression?))
  (app-exp
   (rator expression?)
   (rands (list-of expression?)))
  (begin-exp
   (first expression?)
   (rest (list-of expression?)))
  (letrec-exp
   (proc-names (list-of symbol?))
   (idss (list-of (list-of symbol?)))
   (bodies (list-of expression?))
   (letrec-body expression?))
  (varassign-exp
   (id symbol?)
   (rhs-exp expression?))
  )

(define-datatype
   primitive
   primitive?
   (add-prim)
   (subtract-prim)
   (mult-prim)
   (incr-prim)
   (decr-prim)
   (zero-test-prim)
   (equal-prim)
   (greater-prim)
   (less-prim))

(deftype scan&parse (-> (string) program))
(define scan&parse
  (sllgen:make-string-parser the-lexical-spec the-grammar))

(deftype just-scan (forall (TOK) (-> (string) (list-of TOK))))
(define just-scan
  (sllgen:make-string-scanner the-lexical-spec the-grammar))

(deftype read-eval-print (-> () poof))
(define read-eval-print 
  (sllgen:make-rep-loop  "--> "
    (lambda (pgm) (execute-program pgm))   ; wrapped to avoid load
                                        ; dependency 
    (sllgen:make-stream-parser 
      the-lexical-spec
      the-grammar)))

;;; Domains for this interpreter:
;;;
;;; Expressed-Value = Number + ProcVal + List(Expressed-Value)
;;; Denoted-Value = Ref(Expressed-Value)

;;; ADTs used

(require (lib "environment-3-7.scm" "lib342"))

;;;;;;;;;;;;;;;; ProcVal ;;;;;;;;;;;;;;;;

(deftype procval? (type-predicate-for procval))
(deftype closure (-> ((list-of symbol) expression environment) procval))
(deftype apply-procval
  (-> (procval (list-of Expressed-Value)) Expressed-Value))

(define-datatype procval procval?
  (closure 
    (ids (list-of symbol?)) 
    (body expression?)
    (env environment?)))

(define apply-procval
  (lambda (proc args)
    (cases procval proc
      (closure (ids body env)
        (eval-expression body (extend-env ids args env))))))

;;;;;;;;;;;;;;;; references ;;;;;;;;;;;;;;;;;;

(require (lib "reference-3-7.scm" "lib342"))

;;; Anonymous figure : page 80

(deftype true-value? (-> (Expressed-Value) boolean))
(define true-value?
  (lambda (x)
    (not (zero? (expressed->number x)))))

;;;;;;;;; Expressed-Value ;;;;;;;;;;;;;;;;;;;;;;;;;

;; upcasts
(deftype number->expressed (-> (number) Expressed-Value))
(deftype procval->expressed (-> (Procval) Expressed-Value))
(deftype list->expressed (-> ((list-of Expressed-Value)) Expressed-Value))

;; downcasts
(deftype expressed->number (-> (Expressed-Value) number))
(deftype expressed->procval (-> (Expressed-Value) Procval))
(deftype expressed->list (-> (Expressed-Value) (list-of Expressed-Value)))

;; debugging
(deftype expressed->printable (-> (Expressed-Value) datum))

;; tests
(deftype number->expressed? (-> (Expressed-Value) boolean))
(deftype procval->expressed? (-> (Expressed-Value) boolean))
(deftype list->expressed? (-> (Expressed-Value) boolean))

(define-datatype Expressed-Value expval?
  (number->expressed (num number?))
  (procval->expressed (pv procval?))
  (list->expressed (lst (list-of expval?))))

(define expressed->number
  (lambda (ev)
    (cases Expressed-Value ev
      (number->expressed (num) num)
      (else (error "expressed->number passed non-number argument: " ev)))))

(define expressed->procval
  (lambda (ev)
    (cases Expressed-Value ev
      (procval->expressed (pv) pv)
      (else (error "expressed->procval passed non-procval argument: " ev)))))

(define expressed->list
  (lambda (ev)
    (cases Expressed-Value ev
      (list->expressed (lst) lst)
      (else (error "expressed->list passed non-list argument: " ev)))))

(define expressed->printable
  (lambda (ev)
    (cases Expressed-Value ev
      (number->expressed (num) (has-type datum num))
      (procval->expressed (pv)
        (cases procval pv
          (closure (ids body env)
	   (has-type datum (list 'closure ids body "<env>")))))
      (list->expressed (lst)
	 (has-type datum (map expressed->printable lst))))))

(define number->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (number->expressed (num) #t)
      (else #f))))

(define procval->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (procval->expressed (pv) #t)
      (else #f))))

(define list->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (list->expressed (lst) #t)
      (else #f))))

;; adaptation of environment-3-7 to this file
(deftype extend-env-recursively
  (-> ((list-of symbol) (list-of (list-of symbol)) (list-of expression)
       environment)
      environment))
(define extend-env-recursively
  (extend-env-recursively-maker
   (lambda (ids body env)
     (procval->expressed (closure ids body env)))
   (procval->expressed (closure '() (lit-exp 0) (empty-env)))))

;;; truth and falsity ;;;

(define truth (number->expressed 1))
(define falsity (number->expressed 0))


;;;;;;;;;;;;;;;; the interpreter ;;;;;;;;;;;;;;;;

(deftype execute-program (-> (program) void))
(define execute-program 
  (lambda (pgm)
    (cases program pgm
      (a-program (statement)
        (execute-statement statement (init-env))))))

(deftype execute-statement (-> (statement environment) void))
(define execute-statement
  (lambda (stmt env)
    (cases statement stmt
      (assign-statement (id exp)
        (setref!
          (apply-env-ref env id)
          (eval-expression exp env)))
      (print-statement (exp)
        (write (expressed->printable (eval-expression exp env)))
        (newline))
      (compound-statement (stmts)
        (for-each
          (lambda (stmt)
            (execute-statement stmt env))
          stmts))
      (if-statement (exp true-stmt false-stmt)
        (if (true-value? (eval-expression exp env))
          (execute-statement true-stmt env)
          (execute-statement false-stmt env)))
      (while-statement (exp stmt)
        (let loop ()
          (if (true-value? (eval-expression exp env))
            (begin
              (execute-statement stmt env)
              (loop)))))
      (block-statement (ids stmt)
        (execute-statement stmt
          (extend-env ids (map (lambda (id) (number->expressed 0)) ids) env)))
      )))

(deftype eval-expression (-> (expression environment) Expressed-Value))
(define eval-expression 
  (lambda (exp env)
    (cases expression exp
      (lit-exp (datum)
	(number->expressed datum))
      (var-exp (id)
	(apply-env env id))
      (primapp-exp (prim rands)
        (let ((args (eval-rands rands env)))
          (apply-primitive prim args)))
      (if-exp (test-exp true-exp false-exp)
        (if (true-value? (eval-expression test-exp env))
          (eval-expression true-exp env)
          (eval-expression false-exp env)))
      (begin-exp (exp1 exps)
        (let loop ((acc (eval-expression exp1 env))
                   (exps exps))
          (if (null? exps) acc
            (loop (eval-expression (car exps) env) (cdr exps)))))
      (let-exp (ids rands body)
        (let ((args (eval-rands rands env)))
          (eval-expression
	   body
	   (extend-env ids args env))))
      (proc-exp (ids body) (procval->expressed (closure ids body env)))
      (app-exp (rator rands)
        (let ((proc (eval-expression rator env))
              (args (eval-rands rands env)))
          (if (procval->expressed? proc)
            (apply-procval (expressed->procval proc) args)
            (eopl:error 'eval-expression
              "Attempt to apply non-procedure ~s" proc))))
      (letrec-exp (proc-names idss bodies letrec-body)
        (eval-expression letrec-body
          (extend-env-recursively proc-names idss bodies env)))     
      (varassign-exp (id rhs-exp)
        (begin
          (setref! 
            (apply-env-ref env id)
            (eval-expression rhs-exp env))
          (number->expressed 1)))
      (else (eopl:error 'eval-expression "Missing case:~s" exp))
      )))

(deftype eval-rands
  (-> ((list-of expression) environment) (list-of Expressed-Value)))
(define eval-rands
  (lambda (rands env)
    (map (lambda (x) (eval-rand x env)) rands)))

(deftype eval-rand (-> (expression environment) Expressed-Value))
(define eval-rand
  (lambda (rand env)
    (eval-expression rand env)))

(deftype apply-primitive
  (-> (primitive (list-of Expressed-Value)) Expressed-Value))
(define apply-primitive
  (lambda (prim args)
    (define (check arity)
      (if (not (= arity (length args)))
          (eopl:error 'apply-primitive
                 "wrong number of arguments to primitive ~s"
                 prim)))
    (cases primitive prim
      (add-prim      () (check 2)
		        (number->expressed
			 (+ (expressed->number (car args))
			    (expressed->number (cadr args)))))
      (subtract-prim () (check 2)
		        (number->expressed
			 (- (expressed->number (car args))
			    (expressed->number (cadr args)))))
      (mult-prim     () (check 2)
		        (number->expressed
			 (* (expressed->number (car args))
			    (expressed->number (cadr args)))))
      (incr-prim     () (check 1)
		        (number->expressed
			 (+ (expressed->number (car args)) 1)))
      (decr-prim     () (check 1)
		        (number->expressed
			 (- (expressed->number (car args)) 1)))
      (zero-test-prim() (check 1)
		        (if (zero? (expressed->number (car args)))
			     truth falsity))
      (equal-prim    () (check 2)
                        (if (= (expressed->number (car args))
			       (expressed->number (cadr args)))
			    truth falsity))
      (greater-prim  () (check 2)
                        (if (> (expressed->number (car args))
			       (expressed->number (cadr args)))
			    truth falsity))
      (less-prim     () (check 2)
                        (if (< (expressed->number (car args))
			       (expressed->number (cadr args)))
			    truth falsity))
      )))

(deftype init-env (-> () environment))
(define init-env 
  (lambda ()
    (extend-env
      '(i v x)
      (map number->expressed '(1 5 10))
      (empty-env))))
