;;; $Id: arith-expr-examples.scm,v 1.3 2004/03/29 17:45:42 leavens Exp $

(load-quietly-from-lib "arith-expr.scm")

(deftype copy-arith-expr (-> (arith-expr) arith-expr))
(define copy-arith-expr
  (lambda (ae)
    ;; ENSURES: result is a fresh arith-expr that is just like ae
    (cases arith-expr ae
       (lit-exp (datum)
	  (lit-exp datum))
       (binary-op-exp (left-arg op right-arg)
	  (binary-op-exp (copy-arith-expr left-arg)
			 (copy-arith-expr-infix-op op)
			 (copy-arith-expr right-arg)))
       (unary-minus-exp (arg)
	  (unary-minus-exp (copy-arith-expr arg))))))

(deftype copy-arith-expr-infix-op (-> (infix-op) infix-op))
(define copy-arith-expr-infix-op
  (lambda (op)
    ;; ENSURES: result is a fresh infix-op that is just like op.
    (cases infix-op op
       (plus-op () (plus-op))
       (minus-op () (minus-op))
       (times-op () (times-op))
       (exponent-op () (exponent-op)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftype change-plus-minus (-> (arith-expr) arith-expr))
(define change-plus-minus
  (lambda (ae)
    ;; ENSURES: result is an arith-expr like ae but with each plus-op
    ;; changed to a minus-op, and vice versa.
    (cases arith-expr ae
       (binary-op-exp (left-arg op right-arg)
	  (binary-op-exp (change-plus-minus left-arg)
			 (change-plus-minus-infix-op op)
			 (change-plus-minus right-arg)))
       (unary-minus-exp (arg)
	  (unary-minus-exp (change-plus-minus arg)))
       (else ae))))

(deftype change-plus-minus-infix-op (-> (infix-op) infix-op))
(define change-plus-minus-infix-op
  (lambda (op)
    ;; ENSURES: result is an infix-op like op, but with each
    ;; plus-op replaced by minus-op and vice versa.
    (cases infix-op op
       (plus-op () (minus-op))
       (minus-op () (plus-op))
       (else op))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(deftype subst-literal (-> (number number arith-expr) arith-expr))
(define subst-literal
  (lambda (new old ae)
    ;; ENSURES: result is an arith-expr like ae but with each
    ;; occurrence of the literal old replace by new.
    (cases arith-expr ae
       (binary-op-exp (left-arg op right-arg)
	  (binary-op-exp (subst-literal new old left-arg)
			 op
			 (subst-literal new old right-arg)))
       (unary-minus-exp (arg)
	  (unary-minus-exp (subst-literal new old arg)))
       (lit-exp (datum)
	  (if (= datum old)
	      (lit-exp new)
	      ae)))))
