;;; $Id: arith-expr.scm,v 1.13 2005/03/03 00:18:41 leavens Exp $

;;; Helping procedures for the following grammar.
;;;    <arith-expr> ::=         
;;;        <number>                        "lit-exp (datum)"
;;;     | ( <arith-expr>                   "binary-op-exp (left-arg
;;;         <infix-op>                                     op
;;;         <arith-expr> )                                 right-arg)"
;;;     | (- <arith-expr>)                 "unary-minus-exp (arg)"
				
;;;   <infix-op> ::=			
;;;        +                               "plus-op ()"
;;;      | -                               "minus-op ()"
;;;      | *                               "times-op ()"
;;;      | **                              "exponent-op ()"

(define-datatype arith-expr arith-expr?
  (lit-exp
   (datum number?))
  (binary-op-exp
   (left-arg arith-expr?) (op infix-op?) (right-arg arith-expr?))
  (unary-minus-exp
   (arg arith-expr?)))

(define-datatype infix-op infix-op?
  (plus-op) (minus-op) (times-op) (exponent-op))

(deftype parse-arith-expr (-> (datum) arith-expr))
(define parse-arith-expr
  (has-type-trusted
   (-> (datum) arith-expr)
   (lambda (exp)
    (cond
     ((number? exp) (lit-exp exp))
     ((and (list? exp)
	   (= (length exp) 2)
	   (eq? (car exp) '-))
      (unary-minus-exp
       (parse-arith-expr (cadr exp))))
     ((and (list? exp)
	   (= (length exp) 3))
      (binary-op-exp (parse-arith-expr (car exp))
		     (parse-infix-op (cadr exp))
		     (parse-arith-expr (caddr exp))))
     (else (error "parse-arith-expr: bad syntax: " exp))))))

(deftype parse-infix-op (-> (symbol) infix-op))
(define parse-infix-op
  (lambda (sym)
    (case sym
      ((+) (plus-op))
      ((-) (minus-op))
      ((*) (times-op))
      ((**) (exponent-op))
      (else (error "parse-infix-op: bad syntax: " sym)))))

(deftype unparse-arith-expr (-> (arith-expr) datum))
(define unparse-arith-expr
  (lambda (exp)
    (cases arith-expr exp
     (lit-exp (datum) (has-type datum datum))
     (binary-op-exp (left-arg op right-arg)
       (list (unparse-arith-expr left-arg)
	     (unparse-infix-op op)
	     (unparse-arith-expr right-arg)))
     (unary-minus-exp (arg)
       (list (has-type datum '-) (unparse-arith-expr arg)))
     )))


(deftype unparse-infix-op (-> (infix-op) datum))
(define unparse-infix-op
  (lambda (io)
    (cases infix-op io
      (plus-op () '+)
      (minus-op () '-)
      (times-op () '*)
      (exponent-op () '**)
      )))
