;;; $Id: atomic-exp.scm,v 1.9 2005/03/22 18:08:51 leavens Exp $
;;; "atomic-exps", a form of atomic-expressions, using the grammar
;;;    <(atomic-exp-of T)> ::= <T> | ( {<(atomic-exp-of T)>}* )
;;; where the T's are required to be atomic and not null.

(module lambda-1-exp (lib "typedscm.ss" "lib342")

(provide atomic-exp-of atomic-list-of
	atom->atomic-exp atomic-list->atomic-exp
	atomic-exp-atomic? atomic-exp-atomic-list? atomic-exp->atom
	atomic-exp->atomic-list parse-atomic-exp parse-atomic-list
        atomic?)

(deftype atomic-exp-of
  (forall (T) (-> ((type-predicate-for T))
		  (type-predicate-for (atomic-exp-of T)))))
(deftype atomic-list-of
  (forall (T) (-> ((type-predicate-for T))
		  (type-predicate-for (list-of (atomic-exp-of T))))))
(deftype atom->atomic-exp
  (forall (T) (-> (T) (atomic-exp-of T))))
(deftype atomic-list->atomic-exp
  (forall (T) (-> ((list-of (atomic-exp-of T))) (atomic-exp-of T))))
(deftype atomic-exp-atomic?
  (forall (T) (-> ((atomic-exp-of T)) boolean)))
(deftype atomic-exp-atomic-list?
  (forall (T) (-> ((atomic-exp-of T)) boolean)))
(deftype atomic-exp->atom
  (forall (T) (-> ((atomic-exp-of T)) T)))
(deftype atomic-exp->atomic-list
  (forall (T) (-> ((atomic-exp-of T)) (list-of (atomic-exp-of T)))))
(deftype parse-atomic-exp
  (forall (T) (-> ((type-predicate-for T) datum) (atomic-exp-of T))))
(deftype parse-atomic-list
  (forall (T) (-> ((type-predicate-for T) datum)
		  (list-of (atomic-exp-of T)))))

(require (lib "all-mod.scm" "lib342") (lib "atomic.scm" "lib342"))

(defrep (forall (T) (atomic-exp-of T) datum))

(define atomic-exp-of
  (forall (T?)
    (lambda (d)
       (or (atomic? d)
	   ((atomic-list-of T?) d)))))

(define atomic-list-of
  (forall (T?)
    (list-of (atomic-exp-of T?))))

(define atom->atomic-exp
  (lambda (t)
    (if (and (atomic? t) (not (null? t)))
	t
	(error
	 "atom->atomic-exp: argument must be atomic and not null, but is: "
	 t))))

(define atomic-list->atomic-exp
  (lambda (tr)
    tr))

(define atomic-exp-atomic?
  (lambda (exp)
    (and (atomic? exp) (not (null? exp)))))

(define atomic-exp-atomic-list?
  (lambda (exp)
    (or (null? exp) (pair? exp))))

(define atomic-exp->atom
  (lambda (exp)
    (if (atomic-exp-atomic? exp)
	exp
	(error "atomic-exp->atom: this is not an atom: " exp))))

(define atomic-exp->atomic-list
  (lambda (exp)
    (if (atomic-exp-atomic-list? exp)
	exp
	(error "atomic-exp->atom: this is not an atomic-list: " exp))))

(define parse-atomic-exp
  (lambda (T? exp)
    (has-type-trusted
      (forall (T) (atomic-exp-of T))
      (cond
       ((list? exp) (atomic-list->atomic-exp (parse-atomic-list T? exp)))
       ((atomic? exp)
	(if (T? exp)
	    (atom->atomic-exp exp)
	    (error "parse-atomic-exp: wrong type: " exp)))
       (else (error "parse-atomic-exp: bad syntax: " exp))))))

(define parse-atomic-list
  (lambda (T? exp)
    (has-type-trusted
      (forall (T) (atomic-exp-of T))
      (cond
       ((list? exp) (map (lambda (ae) (parse-atomic-exp T? ae)) exp))
       (else (error "parse-atomic-list: bad syntax: " exp))))))

) ;; end module
