;;; $Id: atomic-expression.scm,v 1.8 2005/02/18 02:39:10 leavens Exp $
;;; Atomic-Expressions, using the grammar:
;;;    <(atomic-expression-of T)> ::= <T> | <(tree-of  T)>
;;; where the T's are required to be atomic and not null.
;;; See the file atomic-tree.scm for the grammar and operations on trees.
;;; See the file atomic-exp.scm for a type doesn't use a tree ADT.

(module atomic-expression (lib "typedscm.ss" "lib342")

(provide atomic-expression-of atom->atomic-expression tree->atomic-expression
	 atomic-expression-atom? atomic-expression-tree?
	 atomic-expression->atom atomic-expression->tree
	 parse-atomic-expression atomic?)
  
(deftype atomic-expression-of
  (forall (T) (-> ((type-predicate-for T))
		  (type-predicate-for (atomic-expression-of T)))))
(deftype atom->atomic-expression
  (forall (T) (-> (T) (atomic-expression-of T))))
(deftype tree->atomic-expression
  (forall (T) (-> ((tree-of  T)) (atomic-expression-of T))))
(deftype atomic-expression-atom?
  (forall (T) (-> ((atomic-expression-of T)) boolean)))
(deftype atomic-expression-tree?
  (forall (T) (-> ((atomic-expression-of T)) boolean)))
(deftype atomic-expression->atom
  (forall (T) (-> ((atomic-expression-of T)) T)))
(deftype atomic-expression->tree
  (forall (T) (-> ((atomic-expression-of T)) (tree-of T))))
(deftype parse-atomic-expression
  (forall (T S)
    (all-of (-> ((list-of T)) (atomic-expression-of T))
            (-> (S) (atomic-expression-of S))
            (-> (datum) (atomic-expression-of datum)))))

(require (lib "atomic.scm" "lib342"))

(defrep (forall (T) (atomic-expression-of T) datum))  ; run-time checked

(define atomic-expression-of
  (forall (T?)
    (lambda (d)
      (or (atomic? d)
	  ((list-of (atomic-expression-of T?)) d)))))
 
(define atom->atomic-expression
  (lambda (t)
    (if (and (atomic? t) (not (null? t)))
	t
	(error
	 "atom->atomic-expression: argument must be atomic and not null, but is: "
	 t))))

(define tree->atomic-expression
  (lambda (tr)
    (if (and (list? tr) (atomic-expression-tree? tr))
	tr
	(error "list->atomic-expression: improper tree argument: " tr))))

(define atomic-expression-atom?
    (lambda (exp)
      (and (atomic? exp) (not (null? exp)))))
      
(define atomic-expression-tree?
  (lambda (exp)
    (or (null? exp) (pair? exp))))

(define atomic-expression->atom
  (lambda (exp)
    (if (atomic-expression-atom? exp)
	exp
	(error "atomic-expression->atom: atomic-expression is a tree " exp))))

(define atomic-expression->tree
  (lambda (exp)
    (if (atomic-expression-tree? exp)
	exp
	(error "atomic-expression->atom: atomic-expression is not a tree "
	       exp))))

(define parse-atomic-expression
  (lambda (exp)
    (has-type
      atomic-expression
      (cond
       ((list? exp) (map parse-atomic-expression exp))
       ((atomic? exp) (atom->atomic-expression exp))
       (else (error "parse-atomic-expression: bad syntax: " exp))))))

) ;; end module
