;;; $Id: combinator-lambda-1-exp.scm,v 1.1 2002/04/09 22:25:24 leavens Exp $

;;; An ADT for the untyped lambda 1-argument calculus with
;;; the three constants LS, LK, and LI, representing the combinators
;;; S, K, and I.

;;; Grammar:
;;;  <combinator-lambda-1-exp>
;;;                 ::= <identifier>
;;;                  | (lambda (<identifier>) <combinator-lambda-1-exp>)
;;;                  | (<combinator-lambda-1-exp> <combinator-lambda-1-exp>)
;;;                  | LS | LK | LI

(define-datatype combinator-lambda-1-exp combinator-lambda-1-exp?
  (var-exp (id symbol?))
  (lambda-exp (formal symbol?) (body combinator-lambda-1-exp?))
  (app-exp (rator combinator-lambda-1-exp?) (rand combinator-lambda-1-exp?))
  (LS)
  (LK)
  (LI))

(deftype parse-combinator-lambda-1-exp (-> (datum) combinator-lambda-1-exp))
(define parse-combinator-lambda-1-exp
  (has-type-trusted
   (-> (datum) combinator-lambda-1-exp)
   (lambda (exp)
     (cond
      ((symbol? exp)
       (case exp
	 ((S) (LS))
	 ((K) (LK))
	 ((I) (LI))
	 (else (var-exp exp))))
      ((and (list? exp) (= (length exp) 3) (eq? (car exp) 'lambda))
       (lambda-exp (caadr exp)
		   (parse-combinator-lambda-1-exp (caddr exp))))
      ((and (list? exp) (= (length exp) 2))
       (app-exp (parse-combinator-lambda-1-exp (car exp))
		(parse-combinator-lambda-1-exp (cadr exp))))
      (else (error "parse-combinator-lambda-1-exp: bad syntax in expression:"
		   exp))))))

(deftype unparse-combinator-lambda-1-exp (-> (combinator-lambda-1-exp) datum))
(define unparse-combinator-lambda-1-exp
  (lambda (exp)
    (cases combinator-lambda-1-exp exp
      (LS () (has-type datum 'S))
      (LK () (has-type datum 'K))
      (LI () (has-type datum 'I))
      (var-exp (id) (has-type datum id))
      (lambda-exp (formal body)
	(list 'lambda (list formal) (unparse-combinator-lambda-1-exp body)))
      (app-exp (rator rand) (list (unparse-combinator-lambda-1-exp rator)
				  (unparse-combinator-lambda-1-exp rand)))
      )))
