;;; $Id: combinator-term.scm,v 1.4 2002/04/09 22:25:24 leavens Exp $

;;; An ADT for combinator terms

;;;  <combinator-term> ::= <identifier> | S | K | I
;;;                     | (<combinator-term> <combinator-term>)

(define-datatype combinator-term combinator-term?
  (var (symbol symbol?))
  (S)
  (K)
  (I)
  (capp (rator combinator-term?) (rand combinator-term?)))

(define S?
  (lambda (trm)
    (cases combinator-term trm
      (S () #t)
      (else #f))))
(define K?
  (lambda (trm)
    (cases combinator-term trm
      (K () #t)
      (else #f))))
(define I?
  (lambda (trm)
    (cases combinator-term trm
      (I () #t)
      (else #f))))

(define capp?
  (lambda (trm)
    (cases combinator-term trm
      (capp (rator rand) #t)
      (else #f))))

(define capp->rator
  (lambda (trm)
    (cases combinator-term trm
      (capp (rator rand) rator)
      (else (error "capp->rand called on non-capp term: " trm)))))

(define capp->rand
  (lambda (trm)
    (cases combinator-term trm
      (capp (rator rand) rand)
      (else (error "capp->rand called on non-capp term: " trm)))))

(deftype parse-combinator-term (-> (datum) combinator-term))
(define parse-combinator-term
  (has-type-trusted
   (-> (datum) combinator-term)
   (lambda (exp)
     (cond
      ((symbol? exp)
       (case exp
	 ((S) (S))
	 ((K) (K))
	 ((I) (I))
	 (else (var exp))))
      ((and (list? exp) (= (length exp) 2))
       (capp (parse-combinator-term (car exp))
	     (parse-combinator-term (cadr exp))))
      (else (error "parse-combinator-term: bad syntax in expression:"
		   exp))))))

(deftype unparse-combinator-term (-> (combinator-term) datum))
(define unparse-combinator-term
  (lambda (exp)
    (cases combinator-term exp
      (S () (has-type datum 'S))
      (K () (has-type datum 'K))
      (I () (has-type datum 'I))
      (var (symbol) (has-type datum symbol))
      (capp (rator rand) (list (unparse-combinator-term rator)
			       (unparse-combinator-term rand)))
      (else (error "unparse-combinator-term: Invalid abstract syntax"
		   exp)))))



