;;; $Id: environment-5-4-2.scm,v 1.1 2005/04/13 22:33:08 leavens Exp $
;;; adapted from 3-5.scm and 3-6-letrec3.scm in the EOPL2e distribution
;;; as modified in section 5.4.2 for the OO interpreters

;;;;;;;;;;;;;;;; environments ;;;;;;;;;;;;;;;;

(module environment-5-4-2 (lib "typedscm.ss" "lib342")

(provide environment? empty-env extend-env extend-env-refs
	 extend-env-recursively-maker
	 apply-env apply-env-ref defined-in-env?)

;; type predicate
(deftype environment? (type-predicate-for environment))

;; constructors
(deftype empty-env (-> () environment))
(deftype extend-env
  (-> ((list-of symbol) (list-of Expressed-Value) environment) environment))
(deftype extend-env-refs
  (-> ((list-of symbol) (vector-of Expressed-Value) environment) environment))
(deftype extend-env-recursively-maker
  (-> ((-> ((list-of symbol) expression environment) Expressed-Value)
       Expressed-Value)
      (-> ((list-of symbol) (list-of (list-of symbol)) (list-of expression)
	   environment)
	  environment)))

(require (lib "reference-3-7.scm" "lib342"))

;; observers
(deftype apply-env
  (-> (environment symbol) Expressed-Value))
(deftype apply-env-ref
  (-> (environment symbol) (ref-of Expressed-Value)))
(deftype defined-in-env?                   ; added
  (-> (environment symbol) boolean))

(define-datatype environment environment?
  (empty-env)
  (extended-env-record
    (syms (list-of symbol?))
    (vec vector?)              ; can use this for anything.
    (env environment?))
  )

(define extend-env
  (lambda (syms vals env)
    (extended-env-record syms (list->vector vals) env)))

(define extend-env-refs
  (lambda (syms vec env)
    (extended-env-record syms vec env)))

(define extend-env-recursively-maker
  (lambda (make-closure init-expressed)
    (lambda (proc-names idss bodies old-env)
      (let ((len (length proc-names)))
	(let ((vec (make-vector len init-expressed)))
	  (let ((env (extended-env-record
		      proc-names vec old-env)))
	    (for-each
	     (lambda (pos ids body)
	       (vector-set! vec pos (make-closure ids body env)))
	     (iota len) idss bodies)
	    env))))))

(define apply-env
  (lambda (env sym)
    (deref (apply-env-ref env sym))))

(define apply-env-ref
  (lambda (env sym)
    (cases environment env
      (empty-env ()
        (eopl:error 'apply-env-ref "No binding for ~s" sym))
      (extended-env-record (syms vals env)
        (let ((pos (rib-find-position sym syms)))
          (if (<= 0 pos)
            (a-ref pos vals)
            (apply-env-ref env sym)))))))

(define defined-in-env?
  (lambda (env sym)
    (cases environment env
      (empty-env () #f)
      (extended-env-record (syms vals env)
        (let ((position (rib-find-position sym syms)))
          (or (<= 0 position)
              (defined-in-env? env sym)))))))

(deftype rib-find-position
  (-> (symbol (list-of symbol)) number))
(define rib-find-position
  (lambda (name symbols)
    (list-find-last-position name symbols)))

(deftype list-find-last-position
  (-> (symbol (list-of symbol)) number))
(define list-find-last-position
  (lambda (sym los)
    (let loop
      ((los los) (curpos 0) (lastpos -1))
      (cond
        ((null? los) lastpos)
        ((eqv? sym (car los))
         (loop (cdr los) (+ curpos 1) curpos))
        (else (loop (cdr los) (+ curpos 1) lastpos))))))

(deftype iota (-> (number) (list-of number)))
(define iota
  (lambda (end)
    (let loop ((next 0))
      (if (>= next end) '()
        (cons next (loop (+ 1 next)))))))

) ;; end module
