;;; $Id: lambda-1-quote-exp-as-ast.scm,v 1.2 2005/02/28 20:05:26 leavens Exp $
;;; An ADT for the untyped 1-argument lambda calculus with quoted symbols
;;; Note that this has a different interface than the ADT in
;;; lambda-1-quote-exp.scm.
;;;
;;; Grammar:
;;;  <expression> ::=
;;;       <identifier>                               "var-exp (id)"
;;;     | (quote <identifier>)                       "quote-exp (symbol)"
;;;     | ( lambda ( <identifier> ) <expression> )   "lambda-exp (id body)"
;;;     | ( <expression> <expression> )              "app-exp (rator rand)"

(module lambda-1-quote-exp-as-ast (lib "typedscm.ss" "lib342")

(provide expression expression?
	 var-exp quote-exp lambda-exp app-exp
	 parse-expression unparse-expression)

;; type predicate
(deftype expression? (type-predicate-for expression))

;; the type itself, to allow use of cases
(deftype expression
  (variant-record
   (var-exp (id symbol))
   (quote-exp (symbol symbol))
   (lambda-exp (id symbol) (body expression))
   (app-exp (rator expression) (rand expression))))

;; constructors
(deftype var-exp (-> (symbol) expression))
(deftype quote-exp (-> (symbol) expression))
(deftype lambda-exp (-> (symbol expression) expression))
(deftype app-exp (-> (expression expression) expression))

;; parsing/bless input correctness
(deftype parse-expression (-> (datum) expression))
;; unparsing
(deftype unparse-expression (-> (expression) datum))
  
;; EOPL2e, p. 48
(define-datatype expression expression? 
  (var-exp
    (id symbol?))
  (quote-exp
    (symbol symbol?))
  (lambda-exp
    (id symbol?)
    (body expression?))
  (app-exp
    (rator expression?)
    (rand expression?)))

;; adapted from EOPL2e, p. 51
(define parse-expression
  (has-type-trusted
    (-> (datum) expression)
    (lambda (datum)
      (cond
       ;; ENSURES: if exp has the syntax of a expression,
       ;;          then result is the representation of exp as an expression,
       ;;          otherwise an error message is given.
       ((symbol? datum) (var-exp datum))
       ((and (list? datum) (= 2 (length datum))
	     (eqv? (car datum) 'quote)
	     (symbol? (cadr datum)))
	(quote-exp (cadr datum)))
       ((and (list? datum) (= 3 (length datum))
	     (eqv? (car datum) 'lambda)
	     (list? (cadr datum))
	     (= 1 (length (cadr datum))))
	(lambda-exp (caadr datum)
		    (parse-expression (caddr datum))))
       ((and (list? datum) (= 2 (length datum)))
	(app-exp
	 (parse-expression (car datum))
	 (parse-expression (cadr datum))))
       (else (eopl:error 'parse-expression
			 "Invalid concrete syntax ~s" datum))))))

(define unparse-expression
  (lambda (exp)
    (cases expression exp
      (var-exp (id) id)
      (quote-exp (symbol) symbol)
      (lambda-exp (id body) 
        (list 'lambda (list id)
          (unparse-expression body)))
      (app-exp (rator rand)
        (list (unparse-expression rator)
              (unparse-expression rand))))))

) ;; end module
