;;; $Id: lexical-addr.scm,v 1.8 2005/04/04 20:35:07 leavens Exp $
;;; An ADT for lexical addresses.  The grammar is as follows.
;;;
;;;  <lexical-addr> ::=
;;;     ( <symbol> : <number> <number> )   "lexical-addr (name depth position)"

(module lexical-addr (lib "typedscm.ss" "lib342")

(provide lexical-addr? lexical-addr
	 lexical-addr->name lexical-addr->depth  lexical-addr->position
	 parse-lexical-addr)

;; type predicate
(deftype lexical-addr? (type-predicate-for lexical-addr))

;; constructor
(deftype lexical-addr (-> (symbol number number) lexical-addr))

;; observers
(deftype lexical-addr->name (-> (lexical-addr) symbol))
(deftype lexical-addr->depth (-> (lexical-addr) number))
(deftype lexical-addr->position (-> (lexical-addr) number))

;; parse/bless input correctness
(deftype parse-lexical-addr (-> (datum) lexical-addr))

(defrep (lexical-addr (list-of datum)))

(define lexical-addr?
  (lambda (d)
    (has-type-trusted
      boolean
      (and (list? d)
	   (= (length d) 4)
	   (symbol? (car d))
	   (eq? ': (cadr d))
	   (number? (caddr d))
	   (number? (cadddr d))))))

(define lexical-addr
  (lambda (v d p)
    ;; ENSURES: result is a lexical-addr representing (v : d p)
    (list v ': d p)))

(define lexical-addr->name
  (lambda (la)
    ;; ENSURES: result is the name represented in la
    (test-type symbol? (car la))))

(define lexical-addr->depth
  (lambda (la)
    ;; ENSURES: result is the lexical depth represented in la
    (test-type number? (caddr la))))

(define lexical-addr->position
  (lambda (la)
    ;; ENSURES: result is the lexical position represented in la
    (test-type number? (cadddr la))))

(define parse-lexical-addr
  (lambda (addr)
    (has-type-trusted
      (list-of datum)
      (if (lexical-addr? addr)
	  addr
	  (error "parse-lexical-addr, syntax error in " addr)))))

) ; end module
