;;; $Id: phone-book-as-ribcage.scm,v 1.7 2005/04/16 21:13:15 dorn Exp $
;;;---------------------------------------------------------------------
;;; An ADT for a simple phone book
;;;---------------------------------------------------------------------

;;;---------------------------------------------------------------------
;;; The declarations for the public (client-visible) interface of the
;;; phone book ADT
;;;---------------------------------------------------------------------

(module phone-book-as-ribcage (lib "typedscm.ss" "lib342")

(provide create-empty-phone-book add-to-phone-book look-up
         add-several-to-phone-book)

(deftype create-empty-phone-book (-> () phone-book))
(deftype add-to-phone-book (-> (symbol number phone-book) phone-book))
(deftype look-up (-> (phone-book symbol) number))


;;;---------------------------------------------------------------------
;;; A ribcage representation of the phone book ADT
;;;---------------------------------------------------------------------

(require (lib "ribassoc.scm" "lib342"))

(defrep (phone-book (list-of (pair-of (list-of symbol)
				     (vector-of number)))))
(define create-empty-phone-book
  (lambda () '()))

(define add-to-phone-book
  (lambda (name number old-book)
    (cons (make-rib (list name) (list number))
	  old-book)))

(deftype make-rib (forall (T) 
                          (-> ((list-of symbol) (list-of T))
                              (pair-of (list-of symbol) (vector-of T)))))
(define make-rib
  (lambda (ls lt)
    (cons ls (list->vector lt))))

(define look-up
  (has-type-trusted			; necessary because typechecker can't
   (-> ((list-of (pair-of (list-of symbol)	; figure out ribassoc
			  (vector-of number)))
	symbol)
       number)
   (lambda (book who)
     (if (null? book)
	 (error "no listing for" who)
	 (let ((first-listings (car book))
	       (old-book (cdr book)))
	   (let ((names (car first-listings))
		 (numbers (cdr first-listings)))
	     (let ((result (ribassoc who names numbers '*fail*)))
	       (if (eq? result '*fail*)
		   (look-up old-book who)
		   result))))))))

;;;---------------------------------------------------------------------
;;; With the ribcage rep it is easy to add a procedure for multiple
;;; names and numbers to the ADT
;;;---------------------------------------------------------------------

(deftype add-several-to-phone-book
  (-> ((list-of symbol) (list-of number) phone-book) phone-book))
(define add-several-to-phone-book
  (lambda (names numbers old-book)
    (cons (make-rib names numbers)
	  old-book)))

) ;; end module	      
