;;; $Id: sxml-node.tst,v 1.4 2005/02/18 02:39:11 leavens Exp $
;;; AUTHOR: Gary T. Leavens

(require (lib "sxml-node.scm" "lib342"))

(run-regression-tests-equal?
 `(

   ( (node? (value "a value")) ==> #t )
   ( (node? (named (top) (list))) ==> #t )
   ( (node? (named (top) (list (value "another")))) ==> #t )
   ( (node? (named (local-name 'foo)
		   (list (value "another") (value "another one also"))))
	    ==> #t )

   ( (node? 'x) ==> #f )
   ( (node? 32.4) ==> #f )
   ( (node? '(3 4 2)) ==> #f )

   ( (value? (value "a value")) ==> #t )
   ( (value? (named (top) (list (value "another")))) ==> #f )
   ( (value? (named (local-name 'foo)
		   (list (value "another") (value "another one also"))))
	    ==> #f )

   ( (named? (value "a value")) ==> #f )
   ( (named? (named (top) (list (value "another")))) ==> #t )
   ( (named? (named (local-name 'foo)
		   (list (value "another") (value "another one also"))))
	    ==> #t )

   ( (attribute? 'my-name) ==> #f )
   ( (attribute? 'e) ==> #f )
   ( (attribute? (attribute)) ==> #t )
   ( (attribute? (top)) ==> #f )
   ( (attribute? (process-instr)) ==> #f )
   ( (attribute? (comment)) ==> #f )
   ( (attribute? (entity)) ==> #f )
   ( (attribute? (namespaces)) ==> #f )
   ( (attribute? 'someurl:somename) ==> #f )

   ( (top? 'my-name) ==> #f )
   ( (top? 'e) ==> #f )
   ( (top? (attribute)) ==> #f )
   ( (top? (top)) ==> #t )
   ( (top? (process-instr)) ==> #f )
   ( (top? (comment)) ==> #f )
   ( (top? (entity)) ==> #f )
   ( (top? (namespaces)) ==> #f )
   ( (top? 'someurl:somename) ==> #f )

   ( (process-instr? 'my-name) ==> #f )
   ( (process-instr? 'e) ==> #f )
   ( (process-instr? (attribute)) ==> #f )
   ( (process-instr? (top)) ==> #f )
   ( (process-instr? (process-instr)) ==> #t )
   ( (process-instr? (comment)) ==> #f )
   ( (process-instr? (entity)) ==> #f )
   ( (process-instr? (namespaces)) ==> #f )
   ( (process-instr? 'someurl:somename) ==> #f )

   ( (comment? 'my-name) ==> #f )
   ( (comment? 'e) ==> #f )
   ( (comment? (attribute)) ==> #f )
   ( (comment? (top)) ==> #f )
   ( (comment? (process-instr)) ==> #f )
   ( (comment? (comment)) ==> #t )
   ( (comment? (entity)) ==> #f )
   ( (comment? (namespaces)) ==> #f )
   ( (comment? 'someurl:somename) ==> #f )

   ( (entity? 'my-name) ==> #f )
   ( (entity? 'e) ==> #f )
   ( (entity? (attribute)) ==> #f )
   ( (entity? (top)) ==> #f )
   ( (entity? (process-instr)) ==> #f )
   ( (entity? (comment)) ==> #f )
   ( (entity? (entity)) ==> #t )
   ( (entity? (namespaces)) ==> #f )
   ( (entity? 'someurl:somename) ==> #f )

   ( (namespaces? 'my-name) ==> #f )
   ( (namespaces? 'e) ==> #f )
   ( (namespaces? (attribute)) ==> #f )
   ( (namespaces? (top)) ==> #f )
   ( (namespaces? (process-instr)) ==> #f )
   ( (namespaces? (comment)) ==> #f )
   ( (namespaces? (entity)) ==> #f )
   ( (namespaces? (namespaces)) ==> #t )
   ( (namespaces? 'someurl:somename) ==> #f )

   ( (local-name? 'my-name) ==> #t )
   ( (local-name? 'e) ==> #t )
   ( (local-name? (attribute)) ==> #f )
   ( (local-name? (top)) ==> #f )
   ( (local-name? (process-instr)) ==> #f )
   ( (local-name? (comment)) ==> #f )
   ( (local-name? (entity)) ==> #f )
   ( (local-name? (namespaces)) ==> #f )
   ( (local-name? 'someurl:somename) ==> #f )

   ( (expanded-name? 'my-name) ==> #f )
   ( (expanded-name? 'e) ==> #f )
   ( (expanded-name? (attribute)) ==> #f )
   ( (expanded-name? (top)) ==> #f )
   ( (expanded-name? (process-instr)) ==> #f )
   ( (expanded-name? (comment)) ==> #f )
   ( (expanded-name? (entity)) ==> #f )
   ( (expanded-name? (namespaces)) ==> #f )
   ( (expanded-name? 'someurl:somename) ==> #t )
   ( (expanded-name? 'cars:part) ==> #t )

   ( (administrative? 'my-name) ==> #f )
   ( (administrative? 'e) ==> #f )
   ( (administrative? (attribute)) ==> #t )
   ( (administrative? (top)) ==> #t )
   ( (administrative? (process-instr)) ==> #t )
   ( (administrative? (comment)) ==> #t )
   ( (administrative? (entity)) ==> #t )
   ( (administrative? (namespaces)) ==> #t )
   ( (administrative? 'someurl:somename) ==> #f )
   ( (administrative? 'cars:part) ==> #f )

   ( (named (local-name 'second-level)
	    (list (value "a value")))
     ==> (second-level "a value") )
   
   ( (named (local-name 'some-name)
	    (list (named (local-name 'second-level)
			 (list (value "a value")))))
     ==> (some-name (second-level "a value")) )

   ( (named->name
      (named (local-name 'some-name)
	     (list (named (local-name 'second-level)
			  (list (value "a value"))))))
     ==> some-name )

   ( (named->nodelist
      (named (local-name 'some-name)
	     (list (named (local-name 'second-level)
			  (list (value "a value"))))))
     ==> ((second-level "a value")) )

   ( (named->name
      (car (named->nodelist
	    (named (local-name 'some-name)
		   (list (named (local-name 'second-level)
				(list (value "a value"))))))))
     ==> second-level )

   ( (named->nodelist
      (car (named->nodelist
	    (named (local-name 'some-name)
		   (list (named (local-name 'second-level)
				(list (value "a value"))))))))
     ==> ("a value") )

   ( (value->string
      (car (named->nodelist
	    (car (named->nodelist
		  (named (local-name 'some-name)
			 (list (named (local-name 'second-level)
				      (list (value "a value"))))))))))
     ==> "a value" )

   ( (local-name->symbol  (local-name 'some-name))
      ==> some-name )

   ( (expanded-name->symbol  (expanded-name 'urn:loc.gov:books:notes))
      ==> urn:loc.gov:books:notes )

   ( (node?
      '(*TOP*
       (@ (*NAMESPACES* 
            (HTML "http://www.w3.org/TR/REC-html40")))
       (RESERVATION
         (NAME (@ (HTML:CLASS "largeSansSerif"))
           "Layman, A")
         (SEAT (@ (HTML:CLASS "largeMonotype")
                  (CLASS "Y"))
            "33B")
         (HTML:A (@ (HREF "/cgi-bin/ResStatus"))
            "Check Status")
         (DEPARTURE "1997-05-24T07:55:00+1"))))
     ==> #t )

    )
 )
