;;; $Id: tc-env.tst,v 1.2 2005/04/01 05:26:06 leavens Exp $


(require (lib "tc-env.scm" "lib342")
	 (lib "tc-hash-table.scm" "lib342"))

(define emptyenv (tc:env-empty))
(define a3 (tc:env-update 'a 3 emptyenv))
(define b4c5a3 (tc:env-update* '(b c) '(4 5) a3))
(define dvar7-evar8-b4c5a3
  (tc:env-update* '(dvar evar) '(7 8) b4c5a3))

(run-regression-tests-equal?
 `(
    ( (not (not (tc:env-bound? 'a emptyenv)))
      ==> #f )

    ( (not (not (tc:env-bound? 'a a3)))
      ==> #t )

    ( (not (not (tc:env-bound? 'b a3)))
      ==> #f )

    ( (not (not (tc:env-bound? 'a b4c5a3)))
      ==> #t )

    ( (not (not (tc:env-bound? 'b b4c5a3)))
      ==> #t )

    ( (not (not (tc:env-bound? 'c b4c5a3)))
      ==> #t )

    ( (not (not (tc:env-bound? 'a dvar7-evar8-b4c5a3)))
      ==> #t )

    ( (not (not (tc:env-bound? 'b dvar7-evar8-b4c5a3)))
      ==> #t )

    ( (not (not (tc:env-bound? 'c dvar7-evar8-b4c5a3)))
      ==> #t )

    ( (not (not (tc:env-bound? 'dvar dvar7-evar8-b4c5a3)))
      ==> #t )

    ( (not (not (tc:env-bound? 'evar dvar7-evar8-b4c5a3)))
      ==> #t )

    ( (not (not (tc:env-bound? 'zero? dvar7-evar8-b4c5a3)))
      ==> #f )

    ( (tc:env-value 'a a3)
      ==> 3 )

    ( (tc:env-value 'a b4c5a3)
      ==> 3 )

    ( (tc:env-value 'b b4c5a3)
      ==> 4 )

    ( (tc:env-value 'c b4c5a3)
      ==> 5 )

    ( (not (not (tc:env-bound? 'a emptyenv)))
      ==> #f )

    ( (not (not (tc:env-bound? 'a (tc:env-empty))))
      ==> #f )

    )
 )

(define newenv (tc:env-join
		(tc:env-update* '(b d) '(99 86) emptyenv)
		b4c5a3))


(run-regression-tests-equal?
 `(
    ( (not (not (tc:env-bound? 'd newenv)))
      ==> #t )

    ( (not (not (tc:env-bound? 'a newenv)))
      ==> #t )

    ( (not (not (tc:env-bound? 'b newenv)))
      ==> #t )

    ( (not (not (tc:env-bound? 'c newenv)))
      ==> #t )

    ( (not (not (tc:env-bound? 'e newenv)))
      ==> #f )

    ( (tc:env-value 'b newenv)
      ==> 99 )

    ( (tc:env-value 'd newenv)
      ==> 86 )

    ( (tc:env-value 'a newenv)
      ==> 3 )

    ( (tc:env-value 'c newenv)
      ==> 5 )

    ( (tc:env-value 'b b4c5a3)
      ==> 4 )
    
    ( (tc:env-value 'b (tc:env-map
			(lambda (key val)
			  (if (equal? key 'b) 342 val))
			b4c5a3))
      ==> 342 )

    ( (tc:env-value 'c (tc:env-map
			(lambda (key val)
			  (if (equal? key 'b) 342 val))
			    b4c5a3))
      ==> 5 )

    ( (tc:env-value 'c (tc:env-filter
			(lambda (key val)
			  (equal? key 'c))
			b4c5a3))
      ==> 5 )

    ( (not (not (tc:env-bound? 'b
			   (tc:hash-table-filter
			    (lambda (key val)
			      (equal? key 'c))
			    b4c5a3))))
      ==> #f )

    )
 )


(tc:env-for-each
 (lambda (key val)
   (displayln "key = " key ", val = " val))
 b4c5a3)

;(pretty-print emptyenv)
;(pretty-print a3)
;(pretty-print b4c5a3)
;(pretty-print newenv)

(define symtab (tc:assoc-list->env
		'(
		  (zero? . zero-val)
		  (null? . null-val)
		  (car . car-val)
		  (cdr . cdr-val)
		  (equal? . equal-val)
		  )))

(set! symtab (tc:env-join
	      symtab
	      (tc:env-update*
	         '(caar cadr caddr)
		 '(caar-val cadr-val caddr-val)
		 (tc:env-empty))))

;; simulate externalizing this hash table
(set! symtab (tc:env-map
	      (lambda (key val) val)
	      symtab))

(run-regression-tests-equal?
 `(
    ( (tc:hash-table? symtab)
      ==> #t )

    ( (not (not (tc:env-bound? 'zero? symtab)))
      ==> #t )

    ( (not (not (tc:env-bound? 'null? symtab)))
      ==> #t )

    ( (not (not (tc:env-bound? 'car symtab)))
      ==> #t )

    ( (not (not (tc:env-bound? 'cdr symtab)))
      ==> #t )

    ( (not (not (tc:env-bound? 'equal? symtab)))
      ==> #t )

    ( (tc:env-value 'zero? symtab)
      ==> zero-val )

    ( (tc:env-value 'null? symtab)
      ==> null-val )

    ( (tc:env-value 'caar symtab)
      ==> caar-val )
   )
 )
