;;; $Id: tc-environments.scm,v 1.20 2005/03/28 22:03:06 leavens Exp $

;;; This module extends the environment model in tc-env.scm
;;; as needed for the current type checker.
;;; Author: Brian Dorn and Gary T. Leavens

(module tc-environments (lib "typedscm.ss" "lib342")
  
(provide tc:environment-of tc:environment?
         tc:extend-env tc:extend-mult-env tc:env-same-names
         tc:env-names tc:env-vals
         ;; all-from tc-env
         tc:env-of
         tc:env-empty
         tc:env-update tc:env-update* tc:env-join tc:env-join*
         tc:env-bound? tc:env-value
         tc:env-map tc:env-filter tc:env-for-each
         tc:assoc-list->env)

(deftype tc:environment-of
  (forall (T) (-> ((type-predicate-for T))
                  (type-predicate-for (tc:environment-of T)))))
(deftype tc:environment? (type-predicate-for tc:environment))
(deftype tc:extend-env
  (all-of (forall (T) (-> ((tc:environment-of T) symbol T)
                          (tc:environment-of T)))
          (-> (tc:environment symbol datum) tc:environment)))
(deftype tc:extend-mult-env
  (all-of (forall (T) (-> ((tc:environment-of T) (list-of symbol) (list-of T))
                          (tc:environment-of T)))
          (-> (tc:environment (list-of symbol) (list-of datum)) tc:environment)))
(deftype tc:same-names
  (-> ((list-of symbol) (list-of symbol)) boolean))
(deftype tc:env-names
  (all-of (forall (T) (-> ((tc:environment-of T))
                          (list-of symbol)))
          (-> (tc:environment) (list-of symbol))))
(deftype tc:env-vals
  (all-of (forall (T) (-> ((tc:environment-of T))
                          (list-of T)))
          (-> (tc:environment) (list-of datum))))

(require (lib "tc-env.scm" "lib342"))

(defrep (forall (T) (tc:environment-of T) (tc:env-of symbol T))
        (tc:environment (tc:environment-of datum)))

(define tc:environment-of
  (forall (T?)
    (tc:env-of symbol? T?)))

(define tc:environment?
  (lambda (d)
    ((tc:environment-of datum?) d)))

(define tc:extend-env
  (lambda (init-env name val)
    (tc:env-update name val init-env)))

(define tc:extend-mult-env
  (lambda (init-env names vals)
    ;; ENSURES: returns an environment where init-env is
    ;;          extended with bindings each corresponding
    ;;          pair in names and vals
    (if (null? names)
	init-env
	(tc:extend-mult-env
	 (tc:extend-env init-env (car names) (car vals))
	 (cdr names)
	 (cdr vals)))))

(define tc:env-same-names
  (lambda (env1-names env2-names)
    (cond
     ((null? env1-names) (null? env2-names))
     ((null? env2-names) #f)
     (else (and
	    (eq? (car env1-names) (car env2-names))
	    (tc:env-same-names (cdr env1-names) (cdr env2-names)))))))

(define tc:env-names
  (lambda (env)
    (let ((result '()))
      (begin
	(tc:env-for-each
	 (lambda (key val)
	   (set! result (cons key result)))
	 env)
	result))))

(define tc:env-vals
  (lambda (env)
    (let ((result '()))
      (begin
	(tc:env-for-each
	 (lambda (key val)
	   (set! result (cons val result)))
	 env)
	result))))
      
) ;; end module
