;;; $Id: tc-global-var-output-types.scm,v 1.10 2005/04/13 22:32:14 leavens Exp $
;;;
;;; External types for built-in Scheme procedures
;;; AUTHORS: Gary T. Leavens and Curtis Clifton
;;;          from an earlier verison by Steven L. Jenkins.

(module tc-global-var-output-types (lib "typedscm.ss" "lib342")

(provide *tc:pre-global-var-output-type-env*)

(deftype *tc:pre-global-var-output-type-env*
  (tc:env-of symbol tc:output-type-expr))

(require (lib "tc-env.scm" "lib342"))

(define *tc:pre-global-var-output-type-env*
  ;; This is the table for output-type-exprs of built in expressions.
  ;; The order follows the listing in the body of R5RS, to make it possible
  ;; to check this list against that report.
  ;; Some additions from SCM and for use with EOPL2e are found at the end.
  ;; Types marked "restrictive" are more restrictive than R5RS.
  ;; Types marked "unsound" make the type system unsound.
  (tc:assoc-list->env
   '(
     ;; 6.1
     (eqv?     . (forall (T) (-> (T T) boolean))) ; restrictive
                 ; the following doesn't work because we commit too early
                 ; to a decision on an intersection-type
                 ;(all-of (-> (boolean boolean) boolean)
		 ;        (-> (number number) boolean)
		 ;        (-> (symbol symbol) boolean)
		 ;        (-> (char char) boolean))) ; restrictive
     (eq?      . (forall (T) (-> (T T) boolean))) ; restrictive
     (equal?   . (forall (T) (-> (T T) boolean))) ; restrictive
     ;; 6.2.5
     (number?  . (type-predicate-for number))
     (complex? . (type-predicate-for number))
     (real?    . (type-predicate-for number))
     (rational? . (type-predicate-for number))
     (integer? . (type-predicate-for number))
     (exact?   . (type-predicate-for number))
     (inexact? . (type-predicate-for number))
     (=        . (-> (number ...) boolean))
     (<        . (-> (number ...) boolean))
     (>        . (-> (number ...) boolean))
     (<=       . (-> (number ...) boolean))
     (>=       . (-> (number ...) boolean))
     (zero?    . (-> (number) boolean))
     (negative? . (-> (number) boolean))
     (positive? . (-> (number) boolean))
     (odd?      . (-> (number) boolean))
     (even?     . (-> (number) boolean))
     (max       . (-> (number number  ...) number))     
     (min       . (-> (number number ...) number))
     (+         . (-> (number ...) number))
     (*         . (-> (number ...) number))
     (-         . (-> (number number ...) number))
     (/         . (-> (number number ...) number))
     (abs       . (-> (number) number))
     (quotient  . (-> (number number) number))     
     (remainder . (-> (number number) number))     
     (modulo    . (-> (number number) number))
     (gcd       . (-> (number ...) number))
     (lcm       . (-> (number ...) number))
     (numerator . (-> (number) number))
     (denominator . (-> (number) number))
     (rationalize . (-> (number number) number))
     (floor      . (-> (number) number))
     (ceiling    . (-> (number) number))
     (truncate   . (-> (number) number))     
     (round      . (-> (number) number))
     (exp        . (-> (number) number))
     (log        . (-> (number) number))          
     (sin        . (-> (number) number))     
     (cos        . (-> (number) number))     
     (asin       . (-> (number) number))     
     (acos       . (-> (number) number))
     (tan        . (-> (number) number))     
     (atan       . (all-of (-> (number) number)
			   (-> (number number) number)))
     (expt       . (-> (number number) number))
     (sqrt       . (-> (number) number))
     (make-rectangular . (-> (number number) number))
     (make-polar . (-> (number number) number))
     (real-part  . (-> (number) number))
     (imag-part  . (-> (number) number))
     (magnitude  . (-> (number) number))
     (angle      . (-> (number) number))
     (exact->inexact . (-> (number) number))
     (inexact->exact . (-> (number) number))
     (number->string . (all-of (-> (number) string)
			       (-> (number number) string)))
     (string->number . (all-of (-> (string) number)
			       (-> (string number) number)))
     ;; 6.3.1
     (not      . (-> (boolean) boolean))	; restrictive
     (boolean? . (type-predicate-for boolean))
     ;; 6.3.2
     (pair?    . (forall (S T) (type-predicate-for (pair-of S T))))
     (cons     . (forall (T U V S W)
			 (all-of (-> (T (list-of T)) (list-of T))
				 (-> (U (list-of V)) (list-of datum))
				 (-> (S W) (pair-of S W)))))
     (car      . (forall (T U V) (all-of (-> ((list-of T)) T)
					 (-> ((pair-of U V)) U))))
     (cdr      . (forall (T U V) (all-of (-> ((list-of T)) (list-of T))
					 (-> ((pair-of U V)) V))))
     (set-car! . (forall (T U V) (all-of (-> ((list-of T) T) void)
					 (-> ((pair-of U V) U) void))))
     (set-cdr! . (forall (T U V) (all-of (-> ((list-of T) (list-of T)) void)
					 (-> ((pair-of U V) V) void))))

     ;; The following types for c...r were computed
     ;; using type-check-compute-cxr-types.scm 
     ;; and an earlier version of the type checker.
     (caar .
           (forall
             (s t u v w x y z)
             (all-of
               (-> ((list-of (list-of s))) s)
               (-> ((pair-of (list-of u) t)) u)
               (-> ((list-of (pair-of w v))) w)
               (-> ((pair-of (pair-of z x) y)) z))))
     (cadr .
           (forall
             (s t u v w x)
             (all-of
               (-> ((list-of s)) s)
               (-> ((pair-of t (list-of u))) u)
               (-> ((pair-of v (pair-of x w))) x))))
     (cdar .
           (forall
             (s t u v w x y z)
             (all-of
               (-> ((list-of (list-of s))) (list-of s))
               (-> ((pair-of (list-of u) t)) (list-of u))
               (-> ((list-of (pair-of v w))) w)
               (-> ((pair-of (pair-of x z) y)) z))))
     (cddr .
           (forall
             (s t u v w x)
             (all-of
               (-> ((list-of s)) (list-of s))
               (-> ((pair-of t (list-of u))) (list-of u))
               (-> ((pair-of v (pair-of w x))) x))))
     (caaar
       .
       (forall
         (s t u v w x y z a b c d e f g h i j k l)
         (all-of
           (-> ((list-of (list-of (list-of s)))) s)
           (-> ((pair-of (list-of (list-of u)) t)) u)
           (-> ((list-of (pair-of (list-of w) v))) w)
           (-> ((pair-of (pair-of (list-of z) x) y)) z)
           (-> ((list-of (list-of (pair-of b a)))) b)
           (-> ((pair-of (list-of (pair-of e c)) d)) e)
           (-> ((list-of (pair-of (pair-of h f) g))) h)
           (-> ((pair-of (pair-of (pair-of l i) j) k)) l))))
     (caadr
       .
       (forall
         (s t u v w x y z a b c d e f g)
         (all-of
           (-> ((list-of (list-of s))) s)
           (-> ((pair-of t (list-of (list-of u)))) u)
           (-> ((pair-of v (pair-of (list-of x) w))) x)
           (-> ((list-of (pair-of z y))) z)
           (-> ((pair-of a (list-of (pair-of c b)))) c)
           (-> ((pair-of d (pair-of (pair-of g e) f))) g))))
     (cadar
       .
       (forall
         (s t u v w x y z a b c d e f g)
         (all-of
           (-> ((list-of (list-of s))) s)
           (-> ((pair-of (list-of u) t)) u)
           (-> ((list-of (pair-of v (list-of w)))) w)
           (-> ((pair-of (pair-of x (list-of z)) y)) z)
           (-> ((list-of (pair-of a (pair-of c b)))) c)
           (-> ((pair-of (pair-of d (pair-of g e)) f)) g))))
     (caddr
       .
       (forall
         (s t u v w x y z a b)
         (all-of
           (-> ((list-of s)) s)
           (-> ((pair-of t (list-of u))) u)
           (-> ((pair-of v (pair-of w (list-of x)))) x)
           (-> ((pair-of y (pair-of z (pair-of b a)))) b))))
     (cdaar
       .
       (forall
         (s t u v w x y z a b c d e f g h i j k l)
         (all-of
           (-> ((list-of (list-of (list-of s)))) (list-of s))
           (-> ((pair-of (list-of (list-of u)) t)) (list-of u))
           (-> ((list-of (pair-of (list-of w) v))) (list-of w))
           (-> ((pair-of (pair-of (list-of z) x) y)) (list-of z))
           (-> ((list-of (list-of (pair-of a b)))) b)
           (-> ((pair-of (list-of (pair-of c e)) d)) e)
           (-> ((list-of (pair-of (pair-of f h) g))) h)
           (-> ((pair-of (pair-of (pair-of i l) j) k)) l))))
     (cdadr
       .
       (forall
         (s t u v w x y z a b c d e f g)
         (all-of
           (-> ((list-of (list-of s))) (list-of s))
           (-> ((pair-of t (list-of (list-of u)))) (list-of u))
           (-> ((pair-of v (pair-of (list-of x) w))) (list-of x))
           (-> ((list-of (pair-of y z))) z)
           (-> ((pair-of a (list-of (pair-of b c)))) c)
           (-> ((pair-of d (pair-of (pair-of e g) f))) g))))
     (cddar
       .
       (forall
         (s t u v w x y z a b c d e f g)
         (all-of
           (-> ((list-of (list-of s))) (list-of s))
           (-> ((pair-of (list-of u) t)) (list-of u))
           (-> ((list-of (pair-of v (list-of w)))) (list-of w))
           (-> ((pair-of (pair-of x (list-of z)) y)) (list-of z))
           (-> ((list-of (pair-of a (pair-of b c)))) c)
           (-> ((pair-of (pair-of d (pair-of e g)) f)) g))))
     (cdddr
       .
       (forall
         (s t u v w x y z a b)
         (all-of
           (-> ((list-of s)) (list-of s))
           (-> ((pair-of t (list-of u))) (list-of u))
           (-> ((pair-of v (pair-of w (list-of x)))) (list-of x))
           (-> ((pair-of y (pair-of z (pair-of a b)))) b))))
     (caaaar
       .
       (forall
         (s t
            u
            v
            w
            x
            y
            z
            a
            b
            c
            d
            e
            f
            g
            h
            i
            j
            k
            l
            m
            n
            o
            p
            q
            r
            ?swn#547
            ?twn#547
            ?uwn#547
            ?swn#550
            ?twn#550
            ?uwn#550
            ?swn#553
            ?twn#553
            ?uwn#553
            ?vwn#553
            ?swn#556
            ?twn#556
            ?uwn#556
            ?swn#559
            ?twn#559
            ?uwn#559
            ?vwn#559
            ?swn#562
            ?twn#562
            ?uwn#562
            ?vwn#562
            ?swn#565
            ?twn#565
            ?uwn#565
            ?vwn#565
            ?wwn#565)
         (all-of
           (-> ((list-of (list-of (list-of (list-of s))))) s)
           (-> ((pair-of (list-of (list-of (list-of u))) t)) u)
           (-> ((list-of (pair-of (list-of (list-of w)) v))) w)
           (-> ((pair-of (pair-of (list-of (list-of z)) x) y)) z)
           (-> ((list-of (list-of (pair-of (list-of b) a)))) b)
           (-> ((pair-of (list-of (pair-of (list-of e) c)) d)) e)
           (-> ((list-of (pair-of (pair-of (list-of h) f) g))) h)
           (-> ((pair-of (pair-of (pair-of (list-of l) i) j) k)) l)
           (-> ((list-of (list-of (list-of (pair-of n m))))) n)
           (-> ((pair-of (list-of (list-of (pair-of q o))) p)) q)
           (-> ((list-of
                  (pair-of (list-of (pair-of ?twn#547 r)) ?swn#547)))
               ?twn#547)
           (-> ((pair-of
                  (pair-of (list-of (pair-of ?uwn#550 ?uwn#547)) ?swn#550)
                  ?twn#550))
               ?uwn#550)
           (-> ((list-of
                  (list-of (pair-of (pair-of ?uwn#553 ?swn#553) ?twn#553))))
               ?uwn#553)
           (-> ((pair-of
                  (list-of (pair-of (pair-of ?uwn#556 ?vwn#553) ?swn#556))
                  ?twn#556))
               ?uwn#556)
           (-> ((list-of
                  (pair-of
                    (pair-of (pair-of ?vwn#559 ?swn#559) ?twn#559)
                    ?uwn#559)))
               ?vwn#559)
           (-> ((pair-of
                  (pair-of
                    (pair-of (pair-of ?swn#565 ?swn#562) ?twn#562)
                    ?uwn#562)
                  ?vwn#562))
               ?swn#565))))
     (caaadr
       .
       (forall
         (s t
            u
            v
            w
            x
            y
            z
            a
            b
            c
            d
            e
            f
            g
            h
            i
            j
            k
            l
            m
            n
            o
            p
            q
            r
            ?uwn#609
            ?swn#612
            ?twn#612
            ?uwn#612
            ?vwn#612
            ?swn#615
            ?twn#615
            ?swn#618
            ?twn#618
            ?uwn#618
            ?swn#621
            ?twn#621
            ?uwn#621
            ?vwn#621
            ?swn#624
            ?twn#624
            ?uwn#624
            ?swn#627
            ?twn#627
            ?uwn#627
            ?vwn#627
            ?swn#630
            ?twn#630
            ?uwn#630
            ?vwn#630
            ?wwn#630)
         (all-of
           (-> ((list-of (list-of (list-of s)))) s)
           (-> ((pair-of t (list-of (list-of (list-of u))))) u)
           (-> ((pair-of v (pair-of (list-of (list-of x)) w))) x)
           (-> ((list-of (pair-of (list-of z) y))) z)
           (-> ((pair-of a (list-of (pair-of (list-of c) b)))) c)
           (-> ((pair-of d (pair-of (pair-of (list-of g) e) f))) g)
           (-> ((list-of (list-of (pair-of i h)))) i)
           (-> ((pair-of j (list-of (list-of (pair-of l k))))) l)
           (-> ((pair-of m (pair-of (list-of (pair-of p n)) o))) p)
           (-> ((list-of (pair-of (pair-of ?uwn#609 q) r))) ?uwn#609)
           (-> ((pair-of
                  ?swn#612
                  (list-of (pair-of (pair-of ?vwn#612 ?twn#612) ?uwn#612))))
               ?vwn#612)
           (-> ((pair-of
                  ?swn#615
                  (pair-of
                    (pair-of (pair-of ?uwn#618 ?twn#615) ?swn#618)
                    ?twn#618)))
               ?uwn#618))))
     (caadar
       .
       (forall
         (s t
            u
            v
            w
            x
            y
            z
            a
            b
            c
            d
            e
            f
            g
            h
            i
            j
            k
            l
            m
            n
            o
            p
            q
            r
            ?uwn#1191
            ?swn#1194
            ?twn#1194
            ?uwn#1194
            ?vwn#1194
            ?swn#1197
            ?twn#1197
            ?swn#1200
            ?twn#1200
            ?uwn#1200
            ?swn#1203
            ?twn#1203
            ?uwn#1203
            ?swn#1206
            ?twn#1206
            ?uwn#1206
            ?vwn#1206
            ?swn#1209
            ?twn#1209
            ?uwn#1209
            ?vwn#1209
            ?swn#1212
            ?twn#1212
            ?uwn#1212
            ?vwn#1212
            ?wwn#1212)
         (all-of
           (-> ((list-of (list-of (list-of s)))) s)
           (-> ((pair-of (list-of (list-of u)) t)) u)
           (-> ((list-of (pair-of v (list-of (list-of w))))) w)
           (-> ((pair-of (pair-of x (list-of (list-of z))) y)) z)
           (-> ((list-of (pair-of a (pair-of (list-of c) b)))) c)
           (-> ((pair-of (pair-of d (pair-of (list-of g) e)) f)) g)
           (-> ((list-of (list-of (pair-of i h)))) i)
           (-> ((pair-of (list-of (pair-of l j)) k)) l)
           (-> ((list-of (pair-of m (list-of (pair-of o n))))) o)
           (-> ((pair-of
                  (pair-of p (list-of (pair-of ?uwn#1191 q)))
                  r))
               ?uwn#1191)
           (-> ((list-of
                  (pair-of
                    ?swn#1194
                    (pair-of (pair-of ?vwn#1194 ?twn#1194) ?uwn#1194))))
               ?vwn#1194)
           (-> ((pair-of
                  (pair-of
                    ?swn#1197
                    (pair-of (pair-of ?uwn#1200 ?twn#1197) ?swn#1200))
                  ?twn#1200))
               ?uwn#1200))))
     (caaddr
       .
       (forall
         (s t u v w x y z a b c d e f g h i j k l m n o p)
         (all-of
           (-> ((list-of (list-of s))) s)
           (-> ((pair-of t (list-of (list-of u)))) u)
           (-> ((pair-of v (pair-of w (list-of (list-of x))))) x)
           (-> ((pair-of y (pair-of z (pair-of (list-of b) a)))) b)
           (-> ((list-of (pair-of d c))) d)
           (-> ((pair-of e (list-of (pair-of g f)))) g)
           (-> ((pair-of h (pair-of i (list-of (pair-of k j))))) k)
           (-> ((pair-of l (pair-of m (pair-of (pair-of p n) o))))
               p))))
     (cadaar
       .
       (forall
         (s t
            u
            v
            w
            x
            y
            z
            a
            b
            c
            d
            e
            f
            g
            h
            i
            j
            k
            l
            m
            n
            o
            p
            q
            r
            ?swn#455
            ?twn#455
            ?uwn#455
            ?swn#458
            ?twn#458
            ?uwn#458
            ?swn#461
            ?twn#461
            ?uwn#461
            ?vwn#461
            ?swn#472
            ?twn#472
            ?uwn#472
            ?swn#475
            ?twn#475
            ?uwn#475
            ?vwn#475
            ?swn#478
            ?twn#478
            ?uwn#478
            ?vwn#478
            ?swn#481
            ?twn#481
            ?uwn#481
            ?vwn#481
            ?wwn#481)
         (all-of
           (-> ((list-of (list-of (list-of s)))) s)
           (-> ((pair-of (list-of (list-of u)) t)) u)
           (-> ((list-of (pair-of (list-of w) v))) w)
           (-> ((pair-of (pair-of (list-of z) x) y)) z)
           (-> ((list-of (list-of (pair-of a (list-of b))))) b)
           (-> ((pair-of (list-of (pair-of c (list-of e))) d)) e)
           (-> ((list-of (pair-of (pair-of f (list-of h)) g))) h)
           (-> ((pair-of (pair-of (pair-of i (list-of l)) j) k)) l)
           (-> ((list-of (list-of (pair-of m (pair-of o n))))) o)
           (-> ((pair-of (list-of (pair-of p (pair-of ?swn#455 q))) r))
               ?swn#455)
           (-> ((list-of
                  (pair-of
                    (pair-of ?twn#455 (pair-of ?twn#458 ?uwn#455))
                    ?swn#458)))
               ?twn#458)
           (-> ((pair-of
                  (pair-of
                    (pair-of ?uwn#458 (pair-of ?vwn#461 ?swn#461))
                    ?twn#461)
                  ?uwn#461))
               ?vwn#461))))
     (cadadr
       .
       (forall
         (s t
            u
            v
            w
            x
            y
            z
            a
            b
            c
            d
            e
            f
            g
            h
            i
            j
            k
            l
            m
            n
            o
            p
            q
            r
            ?swn#665
            ?twn#665
            ?swn#668
            ?twn#668
            ?uwn#668
            ?swn#671
            ?twn#671
            ?swn#674
            ?twn#674
            ?uwn#674
            ?swn#677
            ?twn#677
            ?uwn#677
            ?vwn#677
            ?swn#686
            ?twn#686
            ?uwn#686
            ?swn#689
            ?twn#689
            ?uwn#689
            ?vwn#689
            ?swn#692
            ?twn#692
            ?uwn#692
            ?vwn#692
            ?wwn#692)
         (all-of
           (-> ((list-of (list-of s))) s)
           (-> ((pair-of t (list-of (list-of u)))) u)
           (-> ((pair-of v (pair-of (list-of x) w))) x)
           (-> ((list-of (pair-of y (list-of z)))) z)
           (-> ((pair-of a (list-of (pair-of b (list-of c))))) c)
           (-> ((pair-of d (pair-of (pair-of e (list-of g)) f))) g)
           (-> ((list-of (pair-of h (pair-of j i)))) j)
           (-> ((pair-of k (list-of (pair-of l (pair-of n m))))) n)
           (-> ((pair-of
                  o
                  (pair-of (pair-of p (pair-of ?swn#665 q)) r)))
               ?swn#665))))
     (caddar
       .
       (forall
         (s t u v w x y z a b c d e f g h i j k l m n o p)
         (all-of
           (-> ((list-of (list-of s))) s)
           (-> ((pair-of (list-of u) t)) u)
           (-> ((list-of (pair-of v (list-of w)))) w)
           (-> ((pair-of (pair-of x (list-of z)) y)) z)
           (-> ((list-of (pair-of a (pair-of b (list-of c))))) c)
           (-> ((pair-of (pair-of d (pair-of e (list-of g))) f)) g)
           (-> ((list-of (pair-of h (pair-of i (pair-of k j))))) k)
           (-> ((pair-of (pair-of l (pair-of m (pair-of p n))) o))
               p))))
     (cadddr
       .
       (forall
         (s t u v w x y z a b c d e f g)
         (all-of
           (-> ((list-of s)) s)
           (-> ((pair-of t (list-of u))) u)
           (-> ((pair-of v (pair-of w (list-of x)))) x)
           (-> ((pair-of y (pair-of z (pair-of a (list-of b))))) b)
           (-> ((pair-of c (pair-of d (pair-of e (pair-of g f)))))
               g))))
     (cdaaar
       .
       (forall
         (s t
            u
            v
            w
            x
            y
            z
            a
            b
            c
            d
            e
            f
            g
            h
            i
            j
            k
            l
            m
            n
            o
            p
            q
            r
            ?swn#222
            ?twn#222
            ?uwn#222
            ?swn#225
            ?twn#225
            ?uwn#225
            ?swn#228
            ?twn#228
            ?uwn#228
            ?vwn#228
            ?swn#231
            ?twn#231
            ?uwn#231
            ?swn#234
            ?twn#234
            ?uwn#234
            ?vwn#234
            ?swn#237
            ?twn#237
            ?uwn#237
            ?vwn#237
            ?swn#240
            ?twn#240
            ?uwn#240
            ?vwn#240
            ?wwn#240)
         (all-of
           (-> ((list-of (list-of (list-of (list-of s))))) (list-of s))
           (-> ((pair-of (list-of (list-of (list-of u))) t))
               (list-of u))
           (-> ((list-of (pair-of (list-of (list-of w)) v)))
               (list-of w))
           (-> ((pair-of (pair-of (list-of (list-of z)) x) y))
               (list-of z))
           (-> ((list-of (list-of (pair-of (list-of b) a))))
               (list-of b))
           (-> ((pair-of (list-of (pair-of (list-of e) c)) d))
               (list-of e))
           (-> ((list-of (pair-of (pair-of (list-of h) f) g)))
               (list-of h))
           (-> ((pair-of (pair-of (pair-of (list-of l) i) j) k))
               (list-of l))
           (-> ((list-of (list-of (list-of (pair-of m n))))) n)
           (-> ((pair-of (list-of (list-of (pair-of o q))) p)) q)
           (-> ((list-of
                  (pair-of (list-of (pair-of r ?twn#222)) ?swn#222)))
               ?twn#222)
           (-> ((pair-of
                  (pair-of (list-of (pair-of ?uwn#222 ?uwn#225)) ?swn#225)
                  ?twn#225))
               ?uwn#225)
           (-> ((list-of
                  (list-of (pair-of (pair-of ?swn#228 ?uwn#228) ?twn#228))))
               ?uwn#228)
           (-> ((pair-of
                  (list-of (pair-of (pair-of ?vwn#228 ?uwn#231) ?swn#231))
                  ?twn#231))
               ?uwn#231)
           (-> ((list-of
                  (pair-of
                    (pair-of (pair-of ?swn#234 ?vwn#234) ?twn#234)
                    ?uwn#234)))
               ?vwn#234)
           (-> ((pair-of
                  (pair-of
                    (pair-of (pair-of ?swn#237 ?swn#240) ?twn#237)
                    ?uwn#237)
                  ?vwn#237))
               ?swn#240))))
     (cdaadr
       .
       (forall
         (s t
            u
            v
            w
            x
            y
            z
            a
            b
            c
            d
            e
            f
            g
            h
            i
            j
            k
            l
            m
            n
            o
            p
            q
            r
            ?uwn#284
            ?swn#287
            ?twn#287
            ?uwn#287
            ?vwn#287
            ?swn#290
            ?twn#290
            ?swn#293
            ?twn#293
            ?uwn#293
            ?swn#296
            ?twn#296
            ?uwn#296
            ?vwn#296
            ?swn#299
            ?twn#299
            ?uwn#299
            ?swn#302
            ?twn#302
            ?uwn#302
            ?vwn#302
            ?swn#305
            ?twn#305
            ?uwn#305
            ?vwn#305
            ?wwn#305)
         (all-of
           (-> ((list-of (list-of (list-of s)))) (list-of s))
           (-> ((pair-of t (list-of (list-of (list-of u)))))
               (list-of u))
           (-> ((pair-of v (pair-of (list-of (list-of x)) w)))
               (list-of x))
           (-> ((list-of (pair-of (list-of z) y))) (list-of z))
           (-> ((pair-of a (list-of (pair-of (list-of c) b))))
               (list-of c))
           (-> ((pair-of d (pair-of (pair-of (list-of g) e) f)))
               (list-of g))
           (-> ((list-of (list-of (pair-of h i)))) i)
           (-> ((pair-of j (list-of (list-of (pair-of k l))))) l)
           (-> ((pair-of m (pair-of (list-of (pair-of n p)) o))) p)
           (-> ((list-of (pair-of (pair-of q ?uwn#284) r))) ?uwn#284)
           (-> ((pair-of
                  ?swn#287
                  (list-of (pair-of (pair-of ?twn#287 ?vwn#287) ?uwn#287))))
               ?vwn#287)
           (-> ((pair-of
                  ?swn#290
                  (pair-of
                    (pair-of (pair-of ?twn#290 ?uwn#293) ?swn#293)
                    ?twn#293)))
               ?uwn#293))))
     (cdadar
       .
       (forall
         (s t
            u
            v
            w
            x
            y
            z
            a
            b
            c
            d
            e
            f
            g
            h
            i
            j
            k
            l
            m
            n
            o
            p
            q
            r
            ?uwn#922
            ?swn#925
            ?twn#925
            ?uwn#925
            ?vwn#925
            ?swn#928
            ?twn#928
            ?swn#931
            ?twn#931
            ?uwn#931
            ?swn#934
            ?twn#934
            ?uwn#934
            ?swn#937
            ?twn#937
            ?uwn#937
            ?vwn#937
            ?swn#940
            ?twn#940
            ?uwn#940
            ?vwn#940
            ?swn#943
            ?twn#943
            ?uwn#943
            ?vwn#943
            ?wwn#943)
         (all-of
           (-> ((list-of (list-of (list-of s)))) (list-of s))
           (-> ((pair-of (list-of (list-of u)) t)) (list-of u))
           (-> ((list-of (pair-of v (list-of (list-of w)))))
               (list-of w))
           (-> ((pair-of (pair-of x (list-of (list-of z))) y))
               (list-of z))
           (-> ((list-of (pair-of a (pair-of (list-of c) b))))
               (list-of c))
           (-> ((pair-of (pair-of d (pair-of (list-of g) e)) f))
               (list-of g))
           (-> ((list-of (list-of (pair-of h i)))) i)
           (-> ((pair-of (list-of (pair-of j l)) k)) l)
           (-> ((list-of (pair-of m (list-of (pair-of n o))))) o)
           (-> ((pair-of (pair-of p (list-of (pair-of q ?uwn#922))) r))
               ?uwn#922)
           (-> ((list-of
                  (pair-of
                    ?swn#925
                    (pair-of (pair-of ?twn#925 ?vwn#925) ?uwn#925))))
               ?vwn#925)
           (-> ((pair-of
                  (pair-of
                    ?swn#928
                    (pair-of (pair-of ?twn#928 ?uwn#931) ?swn#931))
                  ?twn#931))
               ?uwn#931))))
     (cdaddr
       .
       (forall
         (s t u v w x y z a b c d e f g h i j k l m n o p)
         (all-of
           (-> ((list-of (list-of s))) (list-of s))
           (-> ((pair-of t (list-of (list-of u)))) (list-of u))
           (-> ((pair-of v (pair-of w (list-of (list-of x)))))
               (list-of x))
           (-> ((pair-of y (pair-of z (pair-of (list-of b) a))))
               (list-of b))
           (-> ((list-of (pair-of c d))) d)
           (-> ((pair-of e (list-of (pair-of f g)))) g)
           (-> ((pair-of h (pair-of i (list-of (pair-of j k))))) k)
           (-> ((pair-of l (pair-of m (pair-of (pair-of n p) o))))
               p))))
     (cddaar
       .
       (forall
         (s t
            u
            v
            w
            x
            y
            z
            a
            b
            c
            d
            e
            f
            g
            h
            i
            j
            k
            l
            m
            n
            o
            p
            q
            r
            ?swn#130
            ?twn#130
            ?uwn#130
            ?swn#133
            ?twn#133
            ?uwn#133
            ?swn#136
            ?twn#136
            ?uwn#136
            ?vwn#136
            ?swn#147
            ?twn#147
            ?uwn#147
            ?swn#150
            ?twn#150
            ?uwn#150
            ?vwn#150
            ?swn#153
            ?twn#153
            ?uwn#153
            ?vwn#153
            ?swn#156
            ?twn#156
            ?uwn#156
            ?vwn#156
            ?wwn#156)
         (all-of
           (-> ((list-of (list-of (list-of s)))) (list-of s))
           (-> ((pair-of (list-of (list-of u)) t)) (list-of u))
           (-> ((list-of (pair-of (list-of w) v))) (list-of w))
           (-> ((pair-of (pair-of (list-of z) x) y)) (list-of z))
           (-> ((list-of (list-of (pair-of a (list-of b)))))
               (list-of b))
           (-> ((pair-of (list-of (pair-of c (list-of e))) d))
               (list-of e))
           (-> ((list-of (pair-of (pair-of f (list-of h)) g)))
               (list-of h))
           (-> ((pair-of (pair-of (pair-of i (list-of l)) j) k))
               (list-of l))
           (-> ((list-of (list-of (pair-of m (pair-of n o))))) o)
           (-> ((pair-of (list-of (pair-of p (pair-of q ?swn#130))) r))
               ?swn#130)
           (-> ((list-of
                  (pair-of
                    (pair-of ?twn#130 (pair-of ?uwn#130 ?twn#133))
                    ?swn#133)))
               ?twn#133)
           (-> ((pair-of
                  (pair-of
                    (pair-of ?uwn#133 (pair-of ?swn#136 ?vwn#136))
                    ?twn#136)
                  ?uwn#136))
               ?vwn#136))))
     (cddadr
       .
       (forall
         (s t
            u
            v
            w
            x
            y
            z
            a
            b
            c
            d
            e
            f
            g
            h
            i
            j
            k
            l
            m
            n
            o
            p
            q
            r
            ?swn#340
            ?twn#340
            ?swn#343
            ?twn#343
            ?uwn#343
            ?swn#346
            ?twn#346
            ?swn#349
            ?twn#349
            ?uwn#349
            ?swn#352
            ?twn#352
            ?uwn#352
            ?vwn#352
            ?swn#361
            ?twn#361
            ?uwn#361
            ?swn#364
            ?twn#364
            ?uwn#364
            ?vwn#364
            ?swn#367
            ?twn#367
            ?uwn#367
            ?vwn#367
            ?wwn#367)
         (all-of
           (-> ((list-of (list-of s))) (list-of s))
           (-> ((pair-of t (list-of (list-of u)))) (list-of u))
           (-> ((pair-of v (pair-of (list-of x) w))) (list-of x))
           (-> ((list-of (pair-of y (list-of z)))) (list-of z))
           (-> ((pair-of a (list-of (pair-of b (list-of c)))))
               (list-of c))
           (-> ((pair-of d (pair-of (pair-of e (list-of g)) f)))
               (list-of g))
           (-> ((list-of (pair-of h (pair-of i j)))) j)
           (-> ((pair-of k (list-of (pair-of l (pair-of m n))))) n)
           (-> ((pair-of
                  o
                  (pair-of (pair-of p (pair-of q ?swn#340)) r)))
               ?swn#340))))
     (cdddar
       .
       (forall
         (s t u v w x y z a b c d e f g h i j k l m n o p)
         (all-of
           (-> ((list-of (list-of s))) (list-of s))
           (-> ((pair-of (list-of u) t)) (list-of u))
           (-> ((list-of (pair-of v (list-of w)))) (list-of w))
           (-> ((pair-of (pair-of x (list-of z)) y)) (list-of z))
           (-> ((list-of (pair-of a (pair-of b (list-of c)))))
               (list-of c))
           (-> ((pair-of (pair-of d (pair-of e (list-of g))) f))
               (list-of g))
           (-> ((list-of (pair-of h (pair-of i (pair-of j k))))) k)
           (-> ((pair-of (pair-of l (pair-of m (pair-of n p))) o))
               p))))
     (cddddr
       .
       (forall
         (s t u v w x y z a b c d e f g)
         (all-of
           (-> ((list-of s)) (list-of s))
           (-> ((pair-of t (list-of u))) (list-of u))
           (-> ((pair-of v (pair-of w (list-of x)))) (list-of x))
           (-> ((pair-of y (pair-of z (pair-of a (list-of b)))))
               (list-of b))
           (-> ((pair-of c (pair-of d (pair-of e (pair-of f g)))))
               g))))

     (null?    . (forall (T U) (all-of (-> ((list-of T)) boolean) ; restrictive
				       (-> (U) boolean))))
     (list?    . (forall (T) (type-predicate-for (list-of T))))
     (list     . (forall (T) (all-of (-> (T ...) (list-of T))
				     (-> (datum ...) (list-of datum)))))
     (length   . (forall (T) (-> ((list-of T)) number)))
     (append   . (forall (T W X)
			 (all-of (-> ((list-of T) ...) (list-of T))
				 (-> ((list-of W) (list-of X) ...) (list-of datum)))))
     (reverse  . (forall (T) (-> ((list-of T)) (list-of T))))
     (list-tail . (forall (T) (-> ((list-of T) number) (list-of T))))
     (list-ref . (forall (T) (-> ((list-of T) number) T)))
     (memq     . (forall (T) (-> (T (list-of T)) boolean)))  ; restrictive
     (memv     . (forall (T) (-> (T (list-of T)) boolean)))  ; restrictive
     (member   . (forall (T) (-> (T (list-of T)) boolean)))  ; restrictive
     (assq     . (forall (T V S)                          ; restrictive
			 (all-of (-> (T (list-of (list-of T))) (list-of T))
				 (-> (V (list-of (pair-of V S))) (pair-of V S)))))
     (assv     . (forall (T V S)                          ; restrictive
			 (all-of (-> (T (list-of (list-of T))) (list-of T))
				 (-> (V (list-of (pair-of V S))) (pair-of V S)))))
     (assoc    . (forall (T V S)                          ; restrictive
			 (all-of (-> (T (list-of (list-of T))) (list-of T))
				 (-> (V (list-of (pair-of V S))) (pair-of V S)))))
     ;; 6.3.3
     (symbol?  . (type-predicate-for symbol))
     (symbol->string . (-> (symbol) string))
     (string->symbol . (-> (string) symbol))
     ;; 6.3.4
     (char?       . (type-predicate-for char))
     (char=?      . (-> (char char) boolean))
     (char<?      . (-> (char char) boolean))
     (char>?      . (-> (char char) boolean))
     (char<=?     . (-> (char char) boolean))
     (char>=?     . (-> (char char) boolean))
     (char-ci=?   . (-> (char char) boolean))
     (char-ci<?   . (-> (char char) boolean))
     (char-ci>?   . (-> (char char) boolean))
     (char-ci<=?  . (-> (char char) boolean))
     (char-ci>=?  . (-> (char char) boolean))
     (char-alphabetic? . (-> (char) boolean))
     (char-numeric?    . (-> (char) boolean))
     (char-whitespace? . (-> (char) boolean))
     (char-upper-case? . (-> (char) boolean))
     (char-lower-case? . (-> (char) boolean))
     (char->integer    . (-> (char) number))
     (integer->char    . (-> (number) char))
     (char-upcase      . (-> (char) char))
     (char-downcase    . (-> (char) char))
     ;; 6.3.5
     (string?          . (type-predicate-for string))
     (make-string      . (all-of (-> (number char) string)
				 (-> (number) string))) ; unsound
     (string           . (-> (char ...) string))
     (string-length    . (-> (string) number))
     (string-ref       . (-> (string number) char))
     (string-set!      . (-> (string number char) void))     
     (string=?         . (-> (string ...) boolean))
     (string-ci=?      . (-> (string ...) boolean))
     (string<?         . (-> (string string) boolean))
     (string>?         . (-> (string string) boolean))
     (string<=?        . (-> (string string) boolean))
     (string>=?        . (-> (string string) boolean))
     (string-ci<?      . (-> (string string) boolean))
     (string-ci>?      . (-> (string string) boolean))
     (string-ci<=?     . (-> (string string) boolean))
     (string-ci>=?     . (-> (string string) boolean))
     (substring        . (-> (string number number) string))     
     (string-append    . (-> (string ...) string))
     (string->list     . (-> (string) (list-of char)))
     (list->string     . (-> ((list-of char)) string))
     (string-copy      . (-> (string) string))
     (string-fill!     . (-> (string) char))
     ;; 6.3.6
     (vector?          . (forall (T) (type-predicate-for (vector-of T))))
     (make-vector      . (forall (T U)
				 (all-of (-> (number T) (vector-of T))
					 (-> (number) (vector-of U))))) ; unsound
     (vector           . (forall (T)
				 (all-of (-> (T ...) (vector-of T))
					 (-> (datum ...) (vector-of datum)))))
     (vector-length    . (forall (T) (-> ((vector-of T)) number)))
     (vector-ref       . (forall (T) (-> ((vector-of T) number) T)))
     (vector-set!      . (forall (T) (-> ((vector-of T) number T) void)))
     (vector->list     . (forall (T) (-> ((vector-of T)) (list-of T))))
     (list->vector     . (forall (T) (-> ((list-of T)) (vector-of T))))
     (vector-fill!     . (forall (T) (-> ((vector-of T) T) void)))
     ;; 6.4
     (procedure?       . (forall (T) (type-predicate-for (-> (datum ...) T))))
     (apply            . (forall
			  (D E T U F G H I V S X W Y Z A C B J K) 
			  (all-of (-> ((-> (D ...) E) (list-of D)) E)
				  (-> ((-> (T) U) (list-of T)) U)
				  (-> ((-> (F F) G) (list-of F)) G)
				  (-> ((-> (H H H) I) (list-of H)) I)
				  (-> ((-> (V S X ...) W)
				       (pair-of V (pair-of S (list-of X))))
				      W)
				  (-> ((-> (Y Z A C ...) B)
				       (pair-of
					Y
					(pair-of Z
						 (pair-of A (list-of C)))))
				      B)
				  (-> ((-> (J ...) K) J ...) K)
				  )))		; restrictive
     (map      . (forall (T U W X Y Z A B C D E)
			 (all-of (-> ((-> (T) U) (list-of T)) (list-of U))
				 (-> ((-> (W X) Y)
				      (list-of W) (list-of X))
				     (list-of Y))
				 (-> ((-> (Z A B) C)
				      (list-of Z) (list-of A) (list-of B))
				     (list-of C))
				 (-> ((-> (D D ...) E) (list-of D)
				      (list-of D) ...)
				     (list-of E))
				 )))			; restrictive
     (for-each . (forall (T W X Z A B D)
			 (all-of (-> ((-> (T) void) (list-of T)) void)
				 (-> ((-> (W X) void)
				      (list-of W) (list-of X))
				     void)
				 (-> ((-> (Z A B) void)
				      (list-of Z) (list-of A) (list-of B))
				     void)
				 (-> ((-> (D ...) void) (list-of D) ...) void)
				 )))			; restrictive
     (force     . (forall (T) (-> ((promise-of T)) T)))
     (call-with-current-continuation  .
				      (forall (T V)
					(-> ((-> ((-> (T) poof)) V)) T)))
     (values                . (-> (datum ...) poof))
     (call-with-values      . (forall (T)
				      (-> ((-> () poof) (-> (poof ...) T)) T)))
     (dynamic-wind          . (forall (T)
				 (-> ((-> () void) (-> () T) (-> () void)) T)))
     ;; 6.5
     (eval                  . (-> (datum scheme-environment) datum))
     (scheme-report-environment . (-> (number) scheme-environment))
     (null-environment          . (-> (number) scheme-environment))
     (interaction-environment   . (-> () scheme-environment))
     ;; 6.6.1
     (call-with-input-file  . (forall (T) (-> (string (-> (input-port) T)) T)))
     (call-with-output-file . (forall (T)
				      (-> (string (-> (output-port) T)) T)))
     (input-port?           . (type-predicate-for input-port))
     (output-port?          . (type-predicate-for output-port))
     (current-input-port    . (-> () input-port))
     (current-output-port   . (-> () output-port))
     (with-input-from-file  . (forall (T) (-> (string (-> () T)) T)))
     (with-output-to-file   . (forall (T) (-> (string (-> () T)) T)))
     (open-input-file       . (-> (string) input-port))
     (open-output-file      . (-> (string) output-port))
     (close-input-port      . (-> (output-port) void))
     (close-output-port     . (-> (output-port) void))
     ;; 6.6.2
     (read       . (all-of (-> () datum)
			   (-> (input-port) datum)))
     (read-char  . (all-of (-> () char)
			   (-> (input-port) char)))
     (peek-char  . (all-of (-> () char)
			   (-> (input-port) char))) ; restrictive
     (eof-object? . (-> (datum) boolean))
     (char-ready? . (-> (input-port) boolean))
     ;; 6.6.3
     (write       . (forall (T U) (all-of (-> (T) void)
					  (-> (U output-port) void))))
     (display     . (forall (T U) (all-of (-> (T) void)
					  (-> (U output-port) void))))
     (newline     . (all-of (-> () void)
			    (-> (output-port) void)))
     (write-char  . (all-of (-> (char) void)
			    (-> (char output-port) void)))
     ;; 6.6.4
     (load           . (-> (string) void))
     (transcript-on  . (-> (string) void))
     (transcript-off . (-> () void))

     ;; procedures added for the class and type checker below

     ;; from localize.scm
     (*cs342:libdir* . string)
     (*cs342:hwdir* . string)
     (*cs342:scheme-implementation-vicinity* . string)
     ;; from load-from-lib.scm (via localize.scm)
     (std:load  . (-> (string) void))
     (load-from-lib  . (-> (string) void))
     (load-quietly-from-lib  . (-> (string) void))
     (load-quietly   . (-> (string) void))
     ;; from displayln.scm (via standardize-*.scm)
     (displayln      . (-> (datum ...) void))
     ;; from writeln.scm (via test-homework.scm, via standardize-*.scm)
     (writeln        . (-> (datum ...) void))
     ;; from test-homework.scm (via standardize-*.scm)
     (test-hw1       . (-> (string) void)) 
     (test-hw2       . (-> (string) void))
     (test-hw2b      . (-> (string) void))
     (test-hw3       . (-> (string) void))
     (test-hw4       . (-> (string) void))
     (test-hw5       . (-> (string) void))
     (test-hw6       . (-> (string) void))
     (test-hw7       . (-> (string) void))
     (test-hw8       . (-> (string) void))
     (test-hw9       . (-> (string) void))
     (test-hw10      . (-> (string) void))
     (test-hw11      . (-> (string) void))
     (test-hw12      . (-> (string) void))
     (test-ex        . (-> (string) void))
     (run-test-case-maker . (-> ((-> (datum) void) (-> (datum) void))
				(-> (datum) void)))
     (run-test-case   . (-> (datum) void))
     (regression-test-maker . (forall (S T) (-> ((-> (S T) boolean)
						 (-> (T) void)
						 (-> (T) void))
						(-> (T symbol T) void))))
     (run-regression-test . (-> (datum symbol datum) void))
     (run-regression-tests-maker . (-> ((-> (datum datum) boolean)
					(-> (datum) void)
					(-> (datum) void))
				       (-> ((list-of (list-of datum))) void)))
     (run-regression-tests-equal? . (-> ((list-of (list-of datum))) void))
     ;; from standardize-{drscheme,chez}.scm (things used from SCM/SLIB)
     (error          . (-> (datum ...) poof))
     (exit           . (all-of (-> () poof)
			       (-> (number) poof)))
     (quit           . (-> () poof))
     (file-exists?   . (-> (string) boolean))
     (current-error-port   . (-> () output-port))
     (force-output   . (-> () void))
     (require        . (-> (symbol) void))
     (verbose        . (all-of (-> (number) void)
			       (-> () number)))
     (abort          . (-> () poof))
     (random         . (-> (number) number))
     (trace          . (-> (datum ...) void))
     (untrace        . (-> (datum ...) void))
     (pretty-print   . (-> (datum) void))
     (void           . (-> () void))
     (promise?       . (-> (datum) boolean))
     ;; EOPL2e additions from r5rs.scm (via {chez,drscheme}-init.scm)
     (eopl:printf    . (-> (string datum ...) void))
     (eopl:pretty-print . (forall (T) (-> (T) void)))
     (eopl:error-stop . (-> () poof))
     (eopl:error . (-> (symbol string datum ...) poof))
     ;; EOPL2e additions from sllgen.scm (via {chez,drscheme}-init.scm)
     (sllgen:make-string-parser . (forall (AST)
				    (-> ((list-of (list-of datum))
					 (list-of (list-of datum)))
					(-> (string) AST))))
     (sllgen:make-stream-parser . (forall (AST)
				    (-> ((list-of (list-of datum))
					 (list-of (list-of datum)))
					(-> (sllgen:stream) AST))))
     (sllgen:make-stream-scanner . (forall (TOK)
				     (-> ((list-of (list-of datum))
					  (list-of (list-of datum)))
					(-> (sllgen:stream) (list-of TOK)))))
     (sllgen:make-string-scanner . (forall (TOK)
				     (-> ((list-of (list-of datum))
					  (list-of (list-of datum)))
					(-> (string) (list-of TOK)))))
     (sllgen:make-define-datatypes . 
				  (-> ((list-of (list-of datum))
				       (list-of (list-of datum)))
				      void))
     (sllgen:show-define-datatypes . 
				  (-> ((list-of (list-of datum))
				       (list-of (list-of datum)))
				      void))
     (sllgen:list-define-datatypes . 
				  (-> ((list-of (list-of datum))
				       (list-of (list-of datum)))
				      (list-of datum)))
     (sllgen:make-rep-loop .     (forall (AST VAL)
				  (-> (string (-> (AST) VAL)
					      (-> (sllgen:stream) AST))
				      poof)))
     ;; EOPL2e from test-harness.scm (via {chez,drscheme}-init.scm)
     (run-experiment . (all-of
			(-> ((-> (string) Expressed-Value)
			    (-> (datum) Expressed-Value)
			    (list-of symbol)
			    (list-of datum))
			   void)
		        (-> ((-> (string) datum)
			     (-> (datum) datum)
			     (list-of symbol)
			     (list-of datum))
			    void)))
     (use-execution-outcome . (-> (datum) datum))
     (all-tests . (list-of datum))
     (run-test . (all-of
		  (-> ((-> (string) Expressed-Value) symbol) void)
	          (-> ((-> (string) datum) symbol) void)))
     (initialize-test-suite! . (-> () void))
     (add-test! . (-> (symbol symbol string datum) void))
     (add-typed-test! . (-> (symbol symbol string datum datum) void))
     ;; from type-predicates.scm (via type-check-ignore-types-at-runtime.scm)
     (datum?          .     (type-predicate-for datum))
     (...             .     (type-predicate-for ...))
     (void?           .     (type-predicate-for void))
     (poof?           .     (type-predicate-for poof))
     (pair-of         .     (forall (S T)
			      (all-of
			       (-> ((type-predicate-for S)
				    (type-predicate-for T))
				   (type-predicate-for (pair-of S T)))
			       (-> ((-> (datum) boolean)
				    (-> (datum) boolean))
				    (-> (datum) boolean)))))
     (vector-of       .     (forall (T)
			      (all-of
			       (-> ((type-predicate-for T))
				   (type-predicate-for (vector-of T)))
			       (-> ((-> (datum) boolean))
				    (-> (datum) boolean)))))
     (<-              .     (forall (R S T U V W X Y Z A B C D E F)
			       (all-of
				(-> ((type-predicate-for R))
				    (type-predicate-for (-> () R)))
				(-> ((type-predicate-for S)
				     (type-predicate-for T))
				    (type-predicate-for (-> (T) S)))
				(-> ((type-predicate-for U)
				     (type-predicate-for V)
				     (type-predicate-for W))
				    (type-predicate-for (-> (V W) U)))
				(-> ((type-predicate-for X)
				     (type-predicate-for Y)
				     (type-predicate-for Z)
				     (type-predicate-for A))
				    (type-predicate-for (-> (Y Z A) X)))
				(-> ((type-predicate-for B)
				     (type-predicate-for C)
				     (type-predicate-for D)
				     (type-predicate-for E)
				     (type-predicate-for F) ...)
				    (type-predicate-for (-> (C D E F ...) B)))
				(-> ((-> (datum) boolean)
				     (-> (datum) boolean) ...)
				    (-> (datum) boolean))
				)))
     (type-predicate-for .  (forall (T)
			        (-> ((type-predicate-for T))
				    (type-predicate-for
				     (type-predicate-for T)))))
     (list-of         .     (forall (T)
			      (all-of
			       (-> ((type-predicate-for T))
				   (type-predicate-for (list-of T)))
			       (-> ((-> (datum) boolean))
				    (-> (datum) boolean)))))
     (all-of          .     (forall (R S T U V W X Y Z A B C D E F)
			       (all-of
				(-> ((type-predicate-for R))
				    (type-predicate-for (all-of R)))
				(-> ((type-predicate-for S)
				     (type-predicate-for T))
				    (type-predicate-for (all-of S T)))
				(-> ((type-predicate-for U)
				     (type-predicate-for V)
				     (type-predicate-for W))
				    (type-predicate-for (all-of U V W)))
				(-> ((type-predicate-for X)
				     (type-predicate-for Y)
				     (type-predicate-for Z)
				     (type-predicate-for A))
				    (type-predicate-for (all-of X Y Z A)))
				(-> ((type-predicate-for B)
				     (type-predicate-for C)
				     (type-predicate-for D)
				     (type-predicate-for E)
				     (type-predicate-for F))
				    (type-predicate-for (all-of B C D E F)))
				(-> ((-> (datum) boolean)
				     (-> (datum) boolean) ...)
				    (-> (datum) boolean))
				)))
     ;; additions from the type checker itself
     (*eval-even-if-error* . boolean)
     (type-check-exit    . (-> () poof))
     (type-check-reset-env! . (-> () void))
     (type-prompting-on! . (-> () void))
     (type-prompting-off! . (-> () void))
     (type-prompting-on? . (-> () boolean))
     (type-checklib-on!  . (-> () void))
     (type-checklib-off! . (-> () void))
     (type-checklib-on?  . (-> () boolean))
     (type-tracing-set!  . (-> (number) void))
     (type-tracing-off!  . (-> () void))
     (type-tracing-level . (-> () number))
     (type-help          . (-> () void))
    )))

) ;; end module
