;;; $Id: tc-ignore-types-at-runtime-chez.scm,v 1.2 2005/03/18 07:07:45 leavens Exp $
;;;
;;; Use this file if you aren't using the type checker but are working with
;;; Scheme files that have deftypes and defreps in them

;;; The following are R5RS macros.

(define-syntax has-type
    (syntax-rules ()
      ((has-type type-exp expression)
       expression)))

(define-syntax has-type-trusted
    (syntax-rules ()
      ((has-type-trusted type-exp expression)
       expression)))

(define-syntax test-type
  (lambda (x)
    (syntax-case x ( -> list-of vector-of pair-of all-of)
      ((test-type (-> (arg-pred ...) res-pred) expression)
       (with-syntax (((arg ...) (generate-temporaries (syntax (arg-pred ...)))))
		    (syntax (let ((val expression))
			      (if (procedure? val)
				  (lambda (arg ...)
				    (test-type
				     res-pred
				     (val (test-type arg-pred arg) ...)))
				  (eopl:error
				   'test-type
				   "Testing type with procedure? failed for ~s"
				   val))))))
      ((test-type (list-of elem-pred) expression)
       (syntax (let ((val expression))
		 (if (list? val)
		     (do ((ls val (cdr ls)))
			 ((null? ls) val)
		       (set-car! ls (test-type elem-pred (car ls))))
		     (eopl:error
		      'test-type
		      "Testing type with list? failed for ~s"
		      val)))))
      ((test-type (vector-of elem-pred) expression)
       (syntax (let ((val expression))
		 (if (vector? val)
		     (do ((i (- (vector-length val) 1) (- i 1)))
			 ((negative? i) val)
		       (vector-set! val i
				    (test-type elem-pred (vector-ref val i))))
		     (eopl:error
		      'test-type
		      "Testing type with vector? failed for ~s"
		      val)))))
      ((test-type (pair-of pred1 pred2) expression)
       (syntax (let ((val expression))
		 (if (pair? val)
		     (begin
		       (set-car! val (test-type pred1 (car val)))
		       (set-cdr! val (test-type pred2 (cdr val)))
		       val)
		     (eopl:error
		      'test-type
		      "Testing type with pair? failed for ~s"
		      val)))))
      ((test-type type-pred expression)
       (syntax (let ((val expression))
		 (if (type-pred val)
		     val
		     (eopl:error
		      'test-type
		      "Testing type with ~s failed for ~s"
		      'type-pred
		      val))))))))

;; The <- procedure is defined in type-predicates.scm
(define-syntax ->
  (syntax-rules ()
    ((-> (argument-predicate ...) result-predicate)
     (<- result-predicate argument-predicate ...))))

(define-syntax forall
  (syntax-rules ()
    ((forall type-predicate-formals predicate-expression)
     (lambda type-predicate-formals predicate-expression))))

;; The following relies on syntax-case and generate-temporaries.
;; These are supposed to be in both MzScheme and Chez Scheme.
;; See below if these don't work for you.
;; However, the more complex translation is preferred, as it makes
;; deftype be a definition form.
(define-syntax deftype
  (lambda (x)
    (syntax-case x ()
      ((deftype name type-exp)
       (with-syntax (((gname) (generate-temporaries (syntax (name)))))
	  (syntax (define gname #f)))))))

;; Use the following if you can't get the above definition of deftype to work
;; in your Scheme system.
; (define-syntax deftype
;   (syntax-rules ()
;     ((deftype name type-exp)
;      (void)))) ; should be a definition form, technically

(define-syntax defrep
  (syntax-rules (forall)
    ((defrep (forall (type-var ...) abstract-type concrete-type))
     (begin))
    ((defrep (abstract-type concrete-type))
     (begin))
    ((defrep (forall (type-var ...) abstract-type concrete-type)
       abs-conc-pair ...)
     (defrep abs-conc-pair ...))
    ((defrep (abstract-type concrete-type) abs-conc-pair ...)
     (defrep abs-conc-pair ...))
    ))

(load-quietly-from-lib "type-predicates.scm")
