;;; $Id: tc-parsing-stream.scm,v 1.1 2005/02/28 06:58:05 leavens Exp $
;;; streams for parsing

(module tc-parsing-stream (lib "typedscm.ss" "lib342")

(provide tc:stream-of tc:stream-car tc:stream-cdr tc:make-stream
         tc:the-null-stream tc:stream-null? tc:stream-for-each
         tc:make-input-stream) 
  
;;; EOPL Figure F.1 : page 493

(deftype tc:stream-of
  (forall (T) (-> ((type-predicate-for T))
		  (type-predicate-for (tc:stream T)))))
(deftype tc:stream-car (forall (T) (-> ((tc:stream T)) T)))
(deftype tc:stream-cdr (forall (T) (-> ((tc:stream T)) (tc:stream T))))
(deftype tc:make-stream
  (forall (T) (-> (T (-> () (tc:stream T))) (tc:stream T))))
(deftype tc:the-null-stream (forall (T) (-> () (tc:stream T))))
(deftype tc:stream-null? (forall (T) (-> ((tc:stream T)) boolean)))
(deftype tc:stream-for-each
  (forall (T) (-> ((-> (T) void) (tc:stream T)) void)))
(deftype tc:make-input-stream (-> () (tc:stream char)))

(defrep
  (forall (T) (tc:stream T) (-> () (pair-of T (-> () (list-of T))))))

(define tc:stream-of
  (forall (T?)
    (lambda (x)
      (procedure? x))))

(define tc:stream-car
  (lambda (s)
    (car (s))))

(define tc:stream-cdr
  (lambda (s)
    (cdr (s))))

(define tc:make-stream
  (lambda (car-val th)
    (lambda ()
      (cons car-val (let ((cdr-val #f))
                      (lambda ()
                        (if (pair? cdr-val)
                            cdr-val
                            (begin (set! cdr-val ((th))) cdr-val))))))))

(define tc:the-null-stream
  (lambda ()
    (tc:make-stream 'end-of-stream (lambda () tc:the-null-stream))))

(define tc:stream-null?
  (lambda (s)
    (eq? (tc:stream-car s) 'end-of-stream)))

;;; Figure F.2 : page 494

(define tc:stream-for-each
  (lambda (proc stream)
    (letrec
      ((loop
         (lambda (stream)
           (if (not (tc:stream-null? stream))
               (begin
                 (proc (tc:stream-car stream))
                 (loop (tc:stream-cdr stream)))))))
      (loop stream))))

(define tc:make-input-stream
  (lambda ()
    (let ((char (read-char)))
      (if (eof-object? char)
          (tc:the-null-stream)
          (tc:make-stream char tc:make-input-stream)))))
)
