;;; $Id: tc-scheme-abstract-syntax.scm,v 1.28 2005/03/30 16:40:32 dorn Exp $
;;;
;;; Abstract syntax for Scheme,
;;; with extensions for EOPL (2e) and our type checker.
;;;
;;; The names of nonterminals are, for the most part,
;;; taken directly from the formal syntax in
;;; The Revised(5) Report on Scheme [R5RS].
;;; A notable exception is the use of "form" for what R5RS calls
;;; "command or definition"

(module tc-scheme-abstract-syntax (lib "typedscm.ss" "lib342")

(provide
 ;; programs
 tc:scheme-program tc:scheme-program?
 tc:scheme-program-normal tc:scheme-program->position tc:scheme-program->forms
 ;; forms (<command or definition>s)
 tc:scheme-form tc:scheme-form?
 tc:command tc:load-command tc:definition tc:syntax-definition
 tc:begin-form tc:module tc:require-form
 tc:scheme-form->position tc:definition? tc:definition->definition
 tc:begin-form? tc:require-form? tc:load-command?
 ;; module-names
 tc:scheme-module-name tc:scheme-module-name?
 tc:module-id tc:unix-relative-path-string tc:file tc:lib
 tc:scheme-module-name->position tc:scheme-module-name->file-location
 ;; abs-conc-pairs
 tc:scheme-abs-conc-pair tc:scheme-abs-conc-pair?
 tc:abs-conc-pair-simple tc:abs-conc-pair-polymorphic
 tc:scheme-abs-conc-pair->position
 ;; definitions
 tc:scheme-definition tc:scheme-definition?
 tc:simple-define tc:procedure-define tc:grouped-define tc:define-datatype
 tc:deftype tc:provide tc:require tc:defrep tc:require-for-syntax
 tc:scheme-definition->position tc:provide->exported-ids
 tc:require->module-names tc:defrep->pairs tc:require? tc:defrep? tc:deftype?
 ;; formals
 tc:scheme-formals tc:scheme-formals?
 tc:normal-formals tc:unrestricted-formals tc:some-normal-formals
 tc:scheme-formals->names tc:scheme-formals->position
 ;; bodies
 tc:scheme-body-record tc:scheme-body?
 tc:scheme-body tc:scheme-body-record->position
;;; variant-decls
 tc:variant-decl-record tc:variant-decl?
 tc:variant-decl tc:variant-decl->variant-name tc:variant-decl->position
 ;; field-pred bindings
 tc:field-pred-binding-record tc:field-pred-binding?
 tc:field-pred-binding tc:field-pred-binding->field-name
 tc:field-pred-binding->predicate-exp tc:field-pred-binding->position
 ;; syntax specs
 tc:scheme-syntax-spec-record tc:scheme-syntax-spec?
 tc:scheme-syntax-spec
 ;; transformer specs
 tc:scheme-transformer-spec-record tc:scheme-transformer-spec?
 tc:scheme-transformer-spec
 ;; syntax rules
 tc:scheme-syntax-rule-record tc:scheme-syntax-rule?
 tc:scheme-syntax-rule
 ;; syntax patterns
 tc:scheme-syntax-pattern tc:scheme-syntax-pattern?
 tc:syntax-pattern-identifier tc:syntax-pattern-normal-list
 tc:syntax-pattern-dotted-list tc:syntax-pattern-ellipsis-list
 tc:syntax-pattern-normal-vector tc:syntax-pattern-ellipsis-vector
 tc:syntax-pattern-datum
 ;; syntax templates
 tc:scheme-syntax-template tc:scheme-syntax-template?
 tc:syntax-template-identifier tc:syntax-template-normal-list
 tc:syntax-template-dotted-list tc:syntax-template-normal-vector
 tc:syntax-template-datum
 ;; syntax template elements
 tc:scheme-syntax-template-element tc:scheme-syntax-template-element?
 tc:scheme-syntax-template-element-normal
 tc:scheme-syntax-template-element-ellipsis
 ;; expressions
 tc:scheme-exp tc:scheme-exp?
 tc:varref tc:quotation tc:self-evaluating tc:procedure-call tc:lambda-exp
 tc:named-lambda-exp tc:conditional-exp tc:conditional-stmt tc:assignment
 tc:macro-use tc:let-syntax-exp tc:letrec-syntax-exp tc:cond-exp tc:cond-stmt
 tc:case-exp tc:case-stmt tc:and-exp tc:or-exp tc:let-exp tc:named-let-exp
 tc:let*-exp tc:letrec-exp tc:begin-exp tc:do-exp tc:do-stmt tc:delay-exp
 tc:quasiquotation tc:cases-exp tc:cases-stmt
 tc:has-type-exp tc:has-type-trusted-exp tc:test-type-exp
 tc:arrow-exp tc:varargs-arrow-exp
 tc:forall-exp tc:define-datatype-exp tc:constructor-for-variant-type-exp tc:module-exp
 tc:scheme-exp->position
 tc:varref? tc:varref->variable tc:self-evaluating->datum
 tc:procedure-call->operator tc:procedure-call->operands
 ;; cond clause
 tc:scheme-cond-clause tc:scheme-cond-clause?
 tc:normal-cond-clause tc:test-only-cond-clause
 tc:test=>cond-clause tc:else-cond-clause
 tc:normal-cond-clause? tc:test=>cond-clause?
 tc:scheme-cond-clause->exp tc:scheme-cond-clause->position
 ;; case clause
 tc:scheme-case-clause tc:scheme-case-clause?
 tc:normal-case-clause tc:else-case-clause
 tc:scheme-case-clause->exp tc:scheme-case-clause->position
 ;; binding spec (for let, let*, letrec)
 tc:scheme-binding-spec-record tc:scheme-binding-spec?
 tc:scheme-binding-spec tc:scheme-binding-spec->position
 tc:scheme-binding-spec->variable tc:scheme-binding-spec->exp
 tc:scheme-binding-spec->varref
 ;; iteration spec (for do)
 tc:scheme-iteration-spec tc:scheme-iteration-spec?
 tc:stepped-iteration-spec tc:init-only-iteration-spec
 tc:scheme-iteration-spec->variable tc:scheme-iteration-spec->init
 tc:scheme-iteration-spec->position
 ;; cases clause
 tc:scheme-cases-clause tc:scheme-cases-clause?
 tc:normal-cases-clause tc:else-cases-clause
 tc:scheme-cases-clause->exp tc:scheme-cases-clause->position
 tc:else-cases-clause?
 ;; scheme-datum (for quotations and quasiquotations)
 tc:scheme-datum tc:scheme-datum?
 tc:simple-datum tc:normal-list-datum tc:improper-list-datum tc:quoted-datum
 tc:quasiquoted-datum tc:unquoted-datum tc:unquoted-exp
 tc:unquote-splicing-datum tc:unquote-splicing-exp tc:vector-datum
 tc:scheme-datum->position tc:simple-datum? tc:simple-datum->datum
 ;; input type expressions added for the type checker
 tc:input-type-exp tc:input-type-exp?
 tc:basic-input-type-exp tc:function-input-type-exp
 tc:varargs-function-input-type-exp tc:intersection-input-type-exp
 tc:applied-input-type-exp tc:polymorphic-input-type-exp
 tc:type-predicate-for-input-type-exp tc:variant-record-input-type-exp
 tc:translated-input-type-exp tc:input-type-exp->position
 ;; variant-type-decls
 tc:variant-type-decl-record tc:variant-type-decl?
 tc:variant-type-decl tc:variant-type-decl-record->position
 tc:variant-type-decl->variant-name
 ;; field-type bindings
 tc:field-type-binding-record tc:field-type-binding?
 tc:field-type-binding tc:field-type-binding->field-name
 tc:field-type-binding->type tc:scheme-syntax->position
 )

(require (lib "tc-position.scm" "lib342"))

;;; programs
(define-datatype tc:scheme-program tc:scheme-program?
  (tc:scheme-program-normal
   (position tc:position?)
   (require-forms (list-of tc:scheme-form?))
   (deftypes (list-of tc:scheme-definition?))
   (forms (list-of tc:scheme-form?))))

(deftype tc:scheme-program->position (-> (tc:scheme-program) tc:position))
(define tc:scheme-program->position
  (lambda (prog)
    (cases tc:scheme-program prog
      (tc:scheme-program-normal (position require-forms deftypes forms) position))))

(deftype tc:scheme-program->forms
  (-> (tc:scheme-program) (list-of tc:scheme-form)))
(define tc:scheme-program->forms
  (lambda (prog)
    (cases tc:scheme-program prog
      (tc:scheme-program-normal (position require-forms deftypes forms) forms))))

;;; forms (<command or definition>s)
(define-datatype tc:scheme-form tc:scheme-form?
  (tc:command 
   (position tc:position?)
   (exp tc:scheme-exp?))
  (tc:load-command
   (position tc:position?)
   (operator symbol?) (filename string?) (program tc:scheme-program?))
  (tc:definition
   (position tc:position?)
   (definition tc:scheme-definition?))
  (tc:syntax-definition
   (position tc:position?)
   (spec tc:scheme-syntax-spec?))
  (tc:begin-form
   (position tc:position?)
   (forms (list-of tc:scheme-form?)))
  ;; the following forms are added for the type checker
  (tc:module
   (position tc:position?)
   (identifier symbol?)
   (initial-required-module-name tc:scheme-module-name?)
   (provides (list-of symbol?))
   (requires-for-syntax (list-of tc:scheme-module-name?))
   (requires (list-of tc:scheme-module-name?))
   (defreps (list-of tc:scheme-abs-conc-pair?))
   (deftypes (list-of tc:scheme-definition?))
   (definitions (list-of tc:scheme-definition?)))
  (tc:require-form
   (position tc:position?)
   (module-names (list-of tc:scheme-module-name?)))
  )

(deftype tc:scheme-form->position (-> (tc:scheme-form) tc:position))
(define tc:scheme-form->position
  (lambda (exp)
    (cases tc:scheme-form exp
      (tc:command (position exp) position)
      (tc:load-command (position operator filename program) position)
      (tc:definition (position definition) position)
      (tc:syntax-definition (position spec) position)
      (tc:begin-form (position forms) position)
      (tc:module (position identifier initial-required-module-name
			   provides requires-for-syntax requires defreps deftypes definitions)
		 position)
      (tc:require-form (position module-names) position))))

(deftype tc:definition? (-> (tc:scheme-form) boolean))
(define tc:definition?
  (lambda (exp)
    (cases tc:scheme-form exp
      (tc:definition (position definition) #t)
      (else #f))))

(deftype tc:definition->definition (-> (tc:scheme-form) tc:scheme-definition))
(define tc:definition->definition
  (lambda (exp)
    (cases tc:scheme-form exp
      (tc:definition (position definition) definition)
      (else (error "tc:definition->definition called with non definition"
		   exp)))))

(deftype tc:begin-form? (-> (tc:scheme-form) boolean))
(define tc:begin-form?
  (lambda (exp)
    (cases tc:scheme-form exp
      (tc:begin-form (position forms) #t)
      (else #f))))

(deftype tc:require-form? (-> (tc:scheme-form) boolean))
(define tc:require-form?
  (lambda (exp)
    (cases tc:scheme-form exp
      (tc:require-form (position module-names) #t)
      (else #f))))

(deftype tc:load-command? (-> (tc:scheme-form) boolean))
(define tc:load-command?
  (lambda (exp)
    (cases tc:scheme-form exp
      (tc:load-command (position operator filename program) #t)
      (else #f))))

;;; module-names
(define-datatype tc:scheme-module-name tc:scheme-module-name?
  (tc:module-id
   (position tc:position?)
   (identifier symbol?))
  (tc:unix-relative-path-string
   (position tc:position?)
   (path string?))
  (tc:file
   (position tc:position?)
   (path string?))
  (tc:lib
   (position tc:position?)
   (filename string?)
   (collects (list-of string?))))

(deftype tc:scheme-module-name->position
  (-> (tc:scheme-module-name) tc:position))
(define tc:scheme-module-name->position
  (lambda (mn)
    (cases tc:scheme-module-name mn
      (tc:module-id (position identifier) position)
      (tc:unix-relative-path-string (position path) position)
      (tc:file (position path) position)
      (tc:lib (position filename collects) position))))

(deftype tc:scheme-module-name->file-location
  (-> (tc:scheme-module-name) string))
(define tc:scheme-module-name->file-location
  (lambda (exp)
    (cases tc:scheme-module-name exp
      (tc:module-id (position identifier)
		    (string-append (symbol->string identifier) ".scm"))
      (tc:unix-relative-path-string (position path) path)
      (tc:file (position path) path)
      (tc:lib (position filename collects)
	      (letrec ((dirname
			(lambda (cs ans)
			  (if (null? cs)
			      ans
			      (dirname
			       (cdr cs)
			       (string-append ans "/" (car cs)))))))
		(string-append (dirname collects "") "/" filename ))))))

;;; abs-conc-pairs
(define-datatype tc:scheme-abs-conc-pair tc:scheme-abs-conc-pair?
  (tc:abs-conc-pair-simple
   (position tc:position?)
   (abstract-type tc:input-type-exp?) (concrete-type tc:input-type-exp?))
  (tc:abs-conc-pair-polymorphic
   (position tc:position?)
   (type-vars (list-of symbol?))
   (abstract-type tc:input-type-exp?) (concrete-type tc:input-type-exp?)))

(deftype tc:scheme-abs-conc-pair->position
  (-> (tc:scheme-abs-conc-pair) tc:position))
(define tc:scheme-abs-conc-pair->position
  (lambda (exp)
    (cases tc:scheme-abs-conc-pair exp
      (tc:abs-conc-pair-simple (position abstract-type concrete-type) position)
      (tc:abs-conc-pair-polymorphic
       (position type-vars abstract-type concrete-type)
       position))))

;;; definitions
(define-datatype tc:scheme-definition tc:scheme-definition?
  (tc:simple-define 
   (position tc:position?)
   (variable symbol?) (exp tc:scheme-exp?))
  (tc:procedure-define
   (position tc:position?)
   (variable symbol?) (formals tc:scheme-formals?) (body tc:scheme-body?))
  (tc:grouped-define
   (position tc:position?)
   (definitions (list-of tc:scheme-definition?)))
  ;; the following definitions are added for EOPL
  (tc:define-datatype
   (position tc:position?)
   (type-name symbol?) (type-predicate-name symbol?)
   (variants (list-of tc:variant-decl?)))
  ;; the following definitions are added for the type checker
  (tc:deftype
   (position tc:position?)
   (variable symbol?) (type tc:input-type-exp?))
  ;; the following are only used internally in the type checker
  ;; (for parsing module bodies)
  (tc:provide
   (position tc:position?)
   (exported-ids (list-of symbol?)))
  (tc:require
   (position tc:position?)
   (module-names (list-of tc:scheme-module-name?)))
  (tc:require-for-syntax
   (position tc:position?)
   (module-names (list-of tc:scheme-module-name?)))
  (tc:defrep
   (position tc:position?)
   (pairs (list-of tc:scheme-abs-conc-pair?)))
  )

(deftype tc:scheme-definition->position
  (-> (tc:scheme-definition) tc:position))
(define tc:scheme-definition->position
  (lambda (exp)
    (cases tc:scheme-definition exp
      (tc:simple-define (position variable exp) position)
      (tc:procedure-define (position variable formals body) position)
      (tc:grouped-define (position definitions) position)
      (tc:define-datatype
       (position type-name type-predicate-name variants) position)
      (tc:deftype (position variable type) position)
      (tc:provide (position exported-ids) position)
      (tc:require (position module-names) position)
      (tc:defrep (position pairs) position)
      (tc:require-for-syntax (position module-names) position)
    )))

(deftype tc:provide->exported-ids
  (-> (tc:scheme-definition) (list-of symbol)))
(define tc:provide->exported-ids
  (lambda (exp)
    (cases tc:scheme-definition exp
      (tc:provide (position exported-ids) exported-ids)
      (else (error "tc:provide->exported-ids passed a non provide: " exp)))))

(deftype tc:require->module-names
  (-> (tc:scheme-definition) (list-of module-name)))
(define tc:require->module-names
  (lambda (exp)
    (cases tc:scheme-definition exp
      (tc:require (position module-names) module-names)
      (else (error "tc:provide->exported-ids passed a non require: " exp)))))

(deftype tc:defrep->pairs
  (-> (tc:scheme-definition) (list-of tc:scheme-abs-conc-pair)))
(define tc:defrep->pairs
  (lambda (exp)
    (cases tc:scheme-definition exp
      (tc:defrep (position pairs) pairs)
      (else (error "tc:defrep->exported-ids passed a non defrep" exp)))))

(deftype tc:require?
  (-> (tc:scheme-definition) boolean))
(define tc:require?
  (lambda (exp)
    (cases tc:scheme-definition exp
      (tc:require (position module-names) #t)
      (else #f))))

(deftype tc:defrep?
  (-> (tc:scheme-definition) boolean))
(define tc:defrep?
  (lambda (exp)
    (cases tc:scheme-definition exp
      (tc:defrep (position pairs) #t)
      (else #f))))

(define tc:deftype?
  (lambda (exp)
    (cases tc:scheme-definition exp
	   (tc:deftype (position variable type) #t)
	   (else #f))))

;;; formals
(define-datatype tc:scheme-formals tc:scheme-formals?
  (tc:normal-formals
   (position tc:position?)
   (variables (list-of symbol?)))
  (tc:unrestricted-formals
   (position tc:position?)
   (variable symbol?))
  (tc:some-normal-formals
   (position tc:position?)
   (before-dot (list-of symbol?)) (after-dot symbol?)))

(define tc:scheme-formals->names
  (lambda (sf)
    (cases tc:scheme-formals sf
	   (tc:normal-formals
	    (position variables) variables)
	   (tc:unrestricted-formals
	    (position variable)
	    (list variable))
	   (tc:some-normal-formals
	    (position before-dot after-dot)
	    (append before-dot (list after-dot))))))

(define tc:scheme-formals->position
  (lambda (sf)
    (cases tc:scheme-formals sf
	   (tc:normal-formals (position variables) position)
	   (tc:unrestricted-formals (position variable) position)
	   (tc:some-normal-formals (position before-dot after-dot) position))))
      

;;; bodies
(define-datatype tc:scheme-body-record tc:scheme-body?
  (tc:scheme-body
   (position tc:position?)
   (definitions (list-of tc:scheme-definition?))
   (sequence tc:scheme-exp?)))

(define tc:scheme-body-record->position
  (lambda (sb)
    (cases tc:scheme-body-record sb
	   (tc:scheme-body
	    (position definitions sequence)
	    position))))

;;; variant-decls
(define-datatype tc:variant-decl-record tc:variant-decl?
  (tc:variant-decl
   (position tc:position?)
   (variant-name symbol?) (fields (list-of tc:field-pred-binding?))))

(deftype tc:variant-decl->variant-name
  (-> (tc:variant-decl-record) symbol))
(define tc:variant-decl->variant-name
  (lambda (vdr)
    (cases tc:variant-decl-record vdr
      (tc:variant-decl (position variant-name fields) variant-name))))

(define tc:variant-decl->position
  (lambda (vdr)
    (cases tc:variant-decl-record vdr
      (tc:variant-decl (position variant-name fields) position))))

;;; field-pred bindings
(define-datatype tc:field-pred-binding-record tc:field-pred-binding?
  (tc:field-pred-binding
   (position tc:position?)
   (field-name symbol?) (predicate-exp tc:scheme-exp?)))

(deftype tc:field-pred-binding->field-name
  (-> (tc:field-pred-binding-record) symbol))
(define tc:field-pred-binding->field-name
  (lambda (fpb)
    (cases tc:field-pred-binding-record fpb
      (tc:field-pred-binding (position field-name predicate-exp)
	 field-name))))

(deftype tc:field-pred-binding->predicate-exp
  (-> (tc:field-pred-binding-record) tc:scheme-exp))
(define tc:field-pred-binding->predicate-exp
  (lambda (fpb)
    (cases tc:field-pred-binding-record fpb
      (tc:field-pred-binding (position field-name predicate-exp)
	 predicate-exp))))

(define tc:field-pred-binding->position
  (lambda (fpb)
    (cases tc:field-pred-binding-record fpb
      (tc:field-pred-binding (position field-name predicate-exp)
	 position))))

;;; syntax specs
(define-datatype tc:scheme-syntax-spec-record tc:scheme-syntax-spec?
  (tc:scheme-syntax-spec
   (position tc:position?)
   (keyword symbol?) (transformer tc:scheme-transformer-spec?)))

;;; transformer specs
(define-datatype tc:scheme-transformer-spec-record
  tc:scheme-transformer-spec?
  (tc:scheme-transformer-spec
   (position tc:position?)
   (identifiers (list-of symbol?))
   (syntax-rules (list-of tc:scheme-syntax-rule?))))

;;; syntax rules
(define-datatype tc:scheme-syntax-rule-record tc:scheme-syntax-rule?
  (tc:scheme-syntax-rule
   (position tc:position?)
   (pattern tc:scheme-syntax-pattern?) (template tc:scheme-syntax-template?)))

;;; syntax patterns
(define-datatype tc:scheme-syntax-pattern tc:scheme-syntax-pattern?
  (tc:syntax-pattern-identifier
   (position tc:position?)
   (identifier symbol?))
  (tc:syntax-pattern-normal-list
   (position tc:position?)
   (patterns (list-of tc:scheme-syntax-pattern?)))
  (tc:syntax-pattern-dotted-list
   (position tc:position?)
   (before-dot (list-of tc:scheme-syntax-pattern?))
   (after-dot tc:scheme-syntax-pattern?))
  (tc:syntax-pattern-ellipsis-list
   (position tc:position?)
   (patterns (list-of tc:scheme-syntax-pattern?))
   (repeated tc:scheme-syntax-pattern?))
  (tc:syntax-pattern-normal-vector
   (position tc:position?)
   (patterns (list-of tc:scheme-syntax-pattern?)))
  (tc:syntax-pattern-ellipsis-vector
   (position tc:position?)
   (patterns (list-of tc:scheme-syntax-pattern?))
   (repeated tc:scheme-syntax-pattern?))
  (tc:syntax-pattern-datum
   (position tc:position?)
   (datum datum?)))

;;; syntax templates
(define-datatype tc:scheme-syntax-template tc:scheme-syntax-template?
  (tc:syntax-template-identifier 
   (position tc:position?)
   (identifier symbol?))
  (tc:syntax-template-normal-list
   (position tc:position?)
   (elements (list-of tc:scheme-syntax-template-element?)))
  (tc:syntax-template-dotted-list
   (position tc:position?)
   (before-dot (list-of tc:scheme-syntax-template?))
   (after-dot tc:scheme-syntax-template-element?))
  (tc:syntax-template-normal-vector
   (position tc:position?)
   (elements (list-of tc:scheme-syntax-template-element?)))
  (tc:syntax-template-datum
   (position tc:position?)
   (datum datum?)))

;;; syntax template elements
(define-datatype tc:scheme-syntax-template-element
  tc:scheme-syntax-template-element?
  (tc:scheme-syntax-template-element-normal
   (position tc:position?)
   (element tc:scheme-syntax-template?))
  (tc:scheme-syntax-template-element-ellipsis
   (position tc:position?)
   (element tc:scheme-syntax-template?)))
  
;;; fundamental expressions
(define-datatype tc:scheme-exp tc:scheme-exp?
  (tc:varref 
   (position tc:position?)
   (variable symbol?))
  (tc:quotation
   (position tc:position?)
   (datum tc:scheme-datum?))
  (tc:self-evaluating
   (position tc:position?)
   (datum datum?))
  (tc:procedure-call
   (position tc:position?)
   (operator tc:scheme-exp?) (operands (list-of tc:scheme-exp?)))
  (tc:lambda-exp
   (position tc:position?)
   (formals tc:scheme-formals?) (body tc:scheme-body?))
  (tc:named-lambda-exp
   (position tc:position?)
   (variable symbol?) (formals tc:scheme-formals?) (body tc:scheme-body?))
  (tc:conditional-exp
   (position tc:position?)
   (test tc:scheme-exp?) (consequent tc:scheme-exp?)
   (alternate tc:scheme-exp?))
  (tc:conditional-stmt
   (position tc:position?)
   (test tc:scheme-exp?) (consequent tc:scheme-exp?))
  (tc:assignment
   (position tc:position?)
   (variable symbol?) (exp tc:scheme-exp?))
  (tc:macro-use
   (position tc:position?)
   (keyword symbol?) (datums (list-of tc:scheme-datum?)))
  (tc:let-syntax-exp
   (position tc:position?)
   (syntax-specs (list-of tc:scheme-syntax-spec?)) (body tc:scheme-body?))
  (tc:letrec-syntax-exp
   (position tc:position?)
   (syntax-specs (list-of tc:scheme-syntax-spec?)) (body tc:scheme-body?))
;; derived expressions
  (tc:cond-exp
   (position tc:position?)
   (clauses (list-of tc:scheme-cond-clause?)))
  (tc:cond-stmt
   (position tc:position?)
   (clauses (list-of tc:scheme-cond-clause?)))
  (tc:case-exp
   (position tc:position?)
   (exp tc:scheme-exp?) (clauses (list-of tc:scheme-case-clause?)))
  (tc:case-stmt
   (position tc:position?)
   (exp tc:scheme-exp?) (clauses (list-of tc:scheme-case-clause?)))
  (tc:and-exp
   (position tc:position?)
   (tests (list-of tc:scheme-exp?)))
  (tc:or-exp
   (position tc:position?)
   (tests (list-of tc:scheme-exp?)))
  (tc:let-exp
   (position tc:position?)
   (decls (list-of tc:scheme-binding-spec?)) (body tc:scheme-body?))
  (tc:named-let-exp
   (position tc:position?)
   (name symbol?)
   (decls (list-of tc:scheme-binding-spec?)) (body tc:scheme-body?))
  (tc:let*-exp
   (position tc:position?)
   (decls (list-of tc:scheme-binding-spec?)) (body tc:scheme-body?))
  (tc:letrec-exp
   (position tc:position?)
   (decls (list-of tc:scheme-binding-spec?)) (body tc:scheme-body?))
  ;; begin-exp is often used in place of <sequence> in the concrete syntax
  (tc:begin-exp
   (position tc:position?)
   (commands (list-of tc:scheme-exp?)) (exp tc:scheme-exp?))
  (tc:do-exp
   (position tc:position?)
   (iteration-specs (list-of tc:scheme-iteration-spec?))
   (test tc:scheme-exp?)
   (do-result tc:scheme-exp?)
   (commands (list-of tc:scheme-exp?)))
  (tc:do-stmt
   (position tc:position?)
   (iteration-specs (list-of tc:scheme-iteration-spec?))
   (test tc:scheme-exp?)
   (commands (list-of tc:scheme-exp?)))
  (tc:delay-exp
   (position tc:position?)
   (exp tc:scheme-exp?))
  (tc:quasiquotation
   (position tc:position?)
   (datum tc:scheme-datum?))
  ;; the following expressions are added for EOPL (second edition)
  (tc:cases-exp
   (position tc:position?)
   (type-name symbol?)
   (exp tc:scheme-exp?)
   (clauses (list-of tc:scheme-cases-clause?)))
  (tc:cases-stmt
   (position tc:position?)
   (type-name symbol?)
   (exp tc:scheme-exp?)
   (clauses (list-of tc:scheme-cases-clause?)))
  ;; the following expressions are added for the type checker
  (tc:has-type-exp
   (position tc:position?)
   (type tc:input-type-exp?) (exp tc:scheme-exp?))
  (tc:has-type-trusted-exp
   (position tc:position?)
   (type tc:input-type-exp?) (exp tc:scheme-exp?))
  (tc:test-type-exp
   (position tc:position?)
   (type-pred tc:scheme-exp?) (exp tc:scheme-exp?))
  (tc:arrow-exp
   (position tc:position?)
   (arg-exps (list-of tc:scheme-exp?)) (result-exp tc:scheme-exp?))
  (tc:varargs-arrow-exp
   (position tc:position?)
   (arg-exps (list-of tc:scheme-exp?))
   (repeated-exp tc:scheme-exp?)
   (result-exp tc:scheme-exp?))
  (tc:forall-exp
   (position tc:position?)
   (formals tc:scheme-formals?) (exp tc:scheme-exp?))
  ;; the following are only used internally by the type checker, not parsed
  (tc:define-datatype-exp
   (position tc:position?)
   (variants (list-of tc:variant-decl?)))
  (tc:constructor-for-variant-type-exp
   (position tc:position?)
   (argument-preds (list-of tc:scheme-exp?))
   (type-name symbol?))
  (tc:module-exp
   (position tc:position?)
   (identifier symbol?)
   (initial-required-module-name tc:scheme-module-name?)
   (provides (list-of symbol?))
   (requires-for-syntax (list-of tc:scheme-module-name?))   
   (requires (list-of tc:scheme-module-name?))
   (defreps (list-of tc:scheme-abs-conc-pair?))
   (deftypes (list-of tc:scheme-definition?))
   (definitions (list-of tc:scheme-definition?)))
  
  )

(deftype tc:scheme-exp->position (-> (tc:scheme-exp) tc:position))
(define tc:scheme-exp->position
  (lambda (exp)
    (cases tc:scheme-exp exp
      (tc:varref (position variable) position)
      (tc:quotation (position datum) position)
      (tc:self-evaluating (position datum) position)
      (tc:procedure-call (position operator operands) position)
      (tc:named-lambda-exp (position name formals body) position)
      (tc:lambda-exp (position formals body) position)
      (tc:conditional-exp (position test consequent alternate) position)
      (tc:conditional-stmt (position test consequent) position)
      (tc:assignment (position variable exp) position)
      (tc:macro-use (position keyword datums) position)
      (tc:let-syntax-exp (position syntax-specs body) position)
      (tc:letrec-syntax-exp (position syntax-specs body) position)
      (tc:cond-exp (position clauses) position)
      (tc:cond-stmt (position clauses) position)
      (tc:case-exp (position exp clauses) position)
      (tc:case-stmt (position exp clauses) position)
      (tc:and-exp (position tests) position)
      (tc:or-exp (position tests) position)
      (tc:let-exp (position decls body) position)
      (tc:named-let-exp (position variable decls body) position)
      (tc:let*-exp (position decls body) position)
      (tc:letrec-exp (position decls body) position)
      (tc:begin-exp (position decls body) position)
      (tc:do-exp (position iteration-specs test do-result commands) position)
      (tc:do-stmt (position iteration-specs test commands) position)
      (tc:delay-exp (position exp) position)
      (tc:quasiquotation (position datum) position)
      (tc:cases-exp (position type-name exp clauses) position)
      (tc:cases-stmt (position type-name exp clauses) position)
      (tc:has-type-exp (position type exp) position)
      (tc:has-type-trusted-exp (position type exp) position)
      (tc:test-type-exp (position type-pred exp) position)
      (tc:arrow-exp (position arg-exps result-exp) position)
      (tc:varargs-arrow-exp (position arg-exps repeaded-exp result-exp)
			    position)
      (tc:forall-exp (position formals exp) position)
      (tc:define-datatype-exp (position variants) position)
      (tc:constructor-for-variant-type-exp
       (position argument-preds type-name) position)
      (tc:module-exp
       (position identifier initial-required-module-name
		 provides requirs-for-syntax requires defreps deftype definitions)
       position)
      )))

(deftype tc:varref? (-> (tc:scheme-exp) boolean))
(define tc:varref?
  (lambda (exp)
    (cases tc:scheme-exp exp
      (tc:varref (position variable) #t)
      (else #f))))

(define tc:varref->variable
  (lambda (exp)
    (cases tc:scheme-exp exp
      (tc:varref (position variable) variable)
      (else (error "tc:varref->variable passed a non varref" exp)))))

(define tc:self-evaluating->datum
  (lambda (exp)
    (cases tc:scheme-exp exp
      (tc:self-evaluating (position datum) datum)
      (else (error "tc:self-evaluating->datum passed non self-evaluating"
		   exp)))))

(define tc:procedure-call->operator
  (lambda (exp)
    (cases tc:scheme-exp exp
      (tc:procedure-call (position operator operands) operator)
      (else (error "tc:procedure-call->operator passed a non procedure-call"
		   exp)))))

(define tc:procedure-call->operands
  (lambda (exp)
    (cases tc:scheme-exp exp
      (tc:procedure-call (position operator operands) operands)
      (else (error "tc:procedure-call->operands passed a non procedure-call"
		   exp)))))

;;; cond clause
(define-datatype tc:scheme-cond-clause tc:scheme-cond-clause?
  (tc:normal-cond-clause 
   (position tc:position?)
   (test tc:scheme-exp?) (sequence tc:scheme-exp?))
  (tc:test-only-cond-clause
   (position tc:position?)
   (test tc:scheme-exp?))
  (tc:test=>cond-clause
   (position tc:position?)
   (test tc:scheme-exp?) (recipient tc:scheme-exp?))
  (tc:else-cond-clause
   (position tc:position?)
   (sequence tc:scheme-exp?)))

(define tc:normal-cond-clause?
  (lambda (cc)
    (cases tc:scheme-cond-clause cc
	   (tc:normal-cond-clause (position test sequence) #t)
	   (else #f))))

(define tc:test=>cond-clause?
  (lambda (cc)
    (cases tc:scheme-cond-clause cc
	   (tc:test=>cond-clause (position test recipient) #t)
	   (else #f))))

(deftype tc:scheme-cond-clause->exp (-> (tc:scheme-cond-clause) tc:scheme-exp?))
(define tc:scheme-cond-clause->exp
  (lambda (cc)
    ;; ENSURES: returns the subexpression which gives the result of the clause
    (cases tc:scheme-cond-clause cc
      (tc:normal-cond-clause (position test sequence) sequence)
      (tc:test-only-cond-clause (position test) test)
      (tc:test=>cond-clause (position test recipient)
        (let* ((tpos (tc:scheme-exp->position test))
	       (rpos (tc:scheme-exp->position recipient))
	       (pos (tc:make-tree-position
                     (tc:position->start-line tpos)
		     (tc:position->end-line rpos)
		     (tc:position->filename rpos))))
	  (tc:procedure-call pos recipient (list test))))
      (tc:else-cond-clause (position sequence) sequence))))

(define tc:scheme-cond-clause->position
  (lambda (cc)
    ;; ENSURES: returns the subexpression which gives the result of the clause
    (cases tc:scheme-cond-clause cc
      (tc:normal-cond-clause (position test sequence) position)
      (tc:test-only-cond-clause (position test) position)
      (tc:test=>cond-clause (position test recipient) position)			    
      (tc:else-cond-clause (position sequence) position))))


;;; case clause
(define-datatype tc:scheme-case-clause tc:scheme-case-clause?
  (tc:normal-case-clause 
   (position tc:position?)
   (datums (list-of tc:scheme-datum?)) (sequence tc:scheme-exp?))
  (tc:else-case-clause
   (position tc:position?)
   (sequence tc:scheme-exp?)))

(deftype tc:scheme-case-clause->exp (-> (tc:scheme-case-clause) tc:scheme-exp?))
(define tc:scheme-case-clause->exp
  (lambda (cc)
    ;; ENSURES: returns the subexpression which gives the result of the clause
    (cases tc:scheme-case-clause cc
      (tc:normal-case-clause (position datums sequence) sequence)
      (tc:else-case-clause (position sequence) sequence))))

(define tc:scheme-case-clause->position
  (lambda (cc)
    (cases tc:scheme-case-clause cc
      (tc:normal-case-clause (position datums sequence) position)
      (tc:else-case-clause (position sequence) position))))



;;; binding spec (for let, let*, letrec)
(define-datatype tc:scheme-binding-spec-record tc:scheme-binding-spec?
  (tc:scheme-binding-spec
   (position tc:position?)
   (variable symbol?) (exp tc:scheme-exp?)))

(deftype tc:scheme-binding-spec->position
  (-> (tc:scheme-binding-spec-record) tc:position))
(define tc:scheme-binding-spec->position
  (lambda (bs)
    (cases tc:scheme-binding-spec-record bs
      (tc:scheme-binding-spec (position variable exp)
        position))))

(deftype tc:scheme-binding-spec->variable
  (-> (tc:scheme-binding-spec-record) symbol))
(define tc:scheme-binding-spec->variable
  (lambda (bs)
    (cases tc:scheme-binding-spec-record bs
      (tc:scheme-binding-spec (position variable exp)
        variable))))

(deftype tc:scheme-binding-spec->exp
  (-> (tc:scheme-binding-spec-record) tc:scheme-exp))
(define tc:scheme-binding-spec->exp
  (lambda (bs)
    (cases tc:scheme-binding-spec-record bs
      (tc:scheme-binding-spec (position variable exp)
        exp))))

(deftype tc:scheme-binding-spec->varref
  (-> (tc:scheme-binding-spec-record) tc:scheme-exp))
(define tc:scheme-binding-spec->varref
  (lambda (bs)
    (cases tc:scheme-binding-spec-record bs
      (tc:scheme-binding-spec (position variable exp)
        (tc:varref position variable)))))

;;; iteration spec (for do)
(define-datatype tc:scheme-iteration-spec tc:scheme-iteration-spec?
  (tc:stepped-iteration-spec
   (position tc:position?)
   (variable symbol?) (init tc:scheme-exp?) (step tc:scheme-exp?))
  (tc:init-only-iteration-spec
   (position tc:position?)
   (variable symbol?) (init tc:scheme-exp?)))

(deftype tc:scheme-iteration-spec->variable
  (-> (tc:scheme-iteration-spec) symbol?))
(define tc:scheme-iteration-spec->variable
  (lambda (sip)
    (cases tc:scheme-iteration-spec sip
      (tc:stepped-iteration-spec (position variable init step) variable)
      (tc:init-only-iteration-spec (position variable init) variable)
      )))

(deftype tc:scheme-iteration-spec->init
  (-> (tc:scheme-iteration-spec) tc:scheme-exp?))
(define tc:scheme-iteration-spec->init
  (lambda (sip)
    (cases tc:scheme-iteration-spec sip
      (tc:stepped-iteration-spec (position variable init step) init)
      (tc:init-only-iteration-spec (position variable init) init)
      )))

(define tc:scheme-iteration-spec->position
  (lambda (sip)
    (cases tc:scheme-iteration-spec sip
      (tc:stepped-iteration-spec (position variable init step) position)
      (tc:init-only-iteration-spec (position variable init) position)
      )))


;;; cases clause
(define-datatype tc:scheme-cases-clause tc:scheme-cases-clause?
  (tc:normal-cases-clause 
   (position tc:position?)
   (variant-name symbol?) (fields (list-of symbol?)) (sequence tc:scheme-exp?))
  (tc:else-cases-clause
   (position tc:position?)
   (sequence tc:scheme-exp?)))

(deftype tc:scheme-cases-clause->exp
  (-> (tc:scheme-cases-clause) tc:scheme-exp))
(define tc:scheme-cases-clause->exp
  (lambda (scc)
    ;; ENSURES: returns a subexpression that determines the result
    (cases tc:scheme-cases-clause scc
      (tc:normal-cases-clause (position variant-name fields sequence)
				     sequence)
      (tc:else-cases-clause (position sequence) sequence))))

(define tc:scheme-cases-clause->position
  (lambda (scc)
    (cases tc:scheme-cases-clause scc
      (tc:normal-cases-clause (position variant-name fields sequence)
			      position)
      (tc:else-cases-clause (position sequence) position))))


(deftype tc:else-cases-clause?
  (-> (tc:scheme-cases-clause) symbol))
(define tc:else-cases-clause?
  (lambda (scc)
    (cases tc:scheme-cases-clause scc
	   (tc:else-cases-clause (position sequence) #t)
	   (else #f))))
    

;;; scheme-datum (for quotations and quasiquotations)
(define-datatype tc:scheme-datum tc:scheme-datum?
  (tc:simple-datum 
   (position tc:position?)
   (datum datum?))
  (tc:normal-list-datum
   (position tc:position?)
   (elements (list-of tc:scheme-datum?)))
  (tc:improper-list-datum
   (position tc:position?)
   (before-dot (list-of tc:scheme-datum?))
   (after-dot tc:scheme-datum?))
  (tc:quoted-datum
   (position tc:position?)
   (datum tc:scheme-datum?))
  (tc:quasiquoted-datum
   (position tc:position?)
   (datum tc:scheme-datum?))
  (tc:unquoted-datum
   (position tc:position?)
   (datum tc:scheme-datum?))
  (tc:unquoted-exp
   (position tc:position?)
   (exp tc:scheme-exp?))
  (tc:unquote-splicing-datum
   (position tc:position?)
   (datum tc:scheme-datum?))
  (tc:unquote-splicing-exp
   (position tc:position?)
   (exp tc:scheme-exp?))
  (tc:vector-datum
   (position tc:position?)
   (elements (list-of tc:scheme-datum?))))

(define tc:scheme-datum->position
  (lambda (d)
    (cases tc:scheme-datum d
	   (tc:simple-datum (position datum) position)
	   (tc:normal-list-datum (position elements) position)
	   (tc:improper-list-datum (position before-dot after-dot) position)
	   (tc:quoted-datum (position datum) position)
	   (tc:quasiquoted-datum (position datum) position)
	   (tc:unquoted-datum (position datum) position)
	   (tc:unquoted-exp (position exp) position)
	   (tc:unquote-splicing-datum (position datum) position)
	   (tc:unquote-splicing-exp (position exp) position)
	   (tc:vector-datum (position elements) position))))
    
(deftype tc:simple-datum? (-> (tc:scheme-datum) boolean))
(define tc:simple-datum?
  (lambda (x)
    (cases tc:scheme-datum x
      (tc:simple-datum (position datum) #t)
      (else #f))))

(deftype tc:simple-datum->datum (-> (tc:scheme-datum) datum))
(define tc:simple-datum->datum
  (lambda (x)
    (cases tc:scheme-datum x
      (tc:simple-datum (position datum) datum)
      (else (error "tc:simple-datum->datum passed a non simple-datum" x)))))

;;; input type expressions added for the type checker
(define-datatype tc:input-type-exp tc:input-type-exp?
  (tc:basic-input-type-exp
   (position tc:position?)
   (identifier symbol?))
  (tc:function-input-type-exp
   (position tc:position?)
   (arg-types (list-of tc:input-type-exp?)) (result-type tc:input-type-exp?))
  (tc:varargs-function-input-type-exp
   (position tc:position?)
   (arg-types (list-of tc:input-type-exp?))
   (repeated-type tc:input-type-exp?)
   (result-type tc:input-type-exp?))
  (tc:intersection-input-type-exp
   (position tc:position?)
   (conjoined-types (list-of tc:input-type-exp?)))
  (tc:applied-input-type-exp
   (position tc:position?)
   (operator symbol?) (operand-types (list-of tc:input-type-exp?)))
  (tc:polymorphic-input-type-exp
   (position tc:position?)
   (type-formals (list-of symbol?)) (monomorphic-type tc:input-type-exp?))
  (tc:type-predicate-for-input-type-exp
   (position tc:position?)
   (type tc:input-type-exp?))
  (tc:variant-record-input-type-exp
   (position tc:position?)
   (variants (list-of tc:variant-type-decl?)))
  ;; The following was added for convenience in translation from
  ;; internal type expressions back (temporarily) to input type expressions
  (tc:translated-input-type-exp
   (position tc:position?)
   (type-expr datum?)))

(deftype tc:input-type-exp->position (-> (tc:input-type-exp) tc:position))
(define tc:input-type-exp->position
  (lambda (ityp)
    (cases tc:input-type-exp ityp
      (tc:basic-input-type-exp (position identifier) position)
      (tc:function-input-type-exp (position arg-types result-type) position)
      (tc:varargs-function-input-type-exp
       (position arg-types repeated-type result-type) position)
      (tc:intersection-input-type-exp (position conjoined-types) position)
      (tc:applied-input-type-exp (position operator operand-types) position)
      (tc:polymorphic-input-type-exp
       (position type-formals monomorphic-type) position)
      (tc:type-predicate-for-input-type-exp (position type) position)
      (tc:variant-record-input-type-exp (position variants) position)
      (tc:translated-input-type-exp (position type-expr) position))))

;;; variant-type-decls
(define-datatype tc:variant-type-decl-record tc:variant-type-decl?
  (tc:variant-type-decl
   (position tc:position?)
   (variant-name symbol?) (fields (list-of tc:field-type-binding?))))

(deftype tc:variant-type-decl->position
  (-> (tc:variant-type-decl-record) tc:position))
(define tc:variant-type-decl-record->position
  (lambda (vtd)
    (cases tc:variant-type-decl-record vtd
      (tc:variant-type-decl (position variant-name fields) position))))

(deftype tc:variant-type-decl->variant-name
  (-> (tc:variant-type-decl-record) symbol))
(define tc:variant-type-decl->variant-name
  (lambda (vtd)
    (cases tc:variant-type-decl-record vtd
      (tc:variant-type-decl (position variant-name fields) variant-name))))

;;; field-type bindings
(define-datatype tc:field-type-binding-record tc:field-type-binding?
  (tc:field-type-binding
   (position tc:position?)
   (field-name symbol?) (type tc:input-type-exp?)))

(deftype tc:field-type-binding->field-name
  (-> (tc:field-type-binding-record) symbol))
(define tc:field-type-binding->field-name
  (lambda (ftb)
    (cases tc:field-type-binding-record ftb
      (tc:field-type-binding (position field-name type) field-name))))

(deftype tc:field-type-binding->type
  (-> (tc:field-type-binding-record) tc:input-type-exp))
(define tc:field-type-binding->type
  (lambda (ftb)
    (cases tc:field-type-binding-record ftb
      (tc:field-type-binding (position field-name type) type))))


;;--------generic extraction helper----------------
(define tc:scheme-syntax->position
  (lambda (syn)
    (let ((syntax->position
	   (cond
	    ((tc:scheme-form? syn) tc:scheme-form->position)
	    ((tc:scheme-exp? syn) tc:scheme-exp->position)
	    ((tc:scheme-datum? syn) tc:scheme-datum->position)
	    ((tc:scheme-formals? syn) tc:scheme-formals->position)
	    ((tc:scheme-body? syn) tc:scheme-body-record->position)
	    ((tc:scheme-definition? syn) tc:scheme-definition->position)
	    ((tc:variant-decl? syn) tc:variant-decl->position)
	    ((tc:field-pred-binding? syn) tc:field-pred-binding->position)
	    ((tc:scheme-cond-clause? syn) tc:scheme-cond-clause->position)
	    ((tc:scheme-case-clause? syn) tc:scheme-case-clause->position)
	    ((tc:scheme-binding-spec? syn) tc:scheme-binding-spec->position)
	    ((tc:scheme-cases-clause? syn) tc:scheme-cases-clause->position)
	    ((tc:scheme-iteration-spec? syn) tc:scheme-iteration-spec->position)
	    ((tc:scheme-program? syn) tc:scheme-program->position)
	    ((tc:input-type-exp? syn) tc:input-type-exp->position)
            ((tc:scheme-module-name? syn) tc:scheme-module-name->position)
            ((tc:scheme-abs-conc-pair? syn) tc:scheme-abs-conc-pair->position)
	    ;; more cases go here	    
	    )))
      (syntax->position syn))))

) ;; end module
