;; $Id: tc-scheme-annotate.tst,v 1.14 2005/05/12 20:06:05 dorn Exp $

;;; Load all the parsing stuff
(require (lib "tc-tracing.scm" "lib342")
	 (lib "tc-types.scm" "lib342")
	 (lib "tc-env.scm" "lib342")
         (lib "tc-scheme-parser.scm" "lib342")
	 (lib "tc-type-translate.scm" "lib342")
	 (lib "tc-scheme-annotate.scm" "lib342"))

;; no excess output, please
(type-tracing-off!)

(define tc:infer-exp-type-main
  (lambda (e)
    (tc:annotation->type (tc:scheme-top-level-annotate (tc:env-empty) e ))))

(define tc:infer-top-level-form-type
  (lambda (pi f c)
    (let ((res (tc:annotation->type (tc:scheme-top-level-annotate pi f))))
      (if (tc:declaration-type-expr? res)
	  (c (map tc:binding->type-expr (tc:env-vals (tc:declaration-type-expr->env res))))
	  res))))

(run-regression-tests-equal?
 `(
    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "1")))
      ==> number )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "#t")))
      ==> boolean )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "\"a string\"")))
      ==> string )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "#\\c")))
      ==> char )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "+")))
      ==> (-> (number ...) number) )


    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(lambda (x) x)")))
      ==> (forall (T) (-> (T) T)) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(+ 2 3)")))
      ==> number )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(lambda x x)")))
      ==> (forall (T) (-> (T ...) (list-of T))) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(lambda x (displayln 2) x)")))
      ==> (forall (T) (-> (T ...) (list-of T))) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(lambda (x . y) (displayln x) y)")))
      ==> (forall (S T) (-> (S T ...) (list-of T))) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "'x")))
      ==> symbol )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "'()")))
      ==> (forall (T) (list-of T)) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "'(x y z)")))
      ==> (list-of symbol) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "'(x y z a b c d)")))
      ==> (list-of symbol) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "'(x y 3 a b c d)")))
      ==> (list-of datum) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "'#(x y a b c d)")))
      ==> (vector-of symbol) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let ((x 1)) x)")))
      ==> number )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let ((x 1)) `(,x))")))
      ==> (list-of number) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let ((x 1)) `(,x 3 #t . ,x))")))
      ==> (pair-of number (pair-of number (pair-of boolean number))) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let ((x 1) (y (+ 1 2))) `(,x . ,y))")))
      ==> (pair-of number number) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let* ((x 1) (y (+ x 2))) `(,x . ,y))")))
      ==> (pair-of number number) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let* ((x 1) (y (+ x 2)) (z (zero? y))) `#(,x ,z 4 ,y))")))
      ==> (vector-of datum) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let* ((x 1) (y (+ x 2)) (z (zero? y))) `(,z . #(,x 4 ,y)))")))
      ==> (pair-of boolean (vector-of number)) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(letrec ((odd? (lambda (n) (if (zero? n) #f (even? (- n 1)))))
                   (even? (lambda (n) (if (zero? n) #t (odd? (- n 1))))))
            (odd? 5))")))
      ==> boolean )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let ((x 1) (y 2))
            (letrec ((addx (lambda (n) (+ n x)))
                     (z (+ y x)))
            (addx (+ x y z))))")))
      ==> number )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(has-type number 1)")))
      ==> number )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(has-type (forall (T) (-> (T) T)) (lambda (x) x))")))
      ==> (forall (T) (-> (T) T)) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(has-type-trusted datum 1)")))
      ==> datum )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(test-type number? 1)")))
      ==> number )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(test-type (list-of symbol?) (read))")))
      ==> (list-of symbol) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(letrec ((oh-no (lambda () (oh-no))))
             (delay `(,oh-no . ,(oh-no))))")))
      ==> (forall (S T) (promise (pair-of (-> () S) T))) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(and (or (zero? 3) #f (or) (and))
               (or (not (list? '()))))")))
      ==> boolean )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "()")))
      ==> (forall (T) (list-of T)) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let ((ls '(a b c)))
            (cond
             ((null? ls) (displayln 3) '())
             ('(a))
             ((not (null? ls)) => (lambda (b) '(c)))
             (else (displayln 'got-here) (cdr ls))))")))
      ==> (list-of symbol) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let ((ls '(a b c)))
            (cond
             ((null? ls) (displayln 3))
             ((not (null? ls)) => (lambda (b) (displayln 4)))
             ((eq? 'b (cadr ls)) (displayln 5))))")))
      ==> void )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(if (zero? (+ 7 -7)) (displayln 'math-works))")))
      ==> void )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(if (zero? (+ 7 -7))
              (begin
                (displayln 'math-works)
                (displayln `(and so does begin))))")))
      ==> void )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let ((x 1)) (set! x 3))")))
      ==> void )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(let ((x 1))
            (define y 3)
            (set! y x))")))
      ==> void )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(case 3
           ((0) 'zero)
           ((1) (displayln 'ok) 'one)
           ((2 3) 'many)
           (else 'lots!))")))
      ==> symbol )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(case 3
           ((0) (displayln 'zero))
           ((1) (displayln 'ok))
           ((2 3) (displayln 'many)))")))
      ==> void )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(lambda (x)
            (define b #t)
            b)")))
      ==> (forall (T) (-> (T) boolean)) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	 "(lambda (x)
            (define (foo y) #t)
            (foo x))")))
      ==> (forall (T) (-> (T) boolean)) )

;    ( (tc:type-unparse
;       (tc:infer-exp-type-main
;	(tc:character-string-exp-parser
;	 "(lambda (x)
;            (define (odd? n) (if (zero? n) #f (even? (- n 1))))
;            (define (even? n) (if (zero? n) #t (odd? (- n 1))))
;            (odd? x))")))
;      ==> (-> (number) boolean) )

;    ( (tc:type-unparse
;       (tc:infer-exp-type-main
;	(tc:character-string-exp-parser
;	 "(lambda (x)
;            (deftype odd? (-> (number) boolean))
;            (deftype even? (-> (number) boolean))
;            (define (odd? n) (if (zero? n) #f (even? (- n 1))))
;            (define (even? n) (if (zero? n) #t (odd? (- n 1))))
;            (odd? x))")))
;      ==> (-> (number) boolean) )

;    ( (tc:type-unparse
;       (tc:infer-exp-type-main
;	(tc:character-string-exp-parser
;	 "(lambda (x)
;            (deftype odd? (-> (number) boolean))
;            (define (odd? n) (if (zero? n) #f (even? (- n 1))))
;            (define (even? n) (if (zero? n) #t (odd? (- n 1))))
;            (deftype even? (-> (number) boolean))
;            (odd? x))")))
;      ==> (-> (number) boolean) )

;    ( (tc:type-unparse
;       (tc:infer-exp-type-main
;	(tc:character-string-exp-parser
;	 "(lambda (x)
;            (begin (deftype odd? (-> (number) boolean))
;                   (define (odd? n) (if (zero? n) #f (even? (- n 1)))))
;            (begin (define (even? n) (if (zero? n) #t (odd? (- n 1))))
;                   (deftype even? (-> (number) boolean)))
;            (odd? x))")))
;      ==> (-> (number) boolean) )


    ( (tc:type-unparse
       (tc:infer-top-level-form-type
	(tc:env-empty)
	(tc:character-string-form-parser "+")
	(lambda (typs) (car typs))))
      ==> (-> (number ...) number) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	  "(let loop ((numbers '(3 -2 1 6 -5))
		  (nonneg '())
		  (neg '()))
	    (cond
	       ((null? numbers) (list nonneg neg))
	       ((>= (car numbers) 0)
		(loop (cdr numbers)
		      (cons (car numbers) nonneg)
		      neg))
	       (else
		(loop (cdr numbers)
		      nonneg
		      (cons (car numbers) neg)))))")))
      ==> (list-of (list-of number)) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	  "(-> (number?) boolean?)")))
      ==> (type-predicate-for (-> (number) boolean)) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	  "(-> (symbol? number?) boolean?)")))
      ==> (type-predicate-for (-> (symbol number) boolean)) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	  "(-> (symbol? (list-of number?) (vector-of number?)) boolean?)")))
      ==> (type-predicate-for
	   (-> (symbol (list-of number) (vector-of number)) boolean)) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	  "(-> (symbol? (list-of number?) (-> (number? number?) boolean?))
               boolean?)")))
      ==> (type-predicate-for
	   (-> (symbol (list-of number)
		       (-> (number number) boolean)) boolean)) )

    ( (tc:type-unparse
       (tc:infer-exp-type-main
	(tc:character-string-exp-parser
	  "(-> (number? boolean? ...) boolean?)")))
      ==> (type-predicate-for (-> (number boolean ...) boolean)) )

    ( (tc:type-unparse
       (tc:infer-top-level-form-type
	(tc:env-empty)
	(tc:character-string-form-parser
	 "(type-predicate-for (list-of number?))")
	(lambda (typs) (car typs))))
      ==> (type-predicate-for (type-predicate-for (list-of number))) )

     ( (tc:type-unparse
	(tc:infer-top-level-form-type
	 (tc:env-empty)
	 (tc:character-string-form-parser
	  "(define double-list-of
           (forall (T?)
              (list-of (list-of T?))))")
	 (lambda (typs) (car typs))))
      ==> (forall (T) (-> ((type-predicate-for T))
			  (type-predicate-for (list-of (list-of T))))) )

    ( (tc:type-unparse
       (tc:infer-top-level-form-type
	(tc:env-empty)
	(tc:character-string-form-parser
	 "(define bad-vector-or-list-of
           (lambda (T?)
            (lambda (val)
              (or ((vector-of T?) val)
                  ((list-of T?) val)))))")
	(lambda (typs) (car typs))))
      ==> (forall (S T) (-> ((type-predicate-for S)) (-> (T) boolean))) )
    
    ( (tc:type-unparse
       (tc:infer-top-level-form-type
	(tc:env-empty) 
	(tc:character-string-form-parser
	 "(define union-of
           (lambda (S? T?)
            (lambda (val)
              (or (T? val)
                  (S? val)))))")
	(lambda (typs) (car typs))))
      ==> (forall (T) (-> ((-> (T) boolean) (-> (T) boolean))
			  (-> (T) boolean))) )

;;;    ( (tc:type-unparse
;;;       (tc:infer-exp-type-main
;;;	(tc:character-string-exp-parser
;;;	 "(do ((i 1 (+ i 1))
;;;               (j 2 (+ i j)))
;;;             ((> i j 200) (+ i j))
;;;             (displayln i j))")))
;;;      ==> number )

;;;    ( (tc:type-unparse
;;;       (tc:infer-exp-type-main
;;;	(tc:character-string-exp-parser
;;;	 "(do ((i 1 (+ i 1))
;;;               (j 2 (+ i j)))
;;;             ((> i j 200))
;;;             (displayln i j))")))
;;;      ==> void )


    
    ))
