;; $Id: tc-subst.tst,v 1.7 2005/03/19 06:17:29 dorn Exp $

(require (lib "tc-subst.scm" "lib342")
         (lib "maybe.scm" "lib342"))

(define-datatype simple-type simple-type?
  (basic (name symbol?))
  (lvar (logical-var tc:logicalvar?))
  (applied (operator symbol?) (operands (list-of simple-type?)))
  (uniont (types (list-of simple-type?)))
  (simple-error))

(define (to-term lv) (lvar lv))
(define (get-var t)
  (cases simple-type t
	 (lvar (logical-var) (make-something logical-var))
	 (else (make-nothing))))
(define (subterms t)
    (cases simple-type t
	   (applied (operator operands) operands)
	   (uniont (types) types)
	   (else '())))
(define (kind t)
  (cases simple-type t
	 (basic (name) 'basic)
	 (applied (operator operands) 'applied)
	 (uniont (types) 'uniont)
	 (lvar (logical-var) 'lvar)
	 (simple-error () 'error)))
  
(define (same-kind t1 t2) (equal? (kind t1) (kind t2)))
(define nullSubst (tc:make-null-subst lvar))
(define (bind lv t)
  (tc:make-subst lv t lvar))


(define (app s)
  (lambda (t)
    (cases simple-type t
      (basic (name) t)
      (lvar (logical-var) (tc:subst-apply s logical-var))
      (applied (operator operands) (applied operator (map (app s) operands)))
      (uniont (types) (uniont (map (app s) types)))
      (simple-error () t))))

(define is-error
  (lambda (t)
    (cases simple-type t
	   (simple-error () #t)
	   (else #f))))

(define gen-error-mismatch
  (lambda (x t1 t2)
    (simple-error)))

(define make-error
  (lambda (syn expected atrees)
    (simple-error)))

(define simple-type-md
  (tc:unifiable-md-dict
   to-term
   get-var
   subterms
   same-kind
   nullSubst
   bind
   app
   ;; No Subtyping so we just ignore this
   (lambda (t) '())
   (lambda (t) '())
   (lambda (t) '())
   (lambda (x y) (cons x y))
   (lambda (t) #f)
   (lambda (t1 t2) '())
   (lambda (t1 t2) '())
   (lambda (x) #f)
   (lambda (x) #f)
   ;; Error considerations
   is-error
   gen-error-mismatch
   make-error
   ;; No declaration types, so ignore this
   (lambda (x) #f)
   (lambda (x) #f))
  )

(define mgu (tc:mgu simple-type-md))
(define list-unify (tc:list-unify simple-type-md))
(define vars-in (tc:vars-in simple-type-md))

(run-regression-tests-equal?
 `(
    ( (subterms (uniont (list
		       (lvar (tc:symbol->logicalvar 'z))
		       (applied 'x
				(list (lvar (tc:symbol->logicalvar 'y))
				      (basic 'number)))
		       (lvar (tc:symbol->logicalvar 't))
		       (basic 'number))))
      = (list (lvar (tc:symbol->logicalvar 'z))
	      (applied 'x
		       (list (lvar (tc:symbol->logicalvar 'y))
			     (basic 'number)))
	      (lvar (tc:symbol->logicalvar 't))
	      (basic 'number)) )
    ( (tc:logicalvar? (tc:symbol->logicalvar 'x)) ==> #t )
    ( (vars-in (uniont (list (lvar (tc:symbol->logicalvar 'x)))))
      ==> ((x . 0)) )
    ( (let ((f (mgu (uniont (list (lvar (tc:symbol->logicalvar 'x))))
		    (uniont (list (basic 'number))))))
	(and (not (null? f)) ((car f) (tc:symbol->logicalvar 'x))))
      = (basic 'number) )
    ( (let ((f (mgu (uniont (list (lvar (tc:symbol->logicalvar 'x))))
		    (uniont (list (basic 'number) (basic 'number))))))
	(null? f))
      = #t )

    ( (vars-in (uniont (list
		       (lvar (tc:symbol->logicalvar 't))
		       (basic 'number))))
      ==> ((t . 0)) )
    
    ( (vars-in (lvar (tc:symbol->logicalvar 'y)))
      ==> ((y . 0)) )

    ( (vars-in (uniont (list
		       (applied 'x
				(list (lvar (tc:symbol->logicalvar 'y))
				      (basic 'number)))
		       (lvar (tc:symbol->logicalvar 't))
		       (basic 'number))))
      ==> ((y . 0) (t . 0)) )

    
    ( (vars-in (uniont (list
		       (lvar (tc:symbol->logicalvar 'z))
		       (applied 'x
				(list (lvar (tc:symbol->logicalvar 'y))
				      (basic 'number)))
		       (lvar (tc:symbol->logicalvar 't))
		       (basic 'number))))
      ==> ((z . 0) (y . 0) (t . 0)) )

    
    )
 )

