;;; $Id: tc-type-check-and-eval-loop.scm,v 1.24 2005/05/12 11:36:54 dorn Exp $
;;;---------------------------------------------------------------------
;;;

;;; We assume that you already have loaded a file like
;;;    drscheme/type-check-ignore-types-at-runtime.scm
;;; if you are using DrScheme, or, if you are using Chez Scheme, then both
;;;    translate-module-syntax.scm  and
;;;    tc-ignore-types-at-runtime-chez.scm .
;;; These defines macros that ignore deftypes and other parts of the
;;; type notation during the evaluation process.
;;; See the shell scripts in the bin directory for details.
;;; These files can also be used if you aren't
;;; using the type checker, but want to run files that have deftypes
;;; in them.

;;; Load all the parsing stuff
(require (lib "tc-scheme-annotate.scm" "lib342")
	 (lib "tc-scheme-parser.scm" "lib342")
	 (lib "tc-scheme-unparser.scm" "lib342")
	 (lib "tc-tracing.scm" "lib342")
         (lib "tc-types.scm" "lib342")
	 (lib "tc-type-translate.scm" "lib342")
	 (lib "tc-error-output.scm" "lib342"))

;;; Set the following to #t to evaluate even when there is a type error.
(deftype *eval-even-if-error* boolean)
(define *eval-even-if-error* #f)

(deftype *tc:type-check-and-eval-loop-users-type-env*
  (tc:environment-of symbol binding))
(define *tc:type-check-and-eval-loop-users-type-env* (tc:env-empty))

(deftype tc:colon string)
(define tc:colon " : ")

(deftype tc:type-check-and-eval-loop-helper
  (-> ((tc:env-of symbol binding)) poof))
(define tc:type-check-and-eval-loop-helper 
  (lambda ()
    ;; EFFECT: prompt with prompt-str, read forms, type check, eval, write, ...
    (display "typed> ")
    (force-output)
    (tc:reset-parsed-file-cache!)
    (let* ((char-seq (tc:stream->char-seq (tc:make-input-stream)))
	   (form (if (tc:char-seq-null? char-seq)
		     (exit)
		     (tc:char-seq-form-parser char-seq))))
      (begin
	(if (> (type-tracing-level) 3)
	    (displayln "FORM: " form))
	(if (> (type-tracing-level) 2)
	    (displayln "UNPARSED: " (tc:unparse-scheme-form form)))
	(let ((result (tc:annotation->type (tc:scheme-top-level-annotate
				   *tc:type-check-and-eval-loop-users-type-env*
				   form))))
	  (cond
	   ((tc:error-type-expr? result)
	    (tc:output-error-message result)
	    (displayln "Skipping evaluation because of the type errors..."))
	   ;; PUT STUFF IN HERE FOR OTHER CASES
	   ;; one exp only below ...

	   ((tc:declaration-type-expr? result)
	    ;; Extend the user's environment
	    (let ((delta-env (tc:declaration-type-expr->env result)))
	      ;; Display all new bindings
	      (tc:env-for-each
	       (lambda (key value)
		 (display key)
		 (display tc:colon)		 
		 (tc:displayln-output-type (tc:type-unparse
					    (tc:binding->type-expr value))))
	       delta-env)
		 
	      (set! *tc:type-check-and-eval-loop-users-type-env*
		    (tc:env-join
		     delta-env
		     *tc:type-check-and-eval-loop-users-type-env*))
	      
	    
	      ;; Then evaluate the form
	      (tc:just-eval-form #f form)))
							 	   
	   (else
	    (write (tc:just-eval-form (tc:error-type-expr? result) form))
	    (display tc:colon)
;;	    (tc:displayln-output-type  result)
	   (tc:displayln-output-type (tc:type-unparse result))))
	  (force-output))
	(tc:type-check-and-eval-loop-helper)))))
 
(deftype tc:just-eval-form
  (-> (boolean tc:scheme-form) datum))
(define tc:just-eval-form
  (lambda (err-typ? form)
    ;; EFFECT: Evaluate form
    (if err-typ?
	(displayln "WARNING: this may have a type error..."))
    (eval (tc:unparse-scheme-form form) (interaction-environment))))

;;; ------------------------------------

;;; The following exits the type checker
(deftype type-check-exit (-> () poof))
(define type-check-exit
  (lambda ()
    ;; (set! error tc:orig-error)
    (newline)
    (displayln "Use `(type-check-and-eval-loop)' to restart the type checker.")
    ;; abort is SCM dependent; it this goes back to the top-level.
    (abort)))

;;; The following resets the type checker's environment
(deftype type-check-reset-env! (-> () void))
(define type-check-reset-env!
  (lambda ()
    (set! *tc:type-check-and-eval-loop-users-type-env* (tc:env-empty))
    (displayln "All user-defined names have been purged from the type checker")
    ))

;;; The following provides a list the public type checker procedures
(deftype type-help (-> () void))
(define type-help
  (lambda ()
    (displayln "The Polymorphic Type Checker for Scheme provides ")
    (displayln "the following public procedures:")
    (displayln "type-check-exit           leave the typed interpreter loop")
    (displayln "type-check-and-eval-loop  a typed Scheme interpreter loop")
    (displayln "type-check-reset-env!     forget all names defined so far")
    (displayln "type-help                 displays this information")
    (displayln "type-tracing-set!         set tracing detail level to given value")
    (displayln "type-tracing-off!         set tracing detail level to 0")
    (displayln "type-tracing-level        returns the numeric level of tracing detail")
    ))

;;; adjusting error and starting the type-check-and-eval-loop below...

(define tc:orig-error error)

;;; Calling the new error procedure (see below)
;;; prints an error message and then calls *tc:type-check-and-eval-start*.
(deftype *tc:type-check-and-eval-start* (-> (datum) poof))
(define *tc:type-check-and-eval-start*
  (lambda (ignored)
    (tc:ignore ignored)
    (displayln "exiting")
    (exit)))

;;; Change what error does, so that we start the type-check-and-eval loop
;;; after issuing an error message, instead of dropping out to Scheme.
(define tc:modified-error
  (lambda args
    (let ((cep
	   ;; if needed, change to (current-output-port)
	   (current-error-port)))
      (display "ERROR: " cep)
      (if (not (null? args))
	  (begin (display (car args) cep)
		 (for-each (lambda (x)
			     (display " " cep)
			     (write x cep))
			   (cdr args))
		 (newline cep)))
      (*tc:type-check-and-eval-start* 'ignored))))

(deftype type-check-and-eval-loop  (-> () poof))
(define type-check-and-eval-loop 
  (lambda ()
    (tc:type-check-and-eval-loop-helper)))
