;; $Id: tc-type-helpers.tst,v 1.49 2005/05/12 19:06:53 dorn Exp $

(require (lib "tc-util.scm" "lib342")
	 (lib "tc-types.scm" "lib342")
         (lib "tc-type-helpers.scm" "lib342"))

(define tc:empty-type-env (tc:env-empty))

(define-datatype tc:lambda-calc tc:lambda-calc?
  (tc:le-varref 
   (variable symbol?))
  (tc:le-self-evaluating
   (datum datum?))
  (tc:le-procedure-call
   (operator tc:lambda-calc?) (operands (list-of tc:lambda-calc?)))  
  (tc:le-lambda-exp
   (formals (list-of symbol?)) (body tc:lambda-calc?)) 
  (tc:le-if-exp
   (test tc:lambda-calc?)
   (true-exp tc:lambda-calc?)
   (false-exp tc:lambda-calc?)) 
  )

(define tc:infer-simple-datum-type
  (lambda (datum)
    ;; REQUIRES: datum is a scheme datum
    (cond
      ((number? datum) (tc:basic-type-expr 'number))
     ((char? datum) (tc:basic-type-expr 'char))
     ((string? datum) (tc:basic-type-expr 'string))
     ((boolean? datum) (tc:basic-type-expr 'boolean))
     ((symbol? datum) (tc:basic-type-expr 'symbol))
     ((null? (has-type datum datum))   ;;;cut out lists for now
      (tc:applied-type-expr
       (tc:basic-type-expr 'list-of)
       (list (tc:new-variable-type-expr))))
     (else (tc:fail "tc:infer-simple-datum-type: unknown constant type for: "
		    datum)))))

(define axiom (tc:axiom-helper (tc:type-expr-as-unifiable)))
(define rule (tc:rule-helper
	      (tc:type-expr-as-unifiable)
	      (lambda (pi e) (tc:lambda-calc-annotate-lower pi e))))
(define rule-if (tc:rule-if-helper
			(tc:type-expr-as-unifiable)
			(lambda (pi e) (tc:lambda-calc-annotate-lower pi e))))
(define rule-seq-helper (tc:rule-seq-helper
			 (tc:type-expr-as-unifiable)
			 (lambda (pi e) (tc:lambda-calc-annotate-lower pi e))))
(define rule-or-helper (tc:rule-or-helper
			(tc:type-expr-as-unifiable)
			(lambda (pi e) (tc:lambda-calc-annotate-lower pi e))))
(define-syntax rule-seq
  (syntax-rules (list)
    ((rule-seq (list h1 ...) conc)
     (rule-seq-helper (list (delay h1) ...) conc))))
(define-syntax rule-or
  (syntax-rules ()
    ((rule-or t1 ...)
     (rule-or-helper (delay t1) ...))))

(define mgu (tc:mgu (tc:type-expr-as-unifiable)))

(define tc:lambda-calc-annotate
  (lambda (pi e)
    (tc:set-current-subst-list! (list (tc:type-expr-null-subst)))
    (tc:lambda-calc-annotate-lower pi e)))

(deftype tc:lambda-calc-annotate-lower
  (forall (syn)
    (-> (tc:type-env syn)
	(pair-of (atree-of syn)
		 (list-of (subst-of tc:type-expr))))))
(define tc:lambda-calc-annotate-lower
  (lambda (pi e)
    (cases tc:lambda-calc e
	   (tc:le-varref
	    (variable)
	    (let ((lv (tc:new-logicalvar)))
	      (rule-if
	       '() ;; no hypotheses here
	       (lambda (ts)
		 (tc:env-bound? variable pi))
	       (lambda (ts)
		 (tc:type-expr-bind lv (tc:env-value variable pi)))
	       ;; ------------------------------
	       (:- pi (: e (tc:variable-type-expr lv))))))
	   
	   (tc:le-self-evaluating
	    (datum)
	    (let ((t (tc:infer-simple-datum-type datum)))
	      (axiom (:- pi (: e t)))))
	   (tc:le-procedure-call  
	    (operator operands)
	    (let ((rt (tc:new-variable-type-expr))
		  (operand-ts (tc:listof-new-variable-type-expr (length operands)))
		  (oper-type (tc:new-variable-type-expr)))
	      (rule (append
		     (list
		      (:- pi (: operator oper-type)))
		     (map
		      (lambda (operand t)
			(:- pi (: operand t)))
		      operands operand-ts)
		     (list
		      (<: oper-type (tc:function-type-expr operand-ts rt))))
		    ;;----------------------------
		    (:- pi (: e rt)))))

	   (tc:le-lambda-exp
	    (formals body)
	    (let ((t2 (tc:new-variable-type-expr))
		  (formal-ts
		   (tc:listof-new-variable-type-expr (length formals))))
	      ;; should be no duplicates in formals
	      (rule-if (list
			(:- (tc:extend-mult-env pi formals formal-ts) ;; Add bindings for all formals
			    (: body t2)))
		       (lambda (ts)
			 (tc:no-duplicates? formals))
		       (lambda (ts)
			 (tc:type-expr-null-subst))
		       ;; ---------------------
		       (:- pi (: e (tc:function-type-expr
				    formal-ts t2))))))
	   (tc:le-if-exp
	    (test true-exp false-exp)
	    (let* ((t1 (tc:new-variable-type-expr))
		   (t2 (tc:new-variable-type-expr))
		   (supertype-var (tc:new-logicalvar))
		   (subtype-var (tc:new-logicalvar))
		   (supertype (tc:variable-type-expr supertype-var)))
	      (rule (list
		     (:- pi (: test (tc:basic-type-expr 'boolean))) 
		     (:- pi (: true-exp t1))
		     (:- pi (: false-exp t2))
		     (<:> t1 t2 subtype-var supertype-var))
		    ;; -------------------------
		    (:- pi (: e supertype)))))
	    )))
			     

(run-regression-tests-equal?
 `(
   ((tc:lambda-calc-annotate tc:empty-type-env (tc:le-self-evaluating 5))
     = (: (tc:le-self-evaluating 5) (tc:node (tc:basic-type-expr 'number) '())))
   
   ( (tc:lambda-calc-annotate
	   (tc:extend-env tc:empty-type-env 'x (tc:basic-type-expr 'symbol))
	   (tc:le-varref 'x))
     = (: (tc:le-varref 'x) (tc:node (tc:basic-type-expr 'symbol) '())))
      
   ( (tc:lambda-calc-annotate
	   (tc:extend-mult-env
	    tc:empty-type-env '(x f)
	    (list (tc:basic-type-expr 'symbol)
		  (tc:function-type-expr (list (tc:basic-type-expr 'symbol))
					 (tc:basic-type-expr 'number))))
			 
	   (tc:le-procedure-call
	    (tc:le-varref 'f)
	    (list (tc:le-varref 'x))))
     =  (: (tc:le-procedure-call
	    (tc:le-varref 'f)
	    (list (tc:le-varref 'x)))
	   (tc:node (tc:basic-type-expr 'number)
		    (list (tc:node (tc:function-type-expr
				    (list (tc:basic-type-expr 'symbol))
				    (tc:basic-type-expr 'number)) '())
		       (tc:node (tc:basic-type-expr 'symbol) '())))))

   ;; Test MGU for this next one first
   ( (let ((f (car (mgu (tc:function-type-expr
			 (list (tc:basic-type-expr 'number))
			 (tc:basic-type-expr 'number))
			(tc:function-type-expr
			 (list (tc:variable-type-expr (tc:symbol->logicalvar 't1)))
			 (tc:variable-type-expr (tc:symbol->logicalvar 't2)))))))
       (cons (tc:subst-apply f (tc:symbol->logicalvar 't1))
	     (tc:subst-apply f (tc:symbol->logicalvar 't2))))
     = (cons (tc:basic-type-expr 'number) (tc:basic-type-expr 'number)))


   ( (let ((f (car (mgu (tc:variable-type-expr (tc:symbol->logicalvar 't1))
			(tc:variable-type-expr (tc:symbol->logicalvar 't3))))))
       (cons (tc:subst-apply f (tc:symbol->logicalvar 't1))
	     (tc:subst-apply f (tc:symbol->logicalvar 't3))))
	    
     = (cons
	(tc:variable-type-expr (tc:symbol->logicalvar 't3))
	(tc:variable-type-expr (tc:symbol->logicalvar 't3))))
   
   
   ( (tc:annotation->type  
      (tc:lambda-calc-annotate
       (tc:extend-env tc:empty-type-env 'add1
		      (tc:function-type-expr (list (tc:basic-type-expr 'number))
					     (tc:basic-type-expr 'number)))
       (tc:le-lambda-exp (list 'x)
			 (tc:le-procedure-call (tc:le-varref 'add1)
					       (list (tc:le-varref 'x))))))
     = (tc:function-type-expr
	(list (tc:basic-type-expr 'number))
	(tc:basic-type-expr 'number)))

   ;; Test contravariant rule on procedure call
   ( (tc:error-type-expr?
      (tc:annotation->type
       (tc:lambda-calc-annotate
	(tc:extend-mult-env tc:empty-type-env '(f x)
			    (list (tc:function-type-expr (list (tc:basic-type-expr 'datum))
							 (tc:basic-type-expr 'number))
				  (tc:basic-type-expr 'number)))
	(tc:le-procedure-call (tc:le-varref 'f) (list (tc:le-varref 'x))))))
     = #f)
			    
   ( (tc:error-type-expr?
      (tc:annotation->type
       (tc:lambda-calc-annotate
	(tc:extend-mult-env tc:empty-type-env '(f x)
			    (list (tc:function-type-expr (list (tc:basic-type-expr 'number))
							 (tc:basic-type-expr 'number))
				  (tc:basic-type-expr 'datum)))
	(tc:le-procedure-call (tc:le-varref 'f) (list (tc:le-varref 'x))))))
     = #t)

   ( (tc:error-type-expr?
      (tc:annotation->type
       (tc:lambda-calc-annotate
	(tc:extend-env tc:empty-type-env 'add
		       (tc:function-type-expr (list (tc:basic-type-expr 'number)
						    (tc:basic-type-expr 'number))
					      (tc:basic-type-expr 'number)))
	(tc:le-lambda-exp (list 'x 'x)
			  (tc:le-procedure-call (tc:le-varref 'add)
						(list (tc:le-varref 'x)
						      (tc:le-varref 'x)))))))
     = #t)

   ( (tc:annotation->type
      (tc:lambda-calc-annotate
       (tc:extend-mult-env tc:empty-type-env
			   '(numnum strnum)
			   (list
			    (tc:function-type-expr (list (tc:basic-type-expr 'number)) (tc:basic-type-expr 'number))
			    (tc:function-type-expr (list (tc:basic-type-expr 'symbol)) (tc:basic-type-expr 'number))))
       (tc:le-lambda-exp (list 'x)
			 (tc:le-procedure-call (tc:le-varref 'numnum)
					       (list
						(tc:le-procedure-call
						 (tc:le-varref 'strnum)
						 (list (tc:le-varref 'x))))))))
     = (tc:function-type-expr
	(list (tc:basic-type-expr 'symbol))
	(tc:basic-type-expr 'number)))
   ( (tc:error-type-expr?
      (tc:annotation->type (tc:lambda-calc-annotate
			    tc:empty-type-env
			    (tc:le-varref 'bad_sym))))
     = #t)

   ( (tc:annotation->type
      (tc:lambda-calc-annotate
       (tc:extend-mult-env tc:empty-type-env
			   '(e e1 e2)
			   (list
			    (tc:basic-type-expr 'boolean)
			    (tc:basic-type-expr 'number)
			    (tc:basic-type-expr 'number)))
       (tc:le-if-exp (tc:le-varref 'e)
		     (tc:le-varref 'e1)
		     (tc:le-varref 'e2))))
     = (tc:basic-type-expr 'number))

   ( (tc:error-type-expr?
      (tc:annotation->type
       (tc:lambda-calc-annotate
	(tc:extend-mult-env tc:empty-type-env
			    '(e e1 e2)
			    (list
			     (tc:basic-type-expr 'number)
			     (tc:basic-type-expr 'number)
			     (tc:basic-type-expr 'number)))
	(tc:le-if-exp (tc:le-varref 'e)
		      (tc:le-varref 'e1)
		      (tc:le-varref 'e2)))))
     = #t)

   ( (tc:error-type-expr?
      (tc:annotation->type
       (tc:lambda-calc-annotate
	(tc:extend-mult-env tc:empty-type-env
			    '(e e1 e2)
			    (list
			     (tc:basic-type-expr 'boolean)
			     (tc:basic-type-expr 'number)
			     (tc:basic-type-expr 'symbol)))
	(tc:le-if-exp (tc:le-varref 'e)
		      (tc:le-varref 'e1)
		      (tc:le-varref 'e2)))))
     = #t)

   ( (tc:error-type-expr?
      (tc:annotation->type
       (tc:lambda-calc-annotate
	tc:empty-type-env
	(tc:le-lambda-exp
	 (list 'x)
	 (tc:le-procedure-call
	  (tc:le-varref 'x)
	  (list (tc:le-varref 'x)))))))

     = #t)
		
				   
      
      

   ( (tc:annotation->type
      (tc:lambda-calc-annotate
       (tc:extend-mult-env tc:empty-type-env
			   '(e e1 e2)
			   (list
			    (tc:basic-type-expr 'boolean)
			    (tc:basic-type-expr 'number)
			    (tc:basic-type-expr 'datum)))
       (tc:le-if-exp (tc:le-varref 'e)
		     (tc:le-varref 'e1)
		     (tc:le-varref 'e2))))
     = (tc:basic-type-expr 'datum))
     
   ( (tc:annotation->type
      (tc:lambda-calc-annotate
       (tc:extend-mult-env tc:empty-type-env
			   '(e e1 e2)
			   (list
			    (tc:basic-type-expr 'boolean)
			    (tc:basic-type-expr 'datum)
			    (tc:basic-type-expr 'symbol)))
       (tc:le-if-exp (tc:le-varref 'e)
		     (tc:le-varref 'e1)
		     (tc:le-varref 'e2))))
     = (tc:basic-type-expr 'datum))

   ( (tc:annotation->type
      (tc:lambda-calc-annotate
       (tc:extend-mult-env tc:empty-type-env
			   '(e e1 e2)
			   (list
			    (tc:basic-type-expr 'boolean)
			    (tc:basic-type-expr 'poof)
			    (tc:basic-type-expr 'symbol)))
       (tc:le-if-exp (tc:le-varref 'e)
		     (tc:le-varref 'e1)
		     (tc:le-varref 'e2))))
     = (tc:basic-type-expr 'symbol))      

   ( (tc:annotation->type
      (tc:lambda-calc-annotate
       (tc:extend-mult-env tc:empty-type-env
			   '(e e1 e2)
			   (list
			    (tc:basic-type-expr 'boolean)
			    (tc:basic-type-expr 'symbol)
			    (tc:basic-type-expr 'poof)))
       (tc:le-if-exp (tc:le-varref 'e)
		     (tc:le-varref 'e1)
		     (tc:le-varref 'e2))))
     = (tc:basic-type-expr 'symbol))    

   ( (tc:annotation->type
      (tc:lambda-calc-annotate
       (tc:extend-mult-env tc:empty-type-env
			   '(plus x y)
			   (list
			    (tc:function-type-expr
			     (list (tc:basic-type-expr 'number)
				   (tc:basic-type-expr 'number))
			     (tc:basic-type-expr 'number))
			    (tc:basic-type-expr 'number)
			    (tc:basic-type-expr 'number)))
       (tc:le-procedure-call
	(tc:le-varref 'plus)
	(list (tc:le-varref 'x)
	      (tc:le-varref 'y)))))
     = (tc:basic-type-expr 'number))

   ( (tc:annotation->type
      (tc:lambda-calc-annotate
       (tc:extend-mult-env tc:empty-type-env
			   '(plus x y)
			   (list
			    (tc:intersection-type-expr
			     (list
			      (tc:function-type-expr
			       (list (tc:basic-type-expr 'number)
				     (tc:basic-type-expr 'number))
			       (tc:basic-type-expr 'number))
			      (tc:function-type-expr
			       (list (tc:basic-type-expr 'symbol)
				     (tc:basic-type-expr 'symbol))
			       (tc:basic-type-expr 'number))))
			    (tc:basic-type-expr 'number)
			    (tc:basic-type-expr 'number)))
       (tc:le-procedure-call
	(tc:le-varref 'plus)
	(list (tc:le-varref 'x)
	      (tc:le-varref 'y)))))
     = (tc:basic-type-expr 'number))
   ))

(displayln "END OF TESTS")

 


