(require (lib "tc-types.scm" "lib342"))

(define mgu (tc:mgu (tc:type-expr-as-unifiable)))
(define list-unify (tc:list-unify (tc:type-expr-as-unifiable)))
(define vars-in (tc:vars-in (tc:type-expr-as-unifiable)))

(define make-type-list
  (lambda (size)
    (make-type-list-helper size '())))

(define make-type-list-helper
  (lambda (size accum)
    (if (= size 0)
	accum
	(make-type-list-helper (- size 1) (cons (tc:basic-type-expr 'number) accum)))))

(run-regression-tests-equal?
 `(

   ;; Test vars-in
   ((vars-in (tc:intersection-type-expr (list
					 (tc:basic-type-expr 'x) (tc:variant-record-type-expr '())
					 (tc:variable-type-expr (tc:symbol->logicalvar 't)))))
    = (list (tc:symbol->logicalvar 't)) )

   ((vars-in (tc:basic-type-expr 'x))
    ==> ())

   ((vars-in (tc:function-type-expr (list
				     (tc:basic-type-expr 'number)
				     (tc:variable-type-expr (tc:symbol->logicalvar 'a))
				     (tc:intersection-type-expr (list
								 (tc:variable-type-expr (tc:symbol->logicalvar 'b))
								 (tc:basic-type-expr 'symbol))))
				    (tc:variable-type-expr (tc:symbol->logicalvar 'c))))
    = (list (tc:symbol->logicalvar 'a)
	    (tc:symbol->logicalvar 'b)
	    (tc:symbol->logicalvar 'c)))
	     
				    
   ;; Test subterms
   
   ((tc:type-expr-subterms (tc:intersection-type-expr (list
						       (tc:basic-type-expr 'x) (tc:variant-record-type-expr '()))))
    =
    (list (tc:basic-type-expr 'x) (tc:variant-record-type-expr '())))

   ((tc:type-expr-subterms (tc:function-type-expr (list
						  (tc:basic-type-expr 'number)
						  (tc:variable-type-expr (tc:symbol->logicalvar 'a))
						  (tc:intersection-type-expr (list
									      (tc:variable-type-expr (tc:symbol->logicalvar 'b))
									      (tc:basic-type-expr 'symbol))))
						 (tc:variable-type-expr (tc:symbol->logicalvar 'c))))
    = (list
       (tc:basic-type-expr 'number)
       (tc:variable-type-expr (tc:symbol->logicalvar 'a))
       (tc:intersection-type-expr (list
				   (tc:variable-type-expr (tc:symbol->logicalvar 'b))
				   (tc:basic-type-expr 'symbol)))
       (tc:variable-type-expr (tc:symbol->logicalvar 'c))))

   ((tc:type-expr-subterms (tc:variant-record-type-expr	(list
							 (tc:field-type-binding-type-expr 'field1 (tc:basic-type-expr 'number))
							 (tc:field-type-binding-type-expr 'field2 (tc:variable-type-expr (tc:symbol->logicalvar 'a))))))
    = (list
       (tc:field-type-binding-type-expr 'field1 (tc:basic-type-expr 'number))
       (tc:field-type-binding-type-expr 'field2 (tc:variable-type-expr (tc:symbol->logicalvar 'a)))))

   ((tc:type-expr-subterms (tc:field-type-binding-type-expr 'field1 (tc:basic-type-expr 'number)))
    = (list (tc:basic-type-expr 'number)))

   ;; Test same-kind
   
   ((tc:type-expr-same-kind (tc:basic-type-expr 'number) (tc:basic-type-expr 'number))
    = #t)
   ((tc:type-expr-same-kind (tc:basic-type-expr 'number) (tc:basic-type-expr 'char))
    = #f)

   ((tc:type-expr-same-kind
     (tc:variant-record-type-expr
      (list
       (tc:field-type-binding-type-expr 'field1 (tc:basic-type-expr 'number))
       (tc:field-type-binding-type-expr 'field2 (tc:variable-type-expr (tc:symbol->logicalvar 'a)))))
     (tc:variant-record-type-expr (list
                                   (tc:field-type-binding-type-expr 'field1 (tc:basic-type-expr 'number))
                                   (tc:field-type-binding-type-expr 'field5 (tc:variable-type-expr (tc:symbol->logicalvar 'a))))))
    ==> #f)

   ((tc:type-expr-same-kind (tc:variant-record-type-expr
                             (list
                              (tc:field-type-binding-type-expr 'field1 (tc:basic-type-expr 'number))
                              (tc:field-type-binding-type-expr 'field2 (tc:variable-type-expr (tc:symbol->logicalvar 'a)))))
			    (tc:variant-record-type-expr
                             (list
                              (tc:field-type-binding-type-expr 'field1 (tc:basic-type-expr 'number))
                              (tc:field-type-binding-type-expr 'field2 (tc:variable-type-expr (tc:symbol->logicalvar 'a))))))
    ==> #t)

   ((tc:type-expr-same-kind (tc:variant-record-type-expr 
                             (list
                              (tc:field-type-binding-type-expr 'field1 (tc:basic-type-expr 'number))
                              (tc:field-type-binding-type-expr 'field2 (tc:variable-type-expr (tc:symbol->logicalvar 'a)))))
			    (tc:basic-type-expr 'number))
    ==> #f)

   ((tc:type-expr-same-kind (tc:declaration-type-expr (tc:extend-mult-env (tc:env-empty)
									  '(a b c d e f g)
									  (make-type-list 7)))
			    (tc:declaration-type-expr (tc:extend-mult-env (tc:env-empty)
									  '(a b c d e f)
									  (make-type-list 6))))
    ==> #f)

   
   ((tc:type-expr-same-kind (tc:declaration-type-expr (tc:extend-mult-env (tc:env-empty)
									  '(a b c d e f g)
									  (make-type-list 7)))
			    (tc:declaration-type-expr (tc:extend-mult-env (tc:env-empty)
									  '(a b c d e f g )
									  (make-type-list 7))))
    ==> #t)

   ((tc:type-expr-same-kind (tc:declaration-type-expr (tc:extend-mult-env (tc:env-empty)
									  '(a b c d e f g)
									  (make-type-list 7)))
			    (tc:declaration-type-expr (tc:extend-mult-env (tc:env-empty)
									  '(a b c d e f g)
									  (cons (tc:basic-type-expr 'symbol) (cdr (make-type-list 7))))))
    ==> #t)
									   
   
   ;; Test mgu
   
   ((tc:subst-apply (car (mgu (tc:basic-type-expr 'number) (tc:variable-type-expr (tc:symbol->logicalvar 't)))) (tc:symbol->logicalvar 't))
    = (tc:basic-type-expr 'number))
   
   ((let ((f (mgu (tc:intersection-type-expr (list (tc:variable-type-expr (tc:symbol->logicalvar 't))))
		  (tc:intersection-type-expr (list (tc:basic-type-expr 'number))))))
      (tc:subst-apply (car f) (tc:symbol->logicalvar 't)))
    = (tc:basic-type-expr 'number))

   ((null? (mgu (tc:intersection-type-expr (list (tc:basic-type-expr 'number)
						 (tc:variable-type-expr (tc:symbol->logicalvar 'x))))
		(tc:basic-type-expr 'symbol)))
    ==> #t)

   ((let ((f (mgu (tc:intersection-type-expr (list (tc:basic-type-expr 'number)
					      (tc:variable-type-expr (tc:symbol->logicalvar 'x))))
		  (tc:intersection-type-expr (list (tc:variable-type-expr (tc:symbol->logicalvar 'y))
					      (tc:variant-record-type-expr '()))))))
      (let ((subX (tc:subst-apply (car f) (tc:symbol->logicalvar 'x)))
	    (subY (tc:subst-apply (car f) (tc:symbol->logicalvar 'y))))
	(list subX subY)))
    = (list
       (tc:variant-record-type-expr '())
       (tc:basic-type-expr 'number)))
	
					     
 ))
