;;; $Id: type-predicates.scm,v 1.1 2002/04/09 22:25:26 leavens Exp $

;;; type predicates that the checker "understands"

(define datum? (lambda (x) #t))
(define ... datum?)

(define void? (lambda (x) #f))
(define poof? void?)

;;; Type predicate transformers that the type checker "understands"

(define pair-of
  (forall (S? T?)
    (if (not (procedure? S?))
	(eopl:error 'pair-of
		    "Expecting type predicates as left argument, not ~s"
		    S?)
	(if (not (procedure? T?))
	    (eopl:error 'pair-of
			"Expecting type predicates as right argument, not ~s"
			T?)
	    (lambda (obj)
	      (and (pair? obj)
		   (S? (car obj))
		   (T? (cdr obj))))))))

(define vector-of
  (letrec ((check-all-elems
	    (lambda (T? vec i)
	      (or (< i 0)
		  (and (T? (vector-ref vec i))
		       (check-all-elems T? vec (- i 1)))))))
    (forall (T?)
      (if (not (procedure? T?))
	  (eopl:error 'vector-of
		      "Expecting a type predicate as argument, not ~s"
		      T?)
	  (lambda (obj)
	    (and (vector? obj)
		 (check-all-elems T? obj (- (vector-length obj) 1))))))))

;;; The predicate transformer <- doesn't do anything with its arguments,
;;; but they are used by the -> syntax in the type checker.
;;; See type-check-ignore-types-at-runtime for the -> syntax.
(define <-
  (forall (result-pred . arg-preds)
    (if (not ((list-of procedure?) arg-preds))
	(eopl:error
	 '<-
	 "Expecting a list of type predicates as first argument, not ~s"
	 arg-preds)
	(if (not (procedure? result-pred))
	    (eopl:error
	     '<-
	     "Expecting a type predicate as second argument, not ~s"
	     result-pred)
	    (lambda (obj)
	      (procedure? obj))))))

;;; Type predicate for type predicates
(define type-predicate-for
  (forall (T?)
    (<- boolean? T?)))
  
;;; adapted from EOPL 2e, p. 45 (these are homogeneous lists)
(define list-of
  (forall (T?)
    (if (not (procedure? T?))
	(eopl:error 'list-of
		    "Expecting a type predicate as argument, not ~s"
		    T?)
	(lambda (val)
	  (or (null? val)
	      (and (pair? val)
		   (T? (car val))
		   ((list-of T?) (cdr val))))))))

(define all-of
  (forall (predicate-1 . predicates)
    (if (not (procedure? predicate-1))
	(eopl:error 'all-of
		    "Expecting a type predicate as argument, not ~s"
		    predicate-1)
	(if (not ((list-of procedure?) predicates))
	    (eopl:error 'all-of
			"Expecting type predicate arguments")
	    (lambda (val)
	      (letrec ((test-all
			(lambda (preds)
			  (or (null? preds)
			      (and ((car preds) val)
				   (test-all (cdr preds)))))))
		(test-all (cons predicate-1 predicates))))))))
