;; $Id: typecase-cloneupto-5-4-4.scm,v 1.1 2005/04/28 03:39:02 leavens Exp $
;; This is 5-4-4.scm: flat methods

(let ((time-stamp "Time-stamp: <2001-02-24 10:31:05 dfried>"))
  (eopl:printf "5-4-4.scm - flat methods ~a~%"
    (substring time-stamp 13 29)))

;;;;;;;;;;;;;;;; top level and tests ;;;;;;;;;;;;;;;;

(deftype run (-> (string) datum))
(define run
  (lambda (string)
    (eval-program (scan&parse string))))

(define functional-groups '(lang3-5 lang3-6 lang3-7))

(define oop-groups '(oop))

(deftype run-all (-> () void))
(define run-all
  (lambda ()
    (run-experiment run use-execution-outcome
      (append functional-groups oop-groups) all-tests)))

(define run-functional
  (lambda ()
    (run-experiment run use-execution-outcome
      functional-groups all-tests)))

(define run-oop
  (lambda ()
    (run-experiment run use-execution-outcome
      oop-groups all-tests)))

(deftype run-one (-> (symbol) void))
(define run-one
  (lambda (test-name)
    (run-test run test-name)))

;; needed for testing
(deftype equal-external-reps? (-> (datum datum) boolean))
(define equal-external-reps? equal?)

;;;;;;;;;;;;;;;; grammatical specification ;;;;;;;;;;;;;;;;

(define the-lexical-spec
  '((whitespace (whitespace) skip)
    (comment ("%" (arbno (not #\newline))) skip)
    (identifier
      (letter (arbno (or letter digit "_" "-" "?")))
      symbol)
    (number (digit (arbno digit)) number)))

(define the-grammar
  '((program ((arbno class-decl) expression) a-program)

    ;; new for oop
    (class-decl                         
      ("class" identifier 
        "extends" identifier                   
         (arbno "field" identifier)
         (arbno method-decl)
         )
      a-class-decl)

    ;; new for oop
    (method-decl
      ("method" identifier 
        "("  (separated-list identifier ",") ")" ; method formals
        expression 
        )
      a-method-decl)

    (expression (number) lit-exp)
    (expression ("quote" identifier) quote-exp)
    (expression (identifier) var-exp)
    (expression
      (primitive "(" (separated-list expression ",") ")")
      primapp-exp)
    (expression
      ("if" expression "then" expression "else" expression)
      if-exp)
    (expression
      ("let" (arbno  identifier "=" expression) "in" expression)
      let-exp)
    (expression
      ("proc" "(" (separated-list identifier ",") ")" expression)
      proc-exp)
    (expression
      ("(" expression (arbno expression) ")")
      app-exp)
    (expression
      ("begin" expression (arbno ";" expression) "end")
      begin-exp)
    (expression                         
      ("letrec"
        (arbno identifier "(" (separated-list identifier ",") ")"
          "=" expression)
        "in" expression)
      letrec-exp)
    (expression ("set" identifier "=" expression) varassign-exp)
    ;; new for oop
    (expression 
      ("new" identifier "(" (separated-list expression ",") ")")
      new-object-exp)
    (expression
      ("send" expression identifier
        "("  (separated-list expression ",") ")")
      method-app-exp)
    (expression                                
      ("super" identifier    "("  (separated-list expression ",") ")")
      super-call-exp)
    ;; typecase exercise
    (expression                                
      ("typecase" expression
         (arbno "when" identifier "(" identifier ")" expression)
	 "end")
      typecase-exp)
    ;; clone-upto exercise
    (expression                                
      ("clone" expression "upto" identifier)
      clone-upto-exp)

    (primitive ("+")     add-prim)
    (primitive ("-")     subtract-prim)
    (primitive ("*")     mult-prim)
    (primitive ("add1")  incr-prim)
    (primitive ("sub1")  decr-prim)
    (primitive ("zero?") zero-test-prim)
    (primitive ("equal?") equal-prim)
    (primitive ("greater?") greater-prim)
    (primitive ("less?") less-prim)
    (primitive ("list") list-prim)
    (primitive ("cons") cons-prim)
    (primitive ("nil")  nil-prim)
    (primitive ("car")  car-prim)
    (primitive ("cdr")  cdr-prim)
    (primitive ("null?") null?-prim)
    ))

(deftype show-the-datatypes (-> () (list-of datum)))
(define show-the-datatypes
  (lambda ()
      (sllgen:make-define-datatypes the-lexical-spec the-grammar)
      (sllgen:list-define-datatypes the-lexical-spec the-grammar)))

;; The following is generated by calling show-the-datatypes
;; (and then removing the outer lists, indenting, and renaming fields).
(define-datatype
  program
  program?
  (a-program
   (class-decls (list-of class-decl?))
   (exp expression?)))

(define-datatype
  class-decl
  class-decl?
  (a-class-decl
   (class-name symbol?)
   (super-name symbol?)
   (field-ids (list-of symbol?))
   (method-decls (list-of method-decl?))))

(define-datatype
  method-decl
  method-decl?
  (a-method-decl
   (method-name symbol?)
   (ids (list-of symbol?))
   (body expression?)))

(define-datatype
  expression
  expression?
  (lit-exp (datum number?))
  (quote-exp (sym symbol?))
  (var-exp (id symbol?))
  (primapp-exp
   (prim primitive?)
   (rands (list-of expression?)))
  (if-exp
   (test-exp expression?)
   (true-exp expression?)
   (false-exp expression?))
  (let-exp
   (ids (list-of symbol?))
   (rands (list-of expression?))
   (body expression?))
  (proc-exp
   (ids (list-of symbol?))
   (body expression?))
  (app-exp
   (rator expression?)
   (rands (list-of expression?)))
  (begin-exp
   (first expression?)
   (rest (list-of expression?)))
  (letrec-exp
   (proc-names (list-of symbol?))
   (idss (list-of (list-of symbol?)))
   (bodies (list-of expression?))
   (letrec-body expression?))
  (varassign-exp
   (id symbol?)
   (rhs-exp expression?))
  (new-object-exp
   (class-name symbol?)
   (rands (list-of expression?)))
  (method-app-exp
   (obj-exp expression?)
   (method-name symbol?)
   (rands (list-of expression?)))
  (super-call-exp
   (method-name symbol?)
   (rands (list-of expression?)))
  ;; typecase exercise
  (typecase-exp
   (tested-exp expression?)
   (class-names (list-of symbol?))
   (formals (list-of symbol?))
   (bodies (list-of expression?)))
  ;; clone-upto exercise
  (clone-upto-exp
   (obj-exp expression?)
   (class-name symbol?))
  )

(define-datatype
  primitive
  primitive?
  (add-prim)
  (subtract-prim)
  (mult-prim)
  (incr-prim)
  (decr-prim)
  (zero-test-prim)
  (equal-prim)
  (greater-prim)
  (less-prim)
  (list-prim)
  (cons-prim)
  (nil-prim)
  (car-prim)
  (cdr-prim)
  (null?-prim)
  )

(deftype scan&parse (-> (string) program))
(define scan&parse
  (sllgen:make-string-parser the-lexical-spec the-grammar))

(deftype just-scan (forall (TOK) (-> (string) (list-of TOK))))
(define just-scan
  (sllgen:make-string-scanner the-lexical-spec the-grammar))

(deftype read-eval-print (-> () poof))
(define read-eval-print 
  (sllgen:make-rep-loop  "--> "
    (lambda (pgm) (eval-program pgm))   ; wrapped to avoid load
                                        ; dependency 
    (sllgen:make-stream-parser 
      the-lexical-spec
      the-grammar)))

;;;;;;;;;;;;;;;; AST (declaration) helpers ;;;;;;;;;;;;;;;;

(deftype class-decl->class-name (-> (class-decl) symbol))
(define class-decl->class-name
  (lambda (c-decl)
    (cases class-decl c-decl
      (a-class-decl (class-name super-name field-ids m-decls)
        class-name))))

(deftype class-decl->super-name  (-> (class-decl) symbol))
(define class-decl->super-name
  (lambda (c-decl)
    (cases class-decl c-decl
      (a-class-decl (class-name super-name field-ids m-decls)
        super-name))))

(deftype class-decl->field-ids  (-> (class-decl) (list-of symbol)))
(define class-decl->field-ids
  (lambda (c-decl)
    (cases class-decl c-decl
      (a-class-decl (class-name super-name field-ids m-decls)
        field-ids))))

(deftype class-decl->method-decls (-> (class-decl) (list-of method-decl)))
(define class-decl->method-decls
  (lambda (c-decl)
    (cases class-decl c-decl
      (a-class-decl (class-name super-name field-ids m-decls)
        m-decls))))

(deftype method-decl->method-name (-> (method-decl) symbol))
(define method-decl->method-name
  (lambda (md)
    (cases method-decl md
      (a-method-decl (method-name ids body) method-name))))

(deftype method-decl->ids (-> (method-decl) (list-of symbol)))
(define method-decl->ids
  (lambda (md)
    (cases method-decl md
      (a-method-decl (method-name ids body) ids))))

(deftype method-decl->body (-> (method-decl) expression))
(define method-decl->body
  (lambda (md)
    (cases method-decl md
      (a-method-decl (method-name ids body) body))))

(deftype method-decls->method-names
  (-> ((list-of method-decl)) (list-of symbol)))
(define method-decls->method-names
  (lambda (mds)
    (map method-decl->method-name mds)))

;;; Domains for this interpreter:
;;;
;;; Expressed-Value = Number + ProcVal + Obj + List(Expressed-Value) + Symbol
;;; Denoted-Value = Ref(Expressed-Value)

;;; ADTs used

(require (lib "environment-5-4-2.scm" "lib342"))

;;;;;;;;;;;;;;;; ProcVal ;;;;;;;;;;;;;;;;

(deftype procval? (type-predicate-for procval))
(deftype closure (-> ((list-of symbol) expression environment) procval))
(deftype apply-procval
  (-> (procval (list-of Expressed-Value)) Expressed-Value))

(define-datatype procval procval?
  (closure 
    (ids (list-of symbol?)) 
    (body expression?)
    (env environment?)))

(define apply-procval
  (lambda (proc args)
    (cases procval proc
      (closure (ids body env)
        (eval-expression body (extend-env ids args env))))))

;;;;;;;;;;;;;;;; classes ;;;;;;;;;;;;;;;;

(define-datatype class class?
  (a-class
    (class-name symbol?)  
    (super-name symbol?) 
    (field-length integer?)  
    (field-ids (list-of symbol?))
    (methods method-environment?)))

;;;;;;;;;;;;;;;; objects ;;;;;;;;;;;;;;;;

;; an object is now just a single part, with a vector representing the
;; managed storage for the all the fields. 

(define-datatype object object? 
  (an-object
    (class-name symbol?)
    (fields vector?)))

(deftype new-object (-> (symbol) object))
(define new-object
  (lambda (class-name)
    (an-object
      class-name
      (make-vector
       (class-name->field-length class-name)
       (number->expressed 0)))))

;;;;;;;;;;;;;;;; methods ;;;;;;;;;;;;;;;;

(define-datatype method method?
  (a-method
    (method-decl method-decl?)
    (super-name symbol?)
    (field-ids (list-of symbol?))))

(deftype find-method-and-apply
  (-> (symbol symbol object (list-of Expressed-Value)) Expressed-Value))
(define find-method-and-apply
  (lambda (m-name host-name self args)
    (let ((declared-methods (class-name->methods host-name)))
	(if (has-method? m-name declared-methods)
          (let ((meth (lookup-method m-name declared-methods)))
	    (apply-method meth host-name self args))
	  (eopl:error 'find-method-and-apply
	    "No method for name ~s" m-name)))))

(deftype apply-method
  (-> (method symbol object (list-of Expressed-Value)) Expressed-Value))
(define apply-method
  (lambda (method host-name self args)
    (let ((ids (method->ids method))
          (body (method->body method))
          (super-name (method->super-name method))
          (field-ids (method->field-ids method))       
          (fields (object->fields self)))
      (eval-expression body
        (extend-env
          (cons '%super (cons 'self ids))
          (cons (symbol->expressed super-name)
		(cons (object->expressed self) args))
          (extend-env-refs field-ids fields (empty-env)))))))

;;;;;;;;;;;;;;;; method environments ;;;;;;;;;;;;;;;;

(define method-environment? (list-of method?)) 

(deftype has-method? (-> (symbol (list-of method)) boolean))
(define has-method?
  (lambda (m-name methods)
    (and (not (null? methods))
	 (or (eqv? m-name (method->method-name (car methods)))
	     (has-method? m-name (cdr methods))))))

(deftype lookup-method (-> (symbol (list-of method)) method))
(define lookup-method                   
  (lambda (m-name methods)
    (cond
      ((null? methods) 
       (eopl:error 'lookup-method "No method named ~s" m-name))
      ((eqv? m-name (method->method-name (car methods)))
       (car methods))
      (else (lookup-method m-name (cdr methods))))))

;;;;;;;;;;;;;;;; class environments ;;;;;;;;;;;;;;;;

;;; we'll just use the list of classes (not class decls)

(deftype the-class-env (list-of class))
(define the-class-env '())

(deftype initialize-class-env! (-> () void))
(define initialize-class-env!
  (lambda ()
    (set! the-class-env '())))

(deftype add-to-class-env! (-> (class) void))
(define add-to-class-env!
  (lambda (class)
    (set! the-class-env (cons class the-class-env))))

(deftype elaborate-class-decls! (-> ((list-of class-decl)) void))
(define elaborate-class-decls!
  (lambda (c-decls)
    (initialize-class-env!)
    (for-each elaborate-class-decl! c-decls)))

(deftype elaborate-class-decl! (-> (class-decl) void))
(define elaborate-class-decl!
  (lambda (c-decl)
    (let ((super-name (class-decl->super-name c-decl)))
      (let ((field-ids  (append
                          (class-name->field-ids super-name)
                          (class-decl->field-ids c-decl))))
        (add-to-class-env!
          (a-class
            (class-decl->class-name c-decl)
            super-name
            (length field-ids)
            field-ids
            (roll-up-method-decls
              c-decl super-name field-ids)))))))

;;; This may change for second printing.
(deftype roll-up-method-decls
  (-> (class-decl symbol (list-of symbol)) (list-of method)))
(define roll-up-method-decls
  (lambda (c-decl super-name field-ids)
    (merge-methods
      (class-name->methods super-name)
      (map
        (lambda (m-decl)
          (a-method m-decl super-name field-ids))
        (class-decl->method-decls c-decl)))))

(deftype merge-methods
  (-> ((list-of method) (list-of method)) (list-of method)))
(define merge-methods
  (lambda (super-methods methods)
    (if (null? super-methods)
	methods
	(let ((m-name (method->method-name (car super-methods))))
	  (if (has-method? m-name methods)
	      (let ((overriding-method (lookup-method m-name methods)))
		(cons overriding-method
		      (merge-methods
		       (cdr super-methods)
		       (remove-method overriding-method methods))))
	      (cons (car super-methods)
		    (merge-methods (cdr super-methods) methods)))))))

(deftype remove-method (-> (method (list-of method)) (list-of method)))
(define remove-method
  (lambda (method methods)
    (remv method methods)))

(deftype remv (forall (T) (-> (T (list-of T)) (list-of T))))
(define remv
  (lambda (x lst)
    (cond
      ((null? lst) '())
      ((eqv? (car lst) x) (remv x (cdr lst)))
      (else (cons (car lst) (remv x (cdr lst)))))))

(deftype lookup-class (-> (symbol) class))
(define lookup-class
  (lambda (name)
    (let loop ((env the-class-env))
      (cond
        ((null? env) (eopl:error 'lookup-class "Unknown class ~s" name))
        ((eqv? (class->class-name (car env)) name) (car env))
        (else (loop (cdr env)))))))

;;;;;;;;;;;;;;;; selectors of all sorts ;;;;;;;;;;;;;;;;

(deftype object->fields (-> (object) (vector-of Expressed-Value)))
(define object->fields
  (lambda (obj)
    (cases object obj
      (an-object (class-decl fields)
        fields))))

(deftype object->field-ids (-> (object) (list-of symbol)))
(define object->field-ids
  (lambda (object)
    (class->field-ids
      (object->class object))))

(deftype object->class-name (-> (object) symbol))
(define object->class-name
  (lambda (obj)
    (cases object obj
      (an-object (class-name fields)
        class-name))))

(deftype object->class (-> (object) class))
(define object->class
  (lambda (obj)
    (lookup-class (object->class-name obj))))

(deftype class-name->field-ids (-> (symbol) (list-of symbol)))
(define class-name->field-ids
  (lambda (class-name)
    (if (eqv? class-name 'object) '()
      (class->field-ids (lookup-class class-name)))))

(deftype class-name->methods (-> (symbol) (list-of method)))
(define class-name->methods
  (lambda (class-name)
    (if (eqv? class-name 'object) '()
      (class->methods (lookup-class class-name)))))

(deftype class-name->super-name (-> (symbol) symbol))
(define class-name->super-name
  (lambda (class-name)
    (class->super-name (lookup-class class-name))))

(deftype class-name->field-length (-> (symbol) number))
(define class-name->field-length
  (lambda (class-name)
    (if (eqv? class-name 'object)
        0
        (class->field-length (lookup-class class-name)))))

(deftype class->class-name (-> (class) symbol))
(define class->class-name
  (lambda (c-struct)
    (cases class c-struct
      (a-class (class-name super-name field-length field-ids methods)
        class-name))))

(deftype class->super-name (-> (class) symbol))
(define class->super-name
  (lambda (c-struct)
    (cases class c-struct
      (a-class (class-name super-name field-length field-ids methods)
        super-name))))

(deftype class->field-length (-> (class) number))
(define class->field-length
  (lambda (c-struct)
    (cases class c-struct
      (a-class (class-name super-name field-length field-ids methods)
        field-length))))

(deftype class->field-ids (-> (class) (list-of symbol)))
(define class->field-ids
  (lambda (c-struct)
    (cases class c-struct
      (a-class (class-name super-name field-length field-ids methods)
        field-ids))))

(deftype class->methods (-> (class) (list-of method)))
(define class->methods
  (lambda (c-struct)
    (cases class c-struct
      (a-class (class-name super-name field-length field-ids methods)
        methods))))

(deftype method->method-decl (-> (method) method-decl))
(define method->method-decl
  (lambda (meth)
    (cases method meth
      (a-method (meth-decl super-name field-ids) meth-decl))))

(deftype method->super-name (-> (method) symbol))
(define method->super-name
  (lambda (meth)
    (cases method meth
      (a-method (meth-decl super-name field-ids) super-name))))

(deftype method->field-ids (-> (method) (list-of symbol)))
(define method->field-ids
  (lambda (meth)
    (cases method meth
      (a-method (method-decl super-name field-ids) field-ids))))

(deftype method->method-name (-> (method) symbol))
(define method->method-name
  (lambda (method)
    (method-decl->method-name (method->method-decl method))))

(deftype method->body (-> (method) expression))
(define method->body
  (lambda (method)
    (method-decl->body (method->method-decl method))))

(deftype method->ids (-> (method) (list-of symbol)))
(define method->ids
  (lambda (method)
    (method-decl->ids (method->method-decl method))))


;;;;;;;;;;;;;;;; references ;;;;;;;;;;;;;;;;;;

(require (lib "reference-3-7.scm" "lib342"))

;;; Anonymous figure : page 80

(deftype true-value? (-> (Expressed-Value) boolean))
(define true-value?
  (lambda (x)
    (not (zero? (expressed->number x)))))

(deftype boolean->expressed (-> (boolean) Expressed-Value))
(define boolean->expressed
  (lambda (b)
    (number->expressed
     (if b 1 0))))

;;;;;;;;; Expressed-Value ;;;;;;;;;;;;;;;;;;;;;;;;;

;; upcasts
(deftype number->expressed (-> (number) Expressed-Value))
(deftype procval->expressed (-> (ProcVal) Expressed-Value))
(deftype object->expressed (-> (object) Expressed-Value))
(deftype list->expressed (-> ((list-of Expressed-Value)) Expressed-Value))
(deftype symbol->expressed (-> (symbol) Expressed-Value))

;; downcasts
(deftype expressed->number (-> (Expressed-Value) number))
(deftype expressed->procval (-> (Expressed-Value) ProcVal))
(deftype expressed->object (-> (Expressed-Value) object))
(deftype expressed->list (-> (Expressed-Value) (list-of Expressed-Value)))
(deftype expressed->symbol (-> (Expressed-Value) symbol))

;; tests
(deftype procval->expressed? (-> (Expressed-Value) boolean))

(define-datatype Expressed-Value expval?
  (number->expressed (number number?))
  (procval->expressed (procval procval?))
  (object->expressed (object object?))
  (list->expressed (list (list-of expval?)))
  (symbol->expressed (symbol symbol?))
  )

(define expressed->number
  (lambda (ev)
    (cases Expressed-Value ev
      (number->expressed (number) number)
      (else (error "expressed->number passed non-number argument: " ev)))))

(define expressed->procval
  (lambda (ev)
    (cases Expressed-Value ev
      (procval->expressed (procval) procval)
      (else (error "expressed->procval passed non-procval argument: " ev)))))

(define expressed->object
  (lambda (ev)
    (cases Expressed-Value ev
      (object->expressed (object) object)
      (else (error "expressed->object passed non-object argument: " ev)))))

(define expressed->list
  (lambda (ev)
    (cases Expressed-Value ev
      (list->expressed (list) list)
      (else (error "expressed->list passed non-list argument: " ev)))))

(define expressed->symbol
  (lambda (ev)
    (cases Expressed-Value ev
      (symbol->expressed (symbol) symbol)
      (else (error "expressed->symbol passed non-symbol argument: " ev)))))

(define procval->expressed?
  (lambda (ev)
    (cases Expressed-Value ev
      (procval->expressed (procval) #t)
      (else #f))))

(deftype expressed->printable (-> (Expressed-Value) datum))
(define expressed->printable
  (lambda (ev)
    (cases Expressed-Value ev
      (number->expressed (number) (has-type datum number))
      (procval->expressed (procval) (has-type datum procval))
      (object->expressed (object) (has-type datum object))
      (list->expressed (list)
	 (has-type datum (map expressed->printable list)))
      (symbol->expressed (symbol) (has-type datum symbol))
      )))

;; adaptation of environment-5-4-2 to this file
(deftype extend-env-recursively
  (-> ((list-of symbol) (list-of (list-of symbol)) (list-of expression)
       environment)
      environment))
(define extend-env-recursively
  (extend-env-recursively-maker
   (lambda (ids body env)
     (procval->expressed (closure ids body env)))
   (procval->expressed (closure '() (lit-exp 0) (empty-env)))))

(require (lib "ignore-mod.scm" "lib342"))

;;;;;;;;;;;;;;;; the interpreter ;;;;;;;;;;;;;;;;

(deftype eval-program (-> (program) datum))
(define eval-program 
  (lambda (pgm)
    (cases program pgm
      (a-program (c-decls body)
        (elaborate-class-decls! c-decls)
        (expressed->printable (eval-expression body (init-env)))))))

(deftype eval-expression (-> (expression environment) Expressed-Value))
(define eval-expression 
  (lambda (exp env)
    (cases expression exp
      (lit-exp (datum)
	(number->expressed datum))
      (quote-exp (sym)
        (symbol->expressed sym))
      (var-exp (id)
	(apply-env env id))
      (primapp-exp (prim rands)
        (let ((args (eval-rands rands env)))
          (apply-primitive prim args)))
      (if-exp (test-exp true-exp false-exp)
        (if (true-value? (eval-expression test-exp env))
          (eval-expression true-exp env)
          (eval-expression false-exp env)))
      (begin-exp (exp1 exps)
        (let loop ((acc (eval-expression exp1 env))
                   (exps exps))
          (if (null? exps) acc
            (loop (eval-expression (car exps) env) (cdr exps)))))
      (let-exp (ids rands body)
        (let ((args (eval-rands rands env)))
          (eval-expression
	   body
	   (extend-env ids args env))))
      (proc-exp (ids body) (procval->expressed (closure ids body env)))
      (app-exp (rator rands)
        (let ((proc (eval-expression rator env))
              (args (eval-rands rands env)))
          (if (procval->expressed? proc)
            (apply-procval (expressed->procval proc) args)
            (eopl:error 'eval-expression
              "Attempt to apply non-procedure ~s" proc))))
      (letrec-exp (proc-names idss bodies letrec-body)
        (eval-expression letrec-body
          (extend-env-recursively proc-names idss bodies env)))     
      (varassign-exp (id rhs-exp)
        (begin
          (setref! 
            (apply-env-ref env id)
            (eval-expression rhs-exp env))
          (number->expressed 1)))
      (new-object-exp (class-name rands)
        (let ((args (eval-rands rands env))
              (obj (new-object class-name)))
          (ignore
	   (find-method-and-apply
            'initialize class-name obj args))
          (object->expressed obj)))
      (method-app-exp (obj-exp method-name rands)
        (let ((args (eval-rands rands env))
              (obj (expressed->object (eval-expression obj-exp env))))
          (find-method-and-apply
            method-name (object->class-name obj) obj args)))
      (super-call-exp (method-name rands)
        (let ((args (eval-rands rands env))
              (obj (expressed->object (apply-env env 'self))))
          (find-method-and-apply
            method-name
	    (expressed->symbol (apply-env env '%super))
	    obj
	    args)))
      ;; typecase exercise
      (typecase-exp (tested-exp class-names formals bodies)
	(let* ((obj (expressed->object (eval-expression tested-exp env)))
	       (dynamic-class (object->class-name obj)))
	  (letrec
	      ((test-them
		(lambda (class-names formals bodies)
		  (cond
		   ((null? class-names)
		    (eopl:error 'eval-expression "Typecase falls off end!"))
		   ((is-subclass-of? dynamic-class (car class-names))
		    (eval-expression
		     (car bodies)
		     (extend-env (list (car formals))
				 (list (object->expressed obj))
				 env)))
		   (else (test-them (cdr class-names) (cdr formals)
				    (cdr bodies)))))))
	    (test-them class-names formals bodies))))
      ;; clone-upto exercise
      (clone-upto-exp (obj-exp class-name)
        (let* ((obj (expressed->object (eval-expression obj-exp env)))
	       (dynamic-class (object->class-name obj)))
	  (if (not (is-subclass-of? dynamic-class class-name))
	      (eopl:error
	       'eval-expression
	       "Cannot clone an object of class ~s upto class ~s!"
	       dynamic-class class-name)
	      (object->expressed
	        (an-object class-name (vector-copy (object->fields obj)))))))
      (else (eopl:error 'eval-expression "Missing case:~s" exp))
      )))

;; for both of the exercises
(deftype is-subclass-of? (-> (symbol symbol) boolean))
(define is-subclass-of?
  (lambda (cn1 cn2)
    (or (eq? cn1 cn2)
	(and (not (eq? cn1 'object))
	     (is-subclass-of? (class-name->super-name cn1) cn2)))))

;; clone-upto exercise
(deftype vector-copy
  (forall (T) (-> ((vector-of T)) (vector-of T))))
(define vector-copy
  (lambda (v)
    (let ((res (make-vector (vector-length v))))
      (letrec ((loop
		(lambda (i)
		  (if (negative? i)
		      res
		      (begin
			(vector-set! res i (vector-ref v i))
			(loop (- i 1)))))))
	(loop (- (vector-length v) 1))))))

(deftype eval-rands
  (-> ((list-of expression) environment) (list-of Expressed-Value)))
(define eval-rands
  (lambda (rands env)
    (map (lambda (x) (eval-expression x env)) rands)))

(deftype apply-primitive
  (-> (primitive (list-of Expressed-Value)) Expressed-Value))
(define apply-primitive
  (lambda (prim args)
    (define (check arity)
      (if (not (= arity (length args)))
          (eopl:error 'apply-primitive
                 "wrong number of arguments to primitive ~s"
                 prim)))
    (cases primitive prim
      (add-prim ()
        (check 2)
        (number->expressed (+ (expressed->number (car args))
                              (expressed->number (cadr args)))))
      (subtract-prim ()
        (check 2)
        (number->expressed (- (expressed->number (car args))
                              (expressed->number (cadr args)))))
      (mult-prim ()
        (check 2)
        (number->expressed (* (expressed->number (car args))
                              (expressed->number (cadr args)))))
      (incr-prim ()
        (check 1)
        (number->expressed (+ (expressed->number (car args)) 1)))
      (decr-prim ()
        (check 1)
        (number->expressed (- (expressed->number (car args)) 1)))
      (zero-test-prim()
        (check 1)
        (boolean->expressed (zero? (expressed->number (car args)))))
      (equal-prim ()
        (check 2)
        (boolean->expressed
         (cases Expressed-Value (car args)
           (number->expressed (n1)
             (= n1 (expressed->number (cadr args))))
           (symbol->expressed (s1)
             (eq? s1 (expressed->symbol (cadr args))))
           (else (eopl:error
                  'apply-primitive
                  "Can only compare numbers and symbols with equal?, not ~s"
                  (car args))))))
      (greater-prim () 
        (check 2)
        (boolean->expressed (> (expressed->number (car args))
                               (expressed->number (cadr args)))))
      (less-prim () 
        (check 2)
        (boolean->expressed (< (expressed->number (car args))
			       (expressed->number (cadr args)))))
      (list-prim ()
        ;; any number of arguments is ok
        (list->expressed args))
      (nil-prim ()
        (check 0)
        (list->expressed '()))
      (car-prim () 
        (check 1)
        (car (expressed->list (car args))))
      (cdr-prim ()
        (check 1)
        (list->expressed (cdr (expressed->list (car args)))))
      (cons-prim () 
        (check 2)
        (list->expressed (cons (car args) (expressed->list (cadr args)))))
      (null?-prim ()
        (check 1)
        (boolean->expressed (null? (expressed->list (car args)))))
      )))

(deftype init-env (-> () environment))
(define init-env 
  (lambda ()
    (empty-env)))
