;;; $Id: typedscm-tool.ss,v 1.6 2005/03/30 13:08:27 dorn Exp $
;;; AUTHOR: adapted by Gary T. Leavens from ../eopl/eopl-tool.ss,
;;;         extended by Brian Dorn to integrate the typechecker with
;;;           DrScheme GUI.

(module typedscm-tool mzscheme
  (require

   ;; Requires for being able to type-check things
   (lib "tc-scheme-parser.scm" "lib342")
   (lib "tc-scheme-annotate.scm" "lib342")
   (lib "tc-types.scm" "lib342")
   (lib "tc-env.scm" "lib342")
   (lib "displayln.scm" "lib342")
   (lib "tc-error-output.scm" "lib342")
   (lib "tc-position.scm" "lib342")

   ;; Requires for the tool definition
   (lib "mred.ss" "mred")
   (lib "framework.ss" "framework")
   (lib "list.ss")
   (lib "unitsig.ss")
   (lib "class.ss")
   (lib "tool.ss" "drscheme")
   (lib "string-constant.ss" "string-constants"))
  
  (provide tool@)

  (define tool@
    (unit/sig drscheme:tool-exports^
      (import drscheme:tool^)

      (define language-base%
	(class* object% (drscheme:language:simple-module-based-language<%>)
	  (define/public (get-language-numbers)
	    '(-500 -300))
	  (define/public (get-language-position)
	    (list (string-constant teaching-languages)
                  "Typedscm, typed extension to EoPL(2e)"))
	  (define/public (get-module)
	    '(lib "typedscm.ss" "lib342"))
	  (define/public (get-one-line-summary)
	    "A type notation for the Friedman, Wand, and Haynes text")
          (define/public (get-language-url)
	    "http://www.cs.iastate.edu/~leavens/ComS342/")
	  (define/public (get-reader)
	    (lambda (name port offsets)
	      (let ([v (read-syntax name port offsets)])
		(if (eof-object? v)
		    v
		    (namespace-syntax-introduce v)))))
	  (super-instantiate ())))

      (define language%
	(class (drscheme:language:module-based-language->language-mixin
		(drscheme:language:simple-module-based-language->module-based-language-mixin
		 language-base%))
	  (rename [super-on-execute on-execute])
	  (define/override (use-namespace-require/copy?) #t)
	  (define/override (on-execute settings run-in-user-thread)
	    (super-on-execute settings run-in-user-thread)
	    ;(run-in-user-thread
	    ; (lambda ()
	    ;   ((namespace-variable-value 'install-eopl-exception-handler))))
	    )
	  (super-instantiate ())
          )
        )

      (define unit-frame-mixin
	(mixin ((class->interface drscheme:unit:frame%)) ()
	       (inherit get-button-panel
			get-interactions-text
			get-definitions-text
			get-definitions-canvas
			get-directory
			begin-container-sequence
			end-container-sequence)

	       (super-instantiate ())

	       (rename [super-disable-evaluation disable-evaluation]
                  [super-enable-evaluation enable-evaluation])

	       (define/override (enable-evaluation)
		 (send type-check-button enable #t)
		 (super-enable-evaluation))

	       (define/override (disable-evaluation)
		 (send type-check-button enable #f)
		 (super-disable-evaluation))
	       
	       (define type-check-callback
		 (lambda (b e)
		   ;; Setup the parser so we get nice positions
		   (tc:start-input-filename! "<definitions>")
		   
		   ;; Typecheck program
		   (let* ((scheme-pgm
			   (tc:character-string-program-parser
			    (send (get-definitions-text) get-text 0 'eof #f #f)))
			  (result-type
			   (tc:annotation->type
			    (tc:scheme-top-level-annotate (tc:env-empty) scheme-pgm))))
		     (if (tc:error-type-expr? result-type)
			 (begin
			   (displayln "-------------BEGIN ERROR REPORT----------------")
			   (tc:output-error-message result-type)
			   (displayln "--------------END ERROR REPORT-----------------"))))

		   ;; Cleanup parser so things work next time
		   (tc:end-input-filename!)
		   ))
	       
	       (send (get-button-panel) begin-container-sequence)
	       (define type-check-button
		 (instantiate button% ()
			      (label ((drscheme:unit:make-bitmap
				       "Type Check"
				       (build-path (collection-path "lib342") "typedscm-icon-small.gif")) this))
			      (parent (get-button-panel))
			      (callback (lambda (b e) (type-check-callback b e)))))
	       (send (get-button-panel) change-children
		     (lambda (loc)
		       (cons type-check-button
			     (filter (lambda (c) (not (eq? type-check-button c))) loc))))
	       (send (get-button-panel) end-container-sequence)
	       ))

      (define (phase1)
        (void))
      (define (phase2)
	(drscheme:language-configuration:add-language 
	 (make-object ((drscheme:language:get-default-mixin) language%))))

      (drscheme:get/extend:extend-unit-frame unit-frame-mixin)
      )
    )
  )
